/**
*** CacheDb.java
*** 
*** @author	Larry Tung
*** @version	$Id: CacheDb.java,v 1.0 2001/07/12 00:25:15 lhtung Exp $
**/

package ostore.cache;

import ostore.util.Assert;
import ostore.util.Carp;
import com.sleepycat.db.*;

class CacheDb {

  private static final String Filename = "cache.db";
  private static boolean DEBUG = ostore.util.DebugFlags.CACHE_DB;
  private Db _table;
  private static LRUCache cache = LRUCache.getInstance();

  /**
   * set up the home directory
   */
  public CacheDb(String home) throws DbException 
  {
    super(home, null, 0); // TODO: may need to add more flags (ie. TXN) 
    // may need to change arguments above for transactional support

    try {
      // Set the OutputStream used for displaying Berkeley DB error messages
      set_error_stream(System.err);

      // TODO - use app_init to configure home and tmp directories.

      // Set the prefix string that appears before error messages 
      // included by Berkeley DB
      set_errpfx("CacheDb");  
    }
    catch( Exception e) {
      System.err.println(e.toString());
      System.exit(1);
    }
  }

  private static void usage() 
  {
    System.err.println("usage: CacheDb [-h home]\n");
    System.exit(1);
  }

  /**
   * Allow user to set home directory
   */
  public static void main(String argv[])
  {
    String home = null;
    for (int i = 0; i < argv.length; ++i) {
      if (argv[i].equals("-h")) {
	home = argv[++i];
      }
      else {
	usage();
      }
    }
    
    try {
      CacheDb app = new CacheDb(home);
      app.run();
    }
    catch( DbException dbe ) {
      System.err.println("CacheDb: " + dbe.toString());
      System.exit(1);
    }
    System.exit(0);
  }

  public void run() throws DbException
  {
    try {
      _table = Db.open(Filename, Db.DB_BTREE, Db.DB_CREATE, 0644, 
		       this, null);
    }
    catch( DbException dbe ) {
      errExit(dbe, "Open of account file failed");
    }
    // create some test strings to see if its working properly
    int num = 10;
    String[] lines = new String[num];
    byte[] buffers = new byte[num];
    Dbt[] keys = new Dbt[num];
    Dbt[] data = new Dbt[num];
    Dbt[] answers = new Dbt[num];

    for (int i = 0; i < num; ++i) {
      lines[i] = new String(i); //data is 0-9
      buffers[i] = lines[i].getBytes();
      keys[i] = new Dbt(buffers[i]);
      keys[i].set_size(buffers[i].length);
      data[i] = new Dbt( (lines[i]+lines[i]).getBytes() );
      data[i].set_size( (lines[i]+lines[i]).getBytes().length );

      try {
	// do not care about duplicates (default behavior of put is desired)
	_table.put(null, keys[i], data[i], 0);
      }
      catch( DbException dbe) {
	System.out.println("put failed: " + dbe.toString());
      }

      // do not need cursors since we are not allowing duplicates
      // for cacheables, will know size due to handle
      answers[i] = new byte[data[i].length];
      System.out.println("getting objects from database:");
      _table.get(null, keys[i], answers[i], 0);
      System.out.println(new String(answers[i]));
    }

    try {
      _table.close(0); // TODO: can throw DbRunRecoveryException
    }
    catch( DbException dbe2 ) {
      errExit(dbe2, "Account file close failed");
    }
  }

  static void errExit(DbException err, String s)
  {
    System.err.print("CacheDb: ");
    if (s != null) {
      System.err.print(s + ": ");
    }
    /*
    System.err.println(err.toString());
    System.exit(1);
    */
    Carp.die(err.toString());
  }

  // Stuff to put into CacheToDb (Translate from CacheToDisk)
  

  /**
   * Puts a CacheLine into the Database
   */
  static void put(CacheLine line) {
    Assert.assert(line != null, "put: line was null");
    if (line.getId() == null) return;
    if (line.backedUp()) return; //TODO - return, or overwrite?
    int len = line.getId().pairSize();
    //TODO naive method, put in all possible ways
    for (int i = 0; i < len; ++i) {
      SecureHash shash = (SecureHash) (line.getId().getName(i));
      byte[] filename = fullString(shash, line.getId().getType(i)).getBytes();
      
      Assert.assert(i < CacheableIdentity.numTypes, 
		    "write: too many attributes");
      try {
	Dbt key = new Dbt(filename); 
	key.set_size(filename.length);

	byte[] obj = QSIO.qsToBytes(line.getCacheable());
	Dbt data = new Dbt(obj);
	data.set_size(obj.length);

	_table.put(null, key, data, 0); //returns 0, DB_KEYEXIST, or exception

	line.hasBeenBackedUp();
      } 
      catch( DbException dbe ) {
	System.err.println("could not put " + line);
      }
    }
  }  

  /**
   * returns block if it finds it from the cache or in DB.
   * if it is not in the cache, but on DB, place the 
   * block in the cache. Return null if the block is not 
   * in the cache and DB.
   */ 
  static CacheableHandle get(CacheableIdentity id) {
    Assert.assert(id != null, "read: id was null");
    CacheableHandle handle = cache.get(id);
    if (handle == null) { // no longer in cache
      int len = id.pairSize(); 
      Assert.assert(len <= CacheableIdentity.numTypes, 
		    "read: too many attributes in id");
      for (int i = 0; i < len; ++i) {
	SecureHash shash = (SecureHash) id.getName(i);
	byte[] filename = fullString(shash, id.getType(i)).getBytes();

	Dbt key = new Dbt(filename);
	key.set_size(filename.length);
	Dbt data = new Dbt();
	data.set_flags(Db.DB_DBT_MALLOC);
	
	// recreate the object
	try {
	  int result = _table.get(null, key, data, 0);
	  if (result == DB_NOTFOUND || result == DB_KEYEMPTY) {
	    if (DEBUG) {
	      System.err.println("Could not find " + new String(filename) +
				 " DB_NOTFOUND: " + result == DB_NOTFOUND + 
				 " DB_KEYEMPTY: " + result == DB_KEYEMPTY);
	    }
	    return null;
	  }
	  int offset[] = { 0 };
	  int len = ByteUtils.bytesToInt( data, offset );
	  int type = ByteUtils.bytesToInt( data, offset );
	  offset[0] = 0;
	  object = TypeTable.new_object(type, data, offset);
	  return cache.put(id, block);
	}
	catch( DbException dbe ) {
	  //TODO - exit for now
	  errExit(dbe, "get " + id + " failed ");
	}
	catch( QSException qse ) {
	  //TODO - exit for now
	  errExit(dbe, "get " + id + " failed - bad object ");
	}
      }
      return null;
    }
  }

  /**
   * Deletes block with id from the Database.
   */
  static boolean delete(CacheableIdentity id) {
    Assert.assert(id != null, "delete: id was null");
    int len = id.pairSize();
    Assert.assert(len <= CacheableIdentity.numTypes, 
		  "delete: too many attributes in id");
    for (int i = 0; i < len; ++i) {
      SecureHash shash = (SecureHash) id.getName(i);
      byte[] filename = fullString(shash, id.getType(i)).getBytes();

      Dbt key = new Dbt(filename);
      key.set_size(filename.length);
      
      try {
	return (_table.del(null, key, 0) == 0);  // 0 for success 
      }
      catch( DbException dbe ) {
	//System.err.println("CacheDb::delete: " + id + "\t" + dbe.toString());
	//TODO exit for now
	errExit(dbe, "delete " + id + " failed ");
      }
    }
  }


  /**
   * Converts all bytes of this SecureHash to a human-readable String
   *
   * @param type
   *    id of object to be written to disk
   * @param shash 
   *    SecureHash to be used to write to disk
   * @return 
   *    human-readable String representation of this Hash
   *    containing every nibble of this Hash's value and prepending 
   *    a path to it, in this case /temp/
   *    tag indicates what type of data is being written to disk
   */
  public static java.lang.String fullString(SecureHash shash, int type) { 
    // copied and modified from SHA1Hash.java::fullString()
    Assert.assert(shash != null, "fullString: shash was null");
    byte[] bytes = shash.bytes();
    String s = path + type +
      ByteUtils.print_bytes( bytes, 0, bytes.length );
    return s;
  }    

	
}








