/**
 * GnutellaOrder.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: GnutellaOrder.java,v 1.3 2002/11/15 23:50:25 strib Exp $
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.apps.gnutella_trace;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Vector;
import java.util.StringTokenizer;
import ostore.util.Carp;
import ostore.util.PriorityQueue;

/**
 * This class reads in a gnutella trace, and generates a priority queue
 * of events, denoting when nodes enter and leave the trace.
 */

public class GnutellaOrder {

    private PriorityQueue pq;
    private int minAlive;
    private int maxAlive;
    private int atLeastNodes;

    // in ms
    private long minStart;
    private long maxStart;
    private long totalTime;
    private long atLeastNodesTime;
    private long exactNodesTime;

    // the up-times for each individual up session
    private Vector up_times;

    // the down-times for each individual down session
    private Vector down_times;
    
    public GnutellaOrder( String traceFilename, long startTime, 
			  int numNodes ) {

	PriorityQueue statPQ = new PriorityQueue( 100 );
	up_times = new Vector();
	down_times = new Vector();

	// read in file
	BufferedReader br = null;
	try {
	    br = new BufferedReader( new FileReader( traceFilename ) );
	} catch( FileNotFoundException fnfe ) {
	    Carp.die( "Could not open trace file: " + fnfe );
	}

	// start parsing
	String line;
	StringTokenizer st;
	String hashed_ip;
	int num_intervals;
	totalTime = -1;
	try {
	    while( (line = br.readLine()) != null ) {
		st = new StringTokenizer( line );
		hashed_ip = st.nextToken();
		num_intervals = Integer.parseInt( st.nextToken() );
		double start_time, end_time = 0, up_length = 0;
		GnutellaNode start_node, end_node;
		// read in each interval, create a start/end node for each
		for( int i = 0; i < num_intervals; i++ ) {
		    // get the time in ms
		    start_time = 1000*Double.parseDouble( st.nextToken() );
		    if( i == 0 ) {
			// if this is the first time on the line, it is the
			// first downtime instance
			down_times.add( new Double( start_time ) );
		    } else {
			// otherwise, the down time was the difference
			// from the last end_time
			// (note: we haven't changed end_time yet, so it's ok)
			down_times.add( new Double( start_time - end_time ) );
		    }
		    end_time = 1000*Double.parseDouble( st.nextToken() );
		    up_length = end_time - start_time;
		    up_times.add( new Double( up_length ) );
		    // make the nodes
		    start_node = new GnutellaNode(hashed_ip, start_time, true);
		    end_node = new GnutellaNode(hashed_ip, end_time, false);
		    // insert into the temp queue
		    statPQ.insert( start_node, (long) start_time, false );
		    statPQ.insert( end_node, (long) end_time, false );
		}
		if( end_time > totalTime ) {
		    totalTime = (long) end_time;
		}
	    }
	} catch( IOException ioe ) {
	    Carp.die( "Could not readLine in gnutella trace: " + ioe );
	}

	pq = new PriorityQueue( statPQ.size() );

	minAlive = Integer.MAX_VALUE;
	maxAlive = -1;
	atLeastNodes = -1;
	atLeastNodesTime = -1;	
	exactNodesTime = -1;
	
	int currAlive = 0;
	long old_prio = -1;
	// go through the queue, and keep track of how many are alive
	// at any one point.
	while( !statPQ.empty() ) {
	    
	    long prio = statPQ.topPriority(false);
	    GnutellaNode currNode = (GnutellaNode) statPQ.extract(false);

	    if( currNode.is_start ) {
		currAlive++;
	    } else {
		currAlive--;
	    }

	    // we only care about the stats starting from startTime
	    if( prio >= startTime /*&& prio < totalTime*/) {
		// insert it into the main pq
		pq.insert( currNode, prio, false );

		// find the first time there are at least the right number
		// of nodes available
		if( atLeastNodesTime < 0 && currAlive >= numNodes ) {
		    atLeastNodes = currAlive;
		    atLeastNodesTime = prio;
		}

		// find the first time there are exactly the right
		// number of nodes available
		if( exactNodesTime < 0 && currAlive == numNodes ) {
		    exactNodesTime = prio;
		}

		// if these priorities are different, then we've reached a new
		// millisecond, so we can examine the times again.
		// we need to treat things that happen in the same millisecond
		// as happening simultaneously
		if( prio != old_prio ) {
		    if( currAlive < minAlive ) {
			minAlive = currAlive;
			minStart = prio;
		    }
		    if( currAlive > maxAlive ) {
			maxAlive = currAlive;
			maxStart = prio;
		    }
		    
		    old_prio = prio;
		    
		}
		
	    }
	}

	// make the totalTime be relative to the start
	totalTime -= startTime;

    }

    public GnutellaOrder( String traceFilename, long startTime ) {
	this( traceFilename, startTime, 1 );
    }
 
    public GnutellaOrder( String traceFilename ) {
	this( traceFilename, 0 );
    }

    /**
     * Maximum number of nodes alive at any one time in the trace
     */
    public int maxAlive() {
	return maxAlive;
    }

    /**
     * Minimum number of nodes alive at any one time in the trace
     */
    public int minAlive() {
	return minAlive;
    }

    /**
     * First time when maximum number of nodes alive at any 
     * one time in the trace
     */
    public long maxAliveStart() {
	return maxStart;
    }

    /**
     * First time when minimum number of nodes alive at any 
     * one time in the trace
     */
    public long minAliveStart() {
	return minStart;
    }

    /**
     * First time when at least the desired number of nodes are alive.
     */
    public long atLeastAliveTime() {
	return atLeastNodesTime;
    }

    /**
     * First time when exactly the desired number of nodes are alive.
     */
    public long exactAliveTime() {
	return exactNodesTime;
    }


    /**
     * Return the queue of ordered events 
     */
    public PriorityQueue getQueue() {
	return pq;
    }

    public void printPDFHelper( Vector times ) {

	// first, set the granularity of the pdf/cdf:

	/*
	///////////  HOURS  ///////////////
	int ms_per = 1000 //ms per sec
	    * 60 // sec per min 
	    * 60; // min per hour
	    int per_hour = 1;
	//////////////////////////////////
	*/

	///////////  MINS  ///////////////
	int ms_per = 1000 //ms per sec 
	    * 60; // sec per min
	int per_hour = 60;
	//////////////////////////////////

	// an array with a spot for each of 60 hours (or mins)
	int pdf[] = new int[61*per_hour];

	//////////////  UPTIMES /////////////////////
	
	// to get the PDF, go through all the up_times and 
	// hash them into buckets based on how long they were up
	for( int i = 0; i < times.size(); i++ ) {
	    
	    double time = ( (Double) times.get(i) ).doubleValue();

	    // round it to a unit
	    int unit = (int) ( time/ms_per );
	    pdf[unit]++;

	}

	System.out.println( "PDF is:" ); 
	for( int i = 0; i < pdf.length; i++ ) {
	    System.out.print( ((double) pdf[i])/((double)times.size()) 
			      + ", " );
	}
	System.out.println( "\n" );

	System.out.println( "CDF is:" ); 
	for( int i = 0; i < pdf.length; i++ ) {
	    double cdf_value = 0;
	    for( int j = 0; j <= i; j++ ) {
		cdf_value += ((double) pdf[j])/((double)times.size());
	    }
	    System.out.print( cdf_value + ", " );
	}
	System.out.println( "" );


    }

    private void printPDF() {

	System.out.println( "//////////////  FAIL_TIMES:" );
	printPDFHelper( up_times );
	System.out.println( "//////////////  START_TIMES:" );
	printPDFHelper( down_times );

    }

    public static void main( String[] args ) {

	// a little test.  figure out max and min for intervals_ip
	GnutellaOrder go = 
	    new GnutellaOrder( args[0] );
	System.out.println( "Max nodes alive at once is " + go.maxAlive() + 
			    ", starting at " + go.maxAliveStart() + " ms.");
	System.out.println( "Min nodes alive at once is " + go.minAlive() + 
			    ", starting at " + go.minAliveStart() + " ms.");
	System.out.println( "Exact nodes available at " + go.exactAliveTime() 
			    + " ms." );
	System.out.println( "At least nodes available at " 
			    + go.atLeastAliveTime() + " ms." );
	go.printPDF();
	
    }

    /**
     * An inner class representing some event (start or end)
     * in a Gnutella trace.
     */
    public class GnutellaNode {

	public String hashed_ip;
	// in ms
	public double event_time;
	public boolean is_start;

	public GnutellaNode( String hashed_ip, double event_time, 
			     boolean is_start ) {
	    
	    this.hashed_ip = hashed_ip;
	    this.event_time = event_time;
	    this.is_start = is_start;

	}

	public String toString() {
	    return "(IP: " + hashed_ip + ", event_time: " + event_time +
		", is_start " + is_start + ")";
	}

    }

}
