/**
 * ClockAlgorithm.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package ostore.cache;
import ostore.util.AssertionViolatedException;
import ostore.util.Debug;

public class ClockAlgorithm implements CachePolicy
{
  ////////////////////       CONSTANTS		////////////////////
  public static final boolean DEBUG = false;
  public static final String tag = "ClockAlgorithm";
  ////////////////////    STATIC VARIABLES 	////////////////////
  ////////////////////       VARIABLES		////////////////////

  protected CachePolicy.Line _next_victim;
  protected int _size;

  ////////////////////  STATIC CONSTRUCTORS	////////////////////
  ////////////////////      CONSTRUCTORS	////////////////////
  ////////////////////     STATIC METHODS 	////////////////////
  public static void main(String [] args)
  {
    CachePolicy c = new ClockAlgorithm();
    System.out.println(c);
    c.insert(new TestLine(1, false));
    System.out.println(c);
    c.insert(new TestLine(2, false));
    System.out.println(c);
    c.insert(new TestLine(3, false));
    System.out.println(c);
    System.out.println(c.select_victim(true));
    System.out.println(c);
    System.out.println(c.select_victim(true));
    System.out.println(c);
    System.out.println(c.select_victim(true));
    System.out.println(c);
  }
  //////////////////// 		METHODS		////////////////////

  /* Specified by ostore.cache.CachePolicy */
  synchronized public int size() { return _size; }


  /* Specified by ostore.cache.CachePolicy */
  synchronized public boolean empty() { return _next_victim == null; }

  /* Specified by ostore.cache.CachePolicy */
  synchronized public void insert(CachePolicy.Line value)
  {
    if(DEBUG) Debug.printtagln(tag + ".insert", value.toString());

    if(_next_victim == null)
    {
      value._next = value;
      value._prev = value;
      _next_victim = value;
    }
    else 
    {
      CachePolicy.Line bef = _next_victim._prev;
      CachePolicy.Line aft = _next_victim;

      value._next = aft;
      value._prev = bef;
      
      bef._next = value;
      aft._prev = value;
    }

    //value.set_referenced();
    _size += 1;
  }

  /* Specified by ostore.cache.CachePolicy */
  synchronized public CachePolicy.Line select_victim(boolean modified_ok)
  {
    if(_next_victim == null) 
	throw new AssertionViolatedException(
		    "ClockAlgorithm.remove: must be non-empty.");

    CachePolicy.Line start = _next_victim;
    int max_value = -1;
    
    CachePolicy.Line result = null;
    while(result == null)
    {
      ++max_value;
      CachePolicy.Line walker = start;
      do {
	  if(DEBUG) Debug.printtagln(tag, walker.toString());
	  
	  // Value	Meaning
	  // 00		not referenced or modified, best to replace
	  // 01		not referenced but modified, next best
	  // 10		referenced but not modified, next again
	  // 11		referenced and modified, worst to replace
	  //
	  // See page 311 of Operating System Concepts, 5th Edition
	  // by Silberschatz and Galvin.
	  
	  int value =(walker.referenced() ? 2 : 0) |
	     (walker.modified()   ? 1 : 0);
	  if(value <= max_value) {
	      result = walker;
	      break;
	  }
	  walker.unset_referenced();
	  walker = walker._next;
      }
      while(walker != start);
    }
    
    if((! result.modified()) || modified_ok) {
	remove(result);
	return result;
    }
    
    return null;
  }

  /* Specified by ostore.cache.CachePolicy */
  synchronized public void remove(CachePolicy.Line value)
  {
    if(DEBUG) Debug.printtagln(tag + ".remove", value.toString());

    if(_next_victim == null) 
	throw new AssertionViolatedException(
		    "ClockAlgorithm.remove: must be non-empty.");

    try 
    {
      value._next._prev = value._prev;
    } 
    catch(NullPointerException npe)
    {
      ostore.util.Carp.die(tag + ".remove: " + 
	        "1) value=" + value +
		" value._next=" + 
		(value==null?null:value._next) +
	        " value._prev=" + 
		(value==null?null:value._prev) +
		" value._next._prev=" + 
		(value==null?null:(value._next==null?null:value._next._prev)) +
		" value._next._next=" + 
	        (value==null?null:(value._next==null?null:value._next._next)) +
		" value._prev._next=" + 
	        (value==null?null:(value._prev==null?null:value._prev._next)) +
		" value._prev._prev=" + 
		(value==null?null:(value._prev==null?null:value._prev._prev)));
    }
    
    try 
    {
      value._prev._next = value._next;
    } 
    catch(NullPointerException npe)
    {
      ostore.util.Carp.die(tag + ".remove: " + 
		"2) value=" + value +
		" value._next=" + 
		(value==null?null:value._next) +
		" value._prev=" + 
	        (value==null?null:value._prev) +
		" value._next._prev=" + 
		(value==null?null:(value._next==null?null:value._next._prev)) +
		" value._next._next=" + 
		(value==null?null:(value._next==null?null:value._next._next)) +
	        " value._prev._next=" + 
	        (value==null?null:(value._prev==null?null:value._prev._next)) +
		" value._prev._prev=" + 
		(value==null?null:(value._prev==null?null:value._prev._prev)));
    }

    if(_next_victim == value)
    {
      if(_next_victim._next == _next_victim)
	  _next_victim = null;
      else
	  _next_victim = value._next;
    }

    _size -= 1;
    
    value._next = value._prev = null;
  }

  /* Specified by java.lang.Object */
  synchronized public String toString()
  {
    if(_next_victim == null) return "Empty ClockAlgorithm";

    String result = "ClockAlgorithm: " + _size + " lines, contents:";
    CachePolicy.Line start = _next_victim;
    CachePolicy.Line walker = start;
    do
    {
      result += "\n" + walker + "\t" + 
	 (walker.modified() ? "1" : "0") +
	 (walker.referenced() ? "1" : "0");
      walker = walker._next;
    }
    while(walker != start);

    return result;
  }

  ////////////////////     INNER CLASSES	////////////////////

  private static class TestLine extends CachePolicy.Line
  {
    ////////////////////     CONSTANTS		////////////////////
    ////////////////////  STATIC VARIABLES 	////////////////////
    ////////////////////     VARIABLES		////////////////////
    public int id;

    ////////////////////    CONSTRUCTORS	////////////////////
    public TestLine(int i, boolean m) { super(m); id = i; }

    ////////////////////	STATIC METHODS 	////////////////////
    //////////////////// 	  METHODS	////////////////////
    public String toString() { return "TestLine " + id; }
  }
}

