/**
 * AuthenticationMsg.java
 *
 * Copyright (c) 2002 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network;
import ostore.util.NodeId;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QSException;
import ostore.util.Carp;

/**
 * The first message sent across a new connection in
 * <code>YANetwork</code>.  Replaces <code>IdentityMessage</code>.
 * <p>
 * This message informs the receiver of the sender's primary contact id.
 * The receiver can then eliminate redundant connections.
 * This message is required because outgoing connections do not in general
 * come from the sender's listen port, and also because virtual node ids
 * may listen on several ports in addition to their "primary" port.
 * <p>
 * This message also optionally includes certificates mapping virtual node
 * ids to the primary contact id.  These certificates are included if and
 * only if the sending node hosts multiple virtual nodes.
 * <p>
 * Eventually the content of this message will be signed by the sender, and
 * the sender's key will be bound to its <code>NodeId</code> using some
 * sort of certificate authority.  The message should also include some
 * sort of timestamp or counter to prevent replay attacks.
 * Only then will the message securely authenticate the sender.
 *
 * @author	Dennis Geels
 * @version	$Id: AuthenticationMsg.java,v 1.1 2003/02/07 00:11:34 geels Exp $
 */

public class AuthenticationMsg extends NetworkMessage
{
    ////////////////////       CONSTANTS	////////////////////
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The primary id to use to contact the sender.   Equal to the true id
     * of one of the virtual nodes on the sender's machine, probably the
     * sender itself.
     */
    private NodeId sender_contact_id;

    /**
     * The address and port used by this specific connection, which may be
     * different from the sender's contact id.  It should always be
     * different if the sender initiated the connection.
     */
    private NodeId this_connection_id;

    /**
     * A non-empty array of <code>VirtualizedNetCert</code>s, one per
     * virtual node on this  machine.
     */
    private VirtualizedNetCert[] certs;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Creates a message mapping the sender's <code>InetAddress</code> and
     * port number for a specific connection to the sender's primary
     * contact <code>NodeId</code>. 
     */
    public AuthenticationMsg( NodeId peer, NodeId local_contact_id,
			      NodeId this_connection_id,
			      VirtualizedNetCert[] certs )

    {
	super( peer, false /* not inbound */ );
	this.sender = this_connection_id;  // usually set by Classifier.
	this.sender_contact_id = local_contact_id;
	this.this_connection_id = this_connection_id;
	this.certs = certs;
	if( (certs == null) || (certs.length == 0) ) {
	    Carp.die( "missing certs" );
	}
    }
    
    /**
     * Constructs an <code>AuthenticationMsg</code> from its
     * <code>serialize</code>d form.
     */
    public AuthenticationMsg( InputBuffer buffer ) throws QSException
    {
	super( buffer );
	this.sender_contact_id = (NodeId)buffer.nextObject();
	this.this_connection_id = (NodeId)buffer.nextObject();
	int num_auths = buffer.nextInt();
	if( num_auths < 1 ) {
	    throw new QSException( "missing certs" );
	} else {
	    this.certs = new VirtualizedNetCert[num_auths];
	    for( int i=0; i<num_auths; i++ ) {
		this.certs[i] = (VirtualizedNetCert)buffer.nextObject();
	    }
	}
    }
			   
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Returns the primary id to use to contact the sender of this message.
     */
    public NodeId get_contact_id()
    {
	return sender_contact_id;
    }

    /**
     * Returns the address and port used by this specific connection, which
     * may be different from the sender's contact id. 
     */
    public NodeId get_connection_id()
    {
	return this_connection_id;
    }

    /**
     * Returns an array of <code>VirtualizedNetCert</code> objects.  Each declares
     * that the specified virtual node can be contacted at the primary id.
     * <p>
     * Returning the whole array is not as safe as cloning it, so be
     * careful not to modify the array itself.
     */
    public VirtualizedNetCert[] get_virtualized_net_certs()
    {
	return certs;
    }
    
    /* Specified by ostore.util.QuickSerializable */
    public void serialize( OutputBuffer buffer )
    {
	super.serialize( buffer );
        buffer.add( sender_contact_id );
	buffer.add( this_connection_id );
	buffer.add( certs.length );
	for( int i=0; i<certs.length; i++ ) {
	    buffer.add( certs[i] );
	}
    }

    /* Specified by ostore.network.NetworkMessage */
    public Object clone() throws CloneNotSupportedException
    {
	AuthenticationMsg dup = (AuthenticationMsg)super.clone();
	dup.sender_contact_id = (NodeId)this.sender_contact_id.clone();
	dup.this_connection_id = (NodeId)this.this_connection_id.clone();
	dup.certs = new VirtualizedNetCert[this.certs.length];
	System.arraycopy( this.certs, 0, dup.certs, 0, this.certs.length );
	return dup;
    }

    /**
     * Returns a human-readable representation of this message.
     */
    public String toString()
    {
	return( "(AuthenticationMsg connection: " + this_connection_id +
		" contact: " + sender_contact_id + " with " +
		certs.length + " VirtualizedNetCerts)" );
    }
}
