/**
 * ConnectionStream.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.network;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.SortedMap;
import java.util.TreeMap;
import seda.sandStorm.lib.aSocket.ATcpInPacket;
import ostore.util.AssertionViolatedException;
import ostore.util.ByteUtils;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.InputBuffer;

/**
 * Used by the network in place of an aSocketInputStream.  This is done for
 * two reasons.  First, the aSocketInputStream never frees any data, and it
 * wasn't immediately clear how to make it do so, mainly b/c the
 * implementation is poor, using a array of arrays rather than a linked
 * list of arrays.  I couldn't immediately figure out how to fix it.
 * Second, for small messages that fit within a packet, it's interface
 * didn't allow for the original packet byte array to be returned to the
 * caller, whereas this classes interface does.  We thus avoid a copy for
 * small messages.
 */
public class ConnectionStream {

    private static final boolean DEBUG = DebugFlags.NETWORK_NETWORK;
    private static final String tag = "ConnectionStream";

    private SortedMap out_of_order_packets = new TreeMap ();
    private LinkedList in_order_packets = new LinkedList ();
    private int next_seq_num = 1;
    private int [] packet_offset = {0};
    private int bytes_available;

    public InputBuffer get_next_message_buffer () {
	int size = read_int ();
	ATcpInPacket packet = (ATcpInPacket) in_order_packets.getFirst ();
	NetworkInputBuffer result = 
	    new NetworkInputBuffer (packet.getSequenceNumber ());
	result.add_packet (packet);
	result.set_first_packet_offset (packet_offset [0]);
	int remaining = size - packet.size () + packet_offset [0];
	while (remaining > 0) {
	    in_order_packets.removeFirst ();
	    packet = (ATcpInPacket) in_order_packets.getFirst ();
	    result.add_packet (packet);
	    remaining -= packet.size ();
	}
	if (remaining == 0) {
	    in_order_packets.removeFirst ();
	    packet_offset [0] = 0;
	}
	else
	    packet_offset [0] = packet.size () + remaining;

	result.set_available (size);
	bytes_available -= size;
	return result;
    }

    public boolean message_available () {
	if (bytes_available < 4) {
	    if (DEBUG) Debug.printtagln (tag, "only " +
		    bytes_available + " bytes available");
	    return false;
	}

	int size = peek_int ();
	if (DEBUG) Debug.printtagln (tag, "msg size is " +
		size + " bytes, " + (bytes_available - 4) + " avail");

	if (bytes_available - 4 < size) {
	    if (DEBUG) Debug.printtagln (tag, "only " +
		    (bytes_available - 4) + " bytes available");
	    // Whole message is not yet here.
	    return false;
	}

	return true;
    }

    public int next_message_size () {
	if (! message_available ()) 
	    throw new AssertionViolatedException ("! message_available");
	int size = peek_int ();
	return size;
    }

    private int peek_int () {
	if (bytes_available < 4) throw new AssertionViolatedException (
		"must have at least four bytes to peek_int");
	ATcpInPacket packet = (ATcpInPacket) in_order_packets.getFirst ();
	if (packet.size () - packet_offset [0] >= 4) {
	    int res = ByteUtils.bytesToInt (packet.getBytes (), packet_offset);
	    packet_offset [0] -= 4;
	    return res;
	}
	else {
	    byte [] bytes = new byte [4];
	    int rem = 4;
	    int boffset = 0;
	    int poffset = packet_offset [0];
	    Iterator i = in_order_packets.iterator ();
	    while (rem > 0) {
		packet = (ATcpInPacket) i.next ();
		byte [] pbytes = packet.getBytes ();
		int toread = Math.min (pbytes.length - poffset, rem);
		System.arraycopy (pbytes, poffset, bytes, boffset, toread);
		poffset = 0;
		boffset += toread;
		rem -= toread;
	    }
	    int [] offset = {0};
	    return ByteUtils.bytesToInt (bytes, offset);
	}
    }

    private int read_int () {
	final int count = 4;
	ATcpInPacket packet = (ATcpInPacket) in_order_packets.getFirst ();
	byte [] out_buffer = null;
	int [] out_offset = {0};

	if (count <= packet.size () - packet_offset [0]) {
	    // If they ask for less than a packet's worth, just give it to
	    // them.

	    out_buffer = packet.getBytes ();
	    out_offset [0] = packet_offset [0];
	    packet_offset [0] += count;

	    if (packet_offset [0] == packet.size ()) {
		in_order_packets.removeFirst ();
		packet_offset [0] = 0;
	    }
	}
	else {
	    // If they ask for more, assemble up a new byte array for them.
	    if (count > bytes_available)
		throw new AssertionViolatedException (
			"not enough data available to read_int");

	    out_buffer = new byte [count];
	    out_offset [0] = 0;
	    int remaining = count;
	    while (remaining > 0) {
		int toread = Math.min (packet.size () - packet_offset [0], 
		                       remaining);
		System.arraycopy (packet.getBytes (), packet_offset [0],
			out_buffer, out_offset [0], toread);
		out_offset [0] += toread;
		packet_offset [0] += toread;
		remaining -= toread;

		if (packet_offset [0] == packet.size ()) {
		    in_order_packets.removeFirst ();
		    if (remaining != 0)
			packet = (ATcpInPacket) in_order_packets.getFirst ();
		    packet_offset [0] = 0;
		}
	    }
	    out_offset [0] = 0;
	}

	bytes_available -= count;
	return ByteUtils.bytesToInt (out_buffer, out_offset);
    }

    public void add_packet (ATcpInPacket packet) {
	long seq = packet.getSequenceNumber ();
	if (seq == next_seq_num) {
	    in_order_packets.addLast (packet);
	    ++next_seq_num;
	    bytes_available += packet.size ();
	}
	else {
	    out_of_order_packets.put (new Long (seq), packet);
	}	

	while (! out_of_order_packets.isEmpty ()) {
	    Long smallest_key = (Long) out_of_order_packets.firstKey ();
	    if ((smallest_key).longValue () != next_seq_num) 
		break;
	    packet = (ATcpInPacket) out_of_order_packets.remove (smallest_key);
	    in_order_packets.addLast (packet);
	    bytes_available += packet.size ();
	    ++next_seq_num;
	}
    }
}


