/**
 * NetworkOutputBuffer.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.core.BufferElement;
import seda.sandStorm.lib.aSocket.ATcpConnection;
import ostore.util.AssertionViolatedException;
import ostore.util.SecureHash;
import ostore.util.OutputBufferImpl;

/**
 * Efficient buffer.
 *
 * @author	Sean C. Rhea
 * @version	$Id: NetworkOutputBuffer.java,v 1.9 2002/09/08 07:37:18 srhea Exp $
 */
public class NetworkOutputBuffer extends OutputBufferImpl {
    
    private static final int MIN_SIZE = 1024;

    private static class Snippet {
	public boolean read_only;
	public Snippet (int size) { 
	    bel = new BufferElement (new byte [size], 0, 0);
	}
	public Snippet () { 
	    bel = new BufferElement (new byte [MIN_SIZE], 0, 0);
	}
	public Snippet (byte [] input, int offset, int length) { 
	    bel = new BufferElement (input, offset, length);
	    read_only = true;
	}
	public BufferElement bel;
	public Snippet next;
    }

    private Snippet _head, _tail;
    private int _size;

    public int size () {
	return _size;
    }

    public void add (byte input) {
	if (_head == null) {
	    _head = _tail = new Snippet ();
	}
	else if ((_tail.read_only) || 
	         (_tail.bel.size == _tail.bel.data.length)) {
	    Snippet s = new Snippet ();
	    _tail.next = s; 
	    _tail = s;
	}

	if (_tail.read_only) 
	    throw new AssertionViolatedException ("should be read/write");

	_tail.bel.data [_tail.bel.size++] = input;
	++_size;
    }

    public void add (byte [] input, int offset, int length) {
	if ((length >= MIN_SIZE) && (_head != null)) {
	    _tail = _tail.next = new Snippet (input, offset, length);
	}
	else {
	    if (_head == null) {
		_head = _tail = new Snippet ();
	    }
	    else if (_tail.read_only) {
		_tail = _tail.next = new Snippet ();
	    }

	    if (_tail.read_only) 
		throw new AssertionViolatedException ("should be read/write");

	    int n = Math.min (length, _tail.bel.data.length - _tail.bel.size);	
	    System.arraycopy (input, offset, _tail.bel.data, _tail.bel.size, n);
	    _tail.bel.size += n;

	    if (n < length) {
		Snippet s = new Snippet (Math.max (length - n, MIN_SIZE));
		System.arraycopy (input, offset + n, s.bel.data, 0, length - n);
		s.bel.size = length - n;
		_tail.next = s;
		_tail = s;
	    }
	}

	_size += length;
    }

    public void overwrite (int offset, int value) {
	if ((_head == null) || (offset != 0))
	    throw new AssertionViolatedException ("Not yet implemented.");
	
	if (_head.read_only) 
	    throw new AssertionViolatedException ("should be read/write");

	for (int j = 3; j >= 0; --j) {
	    _head.bel.data [offset++] = ((byte) (0xff & (value>>(8*j))));
	}
    }

    public void enqueue (ATcpConnection conn) throws SinkException {
	Snippet walker = _head;
	while (walker != null) {
	    //if (! walker.read_only) 
	//	System.out.println ("No copy, size = " + walker.bel.size);
	    conn.enqueue (walker.bel);
	    walker = walker.next;
	}
    }

    public String toString () {
	String result = "(NetworkOutputBuffer ";
	Snippet walker = _head;
	while (walker != null) {
	    result += ostore.util.ByteUtils.print_bytes (
		    walker.bel.data, walker.bel.offset, walker.bel.size);
	    walker = walker.next;
	}
	return result + ")";
    }
}


