/**
 * Copyright (c) 2002 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.network;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.core.ssTimer;
import java.net.InetAddress;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * Allows a SandStorm stage to ping a machine by hostname and not block.
 * Enqueue a PingReq and wait for a PingSuccess or PingFailure in response.
 *
 * @author Sean C. Rhea
 * @version $Id: PingStage.java,v 1.2 2002/11/18 23:39:38 srhea Exp $
 */
public class PingStage implements seda.sandStorm.api.EventHandlerIF {

    public static class PingReq implements QueueElementIF {
	public String hostname;
	public Object user_data;
	public SinkIF return_address;
	public int count;
	public long timeout_ms;
	public PingReq (String h, Object u, SinkIF r, int c, long t) { 
	    hostname = h;  
	    user_data = u; 
	    return_address = r;
	    count = c;
	    timeout_ms = t;
	}
    }

    public static class PingSuccess implements QueueElementIF {
	public String hostname;
	public Object user_data;
	public double ping_time_ms;
	public PingSuccess (PingReq req, double p) {
	    hostname = req.hostname;
	    user_data = req.user_data;
	    ping_time_ms = p;
	}
    }
    
    public static class PingFailure implements QueueElementIF {
	public String hostname;
	public Object user_data;
	public PingFailure (PingReq req) {
	    hostname = req.hostname;
	    user_data = req.user_data;
	}
    }

    protected Set timeouts = new HashSet ();
    protected ssTimer timer = new ssTimer ();
    protected SinkIF sink;

    public void init(ConfigDataIF config) throws Exception {
	sink = config.getStage().getSink();
    }

    public void destroy() {}

    public void handleEvents(QueueElementIF element_array[]) 
    throws EventHandlerException {
	for (int i = 0; i < element_array.length; ++i)
	    handleEvent(element_array[i]);
    }

    protected static class Alarm implements QueueElementIF { 
	public Thread thread;
	public Alarm (Thread t) { thread = t; }
    }

    protected void error (PingReq req) {
	PingFailure resp = new PingFailure (req);
	enqueue (req.return_address, resp);
    }

    public void handleEvent(QueueElementIF item) 
    throws EventHandlerException {
	if (item instanceof PingReq) {
	    PingReq req = (PingReq) item;

	    double pingtime = -1.0;
	    try {
		String cmd = "ping -q -c " + req.count + " " + req.hostname;
		Process p = Runtime.getRuntime().exec (cmd);
		Thread t = Thread.currentThread ();
		synchronized (timeouts) {
		    timeouts.add (t);
		}

		ssTimer.ssTimerEvent token = timer.registerEvent (
			req.timeout_ms, new Alarm (t), sink);
		p.waitFor();
		timer.cancelEvent (token);

		synchronized (timeouts) {
		    if (! timeouts.remove (t))
			t.interrupted (); // clear flag
		}
		try {
		    if (p.exitValue () != 0) {
			System.err.println ("ping " + req.hostname + 
				" returned non-zero exit code " + 
				p.exitValue ());
			error (req);
			return;
		    }
		}
		catch (IllegalThreadStateException e) {
		    System.err.println ("ping " + req.hostname + " timed out");
		    error (req);
		    return;

		}
		InputStream s = p.getInputStream();
		byte[] store = new byte[400]; // TODO
		s.read(store);
		String str = new String(store);
		StringTokenizer st = new StringTokenizer(str);
		while (st.hasMoreTokens()) {
		    String nt = st.nextToken();
		    if (nt.equals("min/avg/max") || 
			nt.equals("min/avg/max/mdev")) {// TODO: ugly

			st.nextToken();
			st.nextToken("/");
			String val = st.nextToken("/");
			pingtime = (new Double(val)).doubleValue();
			break;
		    }
		}
	    } 
	    catch (Exception e) {
		System.err.println ("ping " + req.hostname + 
			" caused exception " + e);
		e.printStackTrace (System.err);
		error (req);
		return;
	    }

	    if (pingtime == -1.0) {
		ostore.util.Carp.die ("ping " + req.hostname + 
			" failed due to bad parser.  Please tell Sean.");
	    }
	    else {
		PingSuccess resp = new PingSuccess (req, pingtime);
		enqueue (req.return_address, resp);
	    }
	}
	else if (item instanceof Alarm) {
	    Alarm alarm = (Alarm) item;
	    synchronized (timeouts) {
		if (timeouts.remove (alarm.thread)) {
		    // it's still waiting
		    alarm.thread.interrupt ();
		}
	    }
	}
	else {
	    System.err.println ("unknown event: " + item);
	    System.exit (1);
	}
    }

    protected void enqueue (SinkIF sink, QueueElementIF item) {
	try {
	    sink.enqueue (item);
	}
	catch (SinkException e) {
	    System.err.println ("ERROR: couldn't enqueue " + item +
		    "onto " + sink + ".  Dropping it.");
	}
    }

    public static class TestHarness 
    implements seda.sandStorm.api.EventHandlerIF {

	protected SinkIF sink, ping_sink;

	public void init(ConfigDataIF config) throws Exception {
	    sink = config.getStage().getSink();
	    try {
		ping_sink = config.getManager ().createStage (
			"PingStage", new PingStage (), null).getSink ();
	    }
	    catch (Exception e) {
		System.err.println ("Caught " + e + " creating ping stage.");
		e.printStackTrace (System.err);
		System.exit (1);
	    }
	    String hostname = config.getString ("hostname");
	    ping_sink.enqueue (new PingReq (hostname, null, sink, 1, 5000));
	}

	public void destroy() {}

	public void handleEvents(QueueElementIF element_array[]) 
	throws EventHandlerException {
	    for (int i = 0; i < element_array.length; ++i)
		handleEvent(element_array[i]);
	}

	public void handleEvent(QueueElementIF item) 
	throws EventHandlerException {
	    if (item instanceof PingSuccess) {
		PingSuccess resp = (PingSuccess) item;
		System.out.println ("Success: " + resp.ping_time_ms + " ms");
		System.exit (0);
	    }
	    else if (item instanceof PingFailure) {
		System.out.println ("Failure.");
		System.exit (1);
	    }
	    else {
		System.err.println ("unknown event: " + item);
		System.exit (1);
	    }
	}
    }
}

