/**
 * VirtualizedNetCert.java
 *
 * Copyright (c) 2002 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network;
import ostore.util.NodeId;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QSException;

/**
 * Certifies that a virtual node can be contacted through a specified
 * proxy.  An <code>AuthenticationMsg</code> includes one
 * <code>VirtualizedNetCert</code> for each virtual node in the system.
 * <p>
 * Eventually should be signed.
 *
 * @author	Dennis Geels
 * @version	$Id: VirtualizedNetCert.java,v 1.1 2003/02/07 00:11:34 geels Exp $
 */
public class VirtualizedNetCert implements QuickSerializable
{
    ////////////////////       CONSTANTS	////////////////////
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The id of the virtual node.
     */
    private NodeId virtual_node_id;

    /**
     * The canonical id for all virtual nodes on this machine.
     */
    private NodeId contact_id;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Creates a new VirtualizedNetCert.
     */
    public VirtualizedNetCert( NodeId virtual_node_id, NodeId contact_id )
    {
	this.virtual_node_id = virtual_node_id;
	this.contact_id = contact_id;
    }
    
    /**
     * Constructs an <code>VirtualizedNetCert</code> from its
     * <code>serialize</code>d form.
     */
    public VirtualizedNetCert( InputBuffer buffer ) throws QSException
    {
	this.virtual_node_id = (NodeId)buffer.nextObject();
	this.contact_id = (NodeId)buffer.nextObject();
    }
    
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Returns the id of the virtual node.
     */
    public NodeId get_virtual_node_id()
    {
	return virtual_node_id;
    }

    /**
     * Returns the id at which to contact the virtual node.
     */
    public NodeId get_contact_id()
    {
	return contact_id;
    }
    
    /* Specified by ostore.util.QuickSerializable */
    public void serialize( OutputBuffer buffer )
    {
	buffer.add( virtual_node_id );
	buffer.add( contact_id );
    }

    /**
     * Returns a human-readable representation of this message.
     */
    public String toString()
    {
	return( "(VirtualizedNetCert virtual node: " + virtual_node_id +
		" contact id: " + contact_id + ")" );
    }
}
