/**
 * YANetwork.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network;

import ostore.cache.CachePolicy;
import ostore.cache.ClockAlgorithm;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.NodeId;
import ostore.util.Pair;
import ostore.util.InputBuffer;
import ostore.util.CountBuffer;
import ostore.util.TypeTable;
import ostore.util.QSClone;
import ostore.dispatch.Classifier;
import ostore.dispatch.Signal;
import ostore.dispatch.Filter;
import ostore.dispatch.Resender;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;
import seda.sandStorm.api.SinkClosedException;
import seda.sandStorm.api.SinkClosedEvent;
import seda.sandStorm.api.StagesInitializedSignal;
import seda.sandStorm.internal.SinkProxy;
import seda.sandStorm.lib.aSocket.ATcpConnection;
import seda.sandStorm.lib.aSocket.ATcpConnectFailedEvent;
import seda.sandStorm.lib.aSocket.ATcpClientSocket;
import seda.sandStorm.lib.aSocket.ATcpServerSocket;
import seda.sandStorm.lib.aSocket.ATcpServerSocketClosedEvent;
import seda.sandStorm.lib.aSocket.ATcpListenFailedEvent;
import seda.sandStorm.lib.aSocket.ATcpListenSuccessEvent;
import seda.sandStorm.lib.aSocket.ATcpInPacket;
import soss.network.TransitStub;
import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.StreamTokenizer;
import java.net.InetAddress;
import java.net.Socket;

/**
 * Yet Another Network stage.  This one provides the same general
 * interfaces as <code>MultiplexedNetwork</code>.  Internally, the main
 * differences are the addition of a connection cache, a handshake protocol
 * for closing connections, and tighter restrictions against multiple
 * connections between physical machines.
 * <p>
 * New authentication algorithm: The first message that each end of a
 * connection sends is a <code>AuthenticationMsg</code>.  It contains the
 * primary listening port of the sender.  Eventually it will contain a
 * certificate chain authenticating the sender and binding it to the
 * current connection and its listening port.
 * <p>
 * The connection cache includes remote-initiated connections, but does not
 * limit the number of connections accepted by the local server (beyond its
 * hardwired limit of 511).  Therefore a burst of incoming connections may
 * temporarily fill the cache beyond its desired limit.  The stage
 * maintains a free list (counter) of connections to minimize the effects
 * of these (or outgoing) bursts.  Unfortunately, remote machines can
 * effectively shutdown our outgoing communication by overfilling our
 * cache with hundreds of incoming connections.  Setting the connection
 * cache size higher than 511 will avoid this particular problem, although
 * a normal bandwidth-swamping DoS attack is still possible.
 *
 * @author	Dennis Geels
 * @version $Id: YANetwork.java,v 1.29 2003/06/19 21:38:06 srhea Exp $
 */

public class YANetwork implements SingleThreadedEventHandlerIF, EventHandlerIF
{
    ////////////////////       CONSTANTS	////////////////////

    private static final boolean DEBUG = DebugFlags.NETWORK_NETWORK;
    //    private static final boolean DEBUG = true;
    private static final boolean DEBUG_TRACE = false;
    private static final boolean DEBUG_VERBOSE = false;
    private static final String class_tag = "YANetwork";

    /**
     * The default value for the "max_conn" configuration variable.
     * Currently: 50.
     */
    private static final int DEFAULT_MAX_CONN = 50;

    /**
     * The default fraction of the connection cache to keep free whenever
     * possible.  Currently: 10%.
     */
    private static final double DEFAULT_FREE_FRACTION = 0.1;

    /**
     * The period at which we wake up and check the cache.
     */
    private static final int ALARM_PERIOD = 10000;

    /**
     * The number of milliseconds to wait for remote peers to send
     * important messages, like <code>AuthenticationMsg</code>s.
     */
    private static final int PEER_TIMEOUT = 30000;
    
    /**
     * The number of milliseconds to wait before trying to reconnect to
     * remote hosts, if <code>infinite_connect_retry==true</code>.
     */
    private static final long RECONNECT_DELAY = 5000;
    
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The sockets on which this stage listens for new connections.
     */
    private List server_sockets;

    /**
     * The maximum number of TCP connections that this node is allowed to
     * have open simultaneously (the size of our connection cache).
     */
    private int max_conn;

    /**
     * The minimum number of TCP connections that this node should have
     * free at all times (a free list for our connection cache).
     */
    private int min_free_conn;

    /**
     * Tracks references and selects victims for clearing connection cache.
     * Connections that are in the process of opening or closing should not
     * be in the clock.
     */
    private CachePolicy cache_clock;

    /**
     * Stores <code>ConnectionState</code> objects for connections in cache
     * that are eligible for ejection (those in <code>cache_clock</code>).
     * Compensates for the lack of a <code>get</code> method in
     * <code>CachePolicy</code> of type <code>ClockAlgorithm</code>.  
     */
    private HashSet connection_cache;

    /**
     * The current number of additional connections we are allowed to make.
     * This variable, <b>not</b> the size of the
     * <code>connection_cache</code> should be used to control admission
     * for new connections.
     */
    private int free_connections;

    /**
     * A list (oldest first) of <code>ConnectionState</code> objects for
     * connections we will make as soon as the connection cache 
     */
    private List waiting_connections;

    /**
     * A <code>ATcpConnection</code> to <code>ConnectionState</code> map
     * for all open and closing connections in the system.
     * Used to find state given references from underlying network.
     */
    private HashMap connection_states_by_connection;

    /**
     * A <code>InetAddress</code> to <code>ConnectionState</code> map for
     * connections that are open or waiting to open.
     * Used to find state given destination address in outbound message.
     */
    private HashMap connection_states_by_contact_id;

    /**
     * Used only by <code>get_contact_id</code>,
     * <code>set_contact_id</code> to name virtual nodes.
     * TODO: Flush occasonally?  Currently is a very slow memory leak.
     */
    private HashMap contact_ids;
    
    /**
     * A <code>ConnectionState</code> to <code>Object</code> map storing
     * incoming connections that are waiting for
     * <code>AuthenticationMsg</code>s.  They map to <code>Resender</code>
     * tokens that will timeout if the <code>AuthenticationMsg</code> does
     * not arrive.
     */
    private HashMap states_pending_authentication;
    
    /**
     * A <code>ConnectionState</code> to <code>Object</code> map storing
     * the connections we are currently trying to close.  They map to
     * <code>Resender</code> tokens that will timeout if the
     * <code>CloseConnectionMsg</code> does not arrive.  The closing
     * process involves a handshake with the remote machine and some delay
     * due to the asynchronous aSocket interface.
     */
    private HashMap closing_connections;

    /**
     * A <code>ATcpClientSocket</code> to <code>ConnectionState</code> map
     * for locally-initiated connections that were abandoned before
     * they opened.  We will close them immediately.
     * <p>
     * Assumes that <code>ATcpClientSocket</code> doesn't do anything
     * screwy with <code>hashCode</code> and <code>equals</code>.
     */
    private HashMap doa_connections;
    
    /**
     * The IP address and port number at which other machines should
     * contact virtual nodes served by this stage.  Equal to that of the
     * first virtual node.
     */
    private NodeId my_contact_id;
    
    /**
     * A <code>NodeId -> Classifier</code> map storing the classifier used
     * by each virtual node.  We use these classifiers to deliver incoming
     * packets.  
     */
    private HashMap node_id_to_classifier;

    /**
     * The main sink for this stage.  We reference this object in most of
     * our communication with the underlying network.
     */
    private SinkIF my_sink;

    /**
     * A <code>Resender</code> used to control how long we wait for certain
     * messages from remote peers, like <code>AuthenticationMsg</code>s.
     */
    private Resender resender;

    /**
     * The name of this stage's <code>Resender</code>.  Used to direct
     * alarm signals.
     */
    private String resender_name;
    
    /**
     * <code>true</code> iff we should delay all outbound messages
     * according to the latencies dictated by an artificial topology.
     */
    private boolean add_delays;
    
    /**
     * The data structure storing the parsed topology information, or
     * <code>null</code> when not using an artificial topology.
     */ 
    private TransitStub topology;

    /**
     * The <code>NodeId</code> to <code>Integer</code> map identifying the
     * destination machine in the artificial topology, or <code>null</code>
     * if <code>add_delays == false</code>.
     */
    private HashMap node_id_to_vertex;

    /**
     * <code>true</code> iff we should record the number of messages and
     * bytes sent and received from each peer.
     */
    private boolean record_peer_stats;
    
    /**
     * A map of maps from <code>NodeId,String</cdoe> pairs to
     * <code>PeerStats</code> objects used to track
     * bandwidth consumed per peer.
     */
    private HashMap peer_stats;

    /**
     * If <code>true</code>, attempt to reconnect on
     * <code>ATcpConnectFailedEvent</code>s and
     * <code>ATcpListenFailedEvent</code>s. 
     * This will attempt to reconnect repeatedly until
     * entire JVM process is terminated. 
     */
    private boolean infinite_connect_retry;

    /**
     * If <code>true</code>, cheat.  Assume that all virtual nodes whose
     * <code>NodeId</code>s have the same <code>InetAddress</code> are
     * hosted by the same network stage.  This simplification avoids the
     * need to open more than one connection to a physical machine,
     * allowing experiments to scale better.  It prevents starting nodes in
     * multiple JVMs or communicating through a NAT box.
     */
    private boolean assume_single_remote_network_stage;
    
    /**
     * The name by which to refer to this stage.  It should be
     * the name of the corresponding <code>StageIF</code>.
     */
    private String name;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /* Specified by seda.sandStorm.api.EventHandlerIF */
    public void handleEvent(QueueElementIF elem) throws UnknownEventException
    {
      try
      {
	final String tag = class_tag+".handleEvent";
	// Don't print out all events -- some contain the word "error".
	//	if( DEBUG_TRACE ) Debug.printtagln( tag, "called:\t" + elem );
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	// Choose the proper handler based on event type
	if( elem instanceof StagesInitializedSignal ) {
	    // Set the first alarm.
	    if( DEBUG ) Debug.printtagln( tag, "Setting first alarm" );	    
	    Signal.getSignal( my_contact_id ).alarm( name, ALARM_PERIOD );
	    
	} else if( elem instanceof Signal.AlarmSignal ) {
	    handle_alarm( (Signal.AlarmSignal)elem );

	} else if( elem instanceof ATcpListenSuccessEvent ) {
	    handle_listen_success( (ATcpListenSuccessEvent)elem );
	    
	} else if( elem instanceof ATcpListenFailedEvent ) {
	    handle_listen_failure( (ATcpListenFailedEvent)elem );
	    
	} else if( elem instanceof ATcpConnectFailedEvent ) {
	    handle_connect_failed( (ATcpConnectFailedEvent)elem );

	} else if( elem instanceof ATcpServerSocketClosedEvent ) {
	    handle_socket_closed( (ATcpServerSocketClosedEvent)elem );

	} else if( elem instanceof ATcpConnection ) {
	    handle_new_connection( (ATcpConnection)elem );
	    
	} else if( elem instanceof ATcpInPacket ) {
	    handle_incoming_packet( (ATcpInPacket)elem );

	} else if( elem instanceof NetworkMessage ) {
	    handle_outbound_message( (NetworkMessage)elem );

	} else if( elem instanceof CloseLocalPortMsg ) {
	    handle_close_local_port( (CloseLocalPortMsg)elem );

	} else if( elem instanceof DelayWrapper ) {
	    send_message( ((DelayWrapper)elem).to_send );

	} else if( elem instanceof SinkClosedEvent ) {
	    handle_sink_closed( (SinkClosedEvent)elem );

	} else if( elem instanceof ClearBandwidthCountsReq ) {
	    handle_clear_bw_req( (ClearBandwidthCountsReq)elem );

	} else if( elem instanceof RecordBytesMsg ) {
	    RecordBytesMsg m = (RecordBytesMsg)elem;
	    record_bytes( m.msg, m.bytes );
	    
	} else {
	    throw new UnknownEventException( "Not a known type", elem);
	}
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
      }
      catch(Exception e)
      {
	Iterator i = node_id_to_classifier.keySet().iterator();
        while(i.hasNext())
	  Carp.dieGracefully(e, (NodeId)i.next()); 
      }
    }
    
    /* Specified by seda.sandStorm.api.EventHandlerIF */
    public void handleEvents(QueueElementIF elemarr[])
	throws UnknownEventException    
    {
	final String tag = class_tag+".handleEvents";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called: "
				      + elemarr.length + " elements" );
	for( int i=0; i<elemarr.length; i++ ) {
	    handleEvent( elemarr[i] );
	}
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
    }

    /* Specified by seda.sandStorm.api.EventHandlerIF */
    public void init(ConfigDataIF config) throws Exception
    {
	final String tag = class_tag+".init";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	// Make sure every node calls this sometime.
	ostore.util.Types.registerAllTypes();
	// And register our new types
	TypeTable.register_type( "ostore.network.CloseConnectionMsg" );
	TypeTable.register_type( "ostore.network.AuthenticationMsg" );
	TypeTable.register_type( "ostore.network.VirtualizedNetCert" );
	
	////////////////////////////////
	// Initialize Private Variables

	name = config.getStage().getName();

	my_sink = config.getStage().getSink();
	if( config.getBoolean( "Profile" ) ) {
	    my_sink = new SinkProxy( my_sink,
				     config.getManager(),
				     config.getStage().getWrapper() );
	    config.getManager().getProfiler().add( "ostore.network.YANetwork",
						   (SinkProxy)my_sink );
	}

	// Set up connection cache:
	// Maximum number of connections
	this.max_conn = config.getInt( "max_conn" );
	if( max_conn <= 0 ) {
	    max_conn = DEFAULT_MAX_CONN;
	}
	if( DEBUG ) Debug.printtagln( tag, "using connection cache of size: "
				      + max_conn );
	double min_free_frac = config.getDouble( "min_free_frac" );
	if( (min_free_frac < 0) || (1 < min_free_frac) ) {
	    if( DEBUG ) Debug.printtagln( tag, "invalid min_free_frac: " +
					  min_free_frac );
	    min_free_frac = DEFAULT_FREE_FRACTION;
	}
	this.min_free_conn = (int)Math.ceil(max_conn * min_free_frac);
	if( DEBUG ) Debug.printtagln( tag, "using free list size: " +
				      min_free_conn );
	
	this.cache_clock = new ClockAlgorithm();
	this.connection_cache = new HashSet();
	this.free_connections = this.max_conn;
	this.waiting_connections = new LinkedList();
	this.connection_states_by_connection = new HashMap();
	this.connection_states_by_contact_id = new HashMap();
	this.states_pending_authentication = new HashMap();
	this.closing_connections = new HashMap();
	this.doa_connections = new HashMap();
	this.contact_ids = new HashMap();
	this.peer_stats = new HashMap();

	this.record_peer_stats = config.getBoolean( "bw_counting" );
	this.infinite_connect_retry =
	    config.getBoolean("infinite_connect_retry");
	this.assume_single_remote_network_stage =
	    config.getBoolean("single_network_hack");

	// NodeIds, Listen sockets, Classifiers for each virtual node.
	int node_id_count = config.getInt( "node_id_count" );
	if( node_id_count <= 0 ) Carp.die( tag + ": invalid node_id_count" );
	this.node_id_to_classifier= new HashMap( node_id_count );
	server_sockets = new LinkedList();	// Filled on SuccessEvents
	InetAddress local_address = null;
	
	for( int i=0; i<node_id_count; i++ ) {
	    NodeId n = new NodeId( config.getString( "node_id_" + i ) );
	    if( DEBUG ) Debug.printtagln (tag, "Local node_id: " + n );
	    // Sanity checks:
	    if( local_address == null ) {
		local_address = n.address();
	    } else if( ! local_address.equals( n.address() ) ) {
		Carp.die( "All NodeIds must have same address" );
	    }

	    if( node_id_to_classifier.containsKey( n ) ) {
		Carp.die( "Duplicate NodeId: " + n );
	    }
	    if( i == 0 ) {
		// Arbitrary choice: Any open port would do.
		my_contact_id = n;
	    }

	    Classifier c = Classifier.getClassifier( n );
	    node_id_to_classifier.put( n, c );
	    
	    int listen_port = n.port();
	    if( listen_port < 1024 ) {	// unspecified or privileged.
		Carp.die( "Invalid listen port: " + listen_port );
	    }
	    if( DEBUG ) Debug.printtagln (tag, "Opening server socket on port "
					  + listen_port + ".");
	    new ATcpServerSocket( listen_port, my_sink );
	}

	// Create a <code>Resender</code> to manage timeouts.
	// It will use the <code>Classifier</code> and <code>Signal</code>
	// of the primary <code>NodeId</code>.
	resender_name = name + "-resender";
	this.resender = new Resender( my_contact_id, resender_name );

	String topology_graph = config.getString ("topology_graph");
	String topology_mapping = config.getString ("topology_mapping");
	if( (topology_graph == null) ^ (topology_mapping == null) ) {
	    Carp.die( tag + ": must specify topology_graph " +
		      " and topology_mapping together.");

	} else if( topology_graph != null ) {
	    // We should use an artificial topology.
	    add_delays = true;
	    try {
		initialize_topology( topology_graph, topology_mapping );
	    } catch( IOException ioe ) {
		Carp.die( tag + ": could not create artificial topology" );
	    }
	}
	
	////////////////////////////////
	// Initialize Filters
	
	// Register for signals
	if( !Signal.getSignal(my_contact_id).register(Signal.INIT,
						      name, my_sink)) {
	    Carp.die( "could not register for Signal.INIT" );
	}
	if( !Signal.getSignal(my_contact_id).register(Signal.ALRM,
						      name, my_sink)) {
	    Carp.die( "could not register for Signal.ALRM" );
	}
	if( !Signal.getSignal(my_contact_id).register(Signal.ALRM,
						      resender_name, my_sink)){
	    Carp.die( "could not register resender for Signal.ALRM" );
	}

	// Now network-related events
	Filter filter = new Filter();
	if( !filter.requireType( "ostore.network.NetworkMessage" ) ||
	    !filter.requireValue( "inbound", new Boolean(false) ) ) {
	    Carp.die( "Could not build NetworkMessage filter." );
	}
	Filter cpFilter = new Filter();
	if( !cpFilter.requireType( "ostore.network.CloseLocalPortMsg" ) ) {
	    Carp.die( "Could not build CloseLocalPort filter." );
	}
	// Subscribe to outbound messages for each node_id.
	Iterator node_id_iterator = node_id_to_classifier.keySet().iterator();
	while( node_id_iterator.hasNext() ) {
	    NodeId n = (NodeId)node_id_iterator.next();
	    Classifier c = (Classifier)node_id_to_classifier.get( n );
	    if( !c.subscribe (filter, my_sink) ) {
		Carp.die( "Classifier does not recognize NetworkMessages" );
	    } else if( DEBUG ) {
		Debug.printtagln( tag, "subscribed to outbound " +
				  "NetworkMessages on " + n );
	    }
	    if( !c.subscribe (cpFilter, my_sink) ) {
		Carp.die( "Classifier does not recognize CloseLocalPortMsgs" );
	    } else if( DEBUG ) {
		Debug.printtagln( tag, "subscribed to " +
				  "CloseLocalPortMsgs on " + n );
	    }
	}
	if( record_peer_stats ) {
	    // Subscribe to bandwidth requests for each node_id.
	    filter = new Filter();
	    if(!filter.requireType("ostore.network.ClearBandwidthCountsReq")) {
		Carp.die( "Could not build NetworkMessage filter." );
	    }
	    // accept byte counts for messages we don't send out . . .
	    Filter rbFilter = new Filter();
	    if( !rbFilter.requireType( "ostore.network.RecordBytesMsg" ) ) {
		Carp.die( "Could not build RecordBytes filter." );
	    }

	    node_id_iterator = node_id_to_classifier.keySet().iterator();
	    while( node_id_iterator.hasNext() ) {
		NodeId n = (NodeId)node_id_iterator.next();
		Classifier c = (Classifier)node_id_to_classifier.get( n );
		if( !c.subscribe (filter, my_sink) ) {
		    Carp.die( "Classifier does not recognize " +
			      "ClearBandwidthCountsReqs" );
		} else if( DEBUG ) {
		    Debug.printtagln( tag, "subscribed to outbound " +
				      "ClearBandwidthCountsReqs on " + n );
		}
		if( !c.subscribe (rbFilter, my_sink) ) {
		    Carp.die( "Classifier does not recognize " +
			      "RecordByteMsgs" );
		} else if( DEBUG ) {
		    Debug.printtagln( tag, "subscribed to " +
				      "RecordByteMsgs on " + n );
		}
	    }
	}
	// Only subscribe to delayed messages on the first classifier..
	filter = new Filter();
	Classifier classifier =
	    (Classifier)node_id_to_classifier.get(my_contact_id);
	if( !filter.requireType( "ostore.network.YANetwork$DelayWrapper" ) ||
	    !classifier.subscribe(filter, my_sink) ) {
	    Carp.die( "Classifier does not recognize DelayWrappers" );
	} else if( DEBUG ) {
	    Debug.printtagln( tag, "subscribed to DelayWrappers" );
	}

	if( DEBUG_TRACE ) Debug.printtagln( tag, "done");
    }

    /**
     * Parses the artificial topology graph and maps <code>NodeId</code>s
     * to nodes in that topology.
     * <p>
     * Initializes the <code>topology</code> and
     * <code>node_id_to_vertex</code> global data structures.
     * <p>
     * Implementation taken directly from original Network stage.
     *
     * @param topology_graph	The name of the topology file (GT-ITM format)
     * @param topology_mapping	The name of the file mapping
     *		<code>NodeId</code>s to indices of the artificial topology.
     */
    private void initialize_topology( String topology_graph,
				      String topology_mapping )
	throws IOException
    {
	final String tag = class_tag+".initialize_topology";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );
	
	// Step 1: Parse the graph file.
	try {
	    topology = new TransitStub( topology_graph );
	} catch (IOException ioe) {
	    Carp.die( tag + ": caught exception when creating " +
		      "graph: " + ioe + ".");
	}
	
	// Step 2: Parse the mapping file.
	// Read the IP:Port to Vertex mapping
	node_id_to_vertex = new HashMap();
	StreamTokenizer tok = null;
	try {
	    BufferedReader reader =
		new BufferedReader( new FileReader( topology_mapping ) ); 
	    tok = new StreamTokenizer( reader );
	} catch (Exception e) {
	    Carp.die( tag + ": caught exception " + e + " trying " +
		      "to read topology mapping " + topology_mapping );
	}
	tok.resetSyntax ();
	tok.whitespaceChars ((int) ' ', (int) ' ');
	tok.wordChars ((int) 'a', (int) 'z');
	tok.wordChars ((int) 'A', (int) 'Z');
	tok.wordChars ((int) '0', (int) '9');
	tok.wordChars ((int) ':', (int) ':');
	tok.wordChars ((int) '.', (int) '.');
	tok.wordChars ((int) '-', (int) '-');
	
	int lineno = 1;
	tok.nextToken ();
	while (tok.ttype == StreamTokenizer.TT_EOL) {
	    tok.nextToken ();
	    lineno++;
	}

	while (tok.ttype != StreamTokenizer.TT_EOF) {
	    
	    // Read the IP:Port
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		Carp.die (tag + ": expected IP:Port on line " + 
			  lineno + ": " + tok);
	    }
	    NodeId nid = null;
	    try {
		nid = new NodeId (tok.sval);
	    } catch (Exception e) {
		Carp.die (tag + ": Bad IP:Port string on " +
			  "line " + lineno + ": " + e);
	    }
	    
	    // Read the Vertex
	    tok.nextToken ();
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		Carp.die (tag + ": expected IP:Port on line " + 
			  lineno + ": " + tok);
	    }
	    Integer vertex = null;
	    try {
		vertex = new Integer (Integer.parseInt (tok.sval));
	    } catch (NumberFormatException e) {
		Carp.die (tag + ": number format error reading " +
			  "vertex on line " + lineno + ": " + e);
	    }
	    
	    // Put them in the mapping.
	    if( node_id_to_vertex.containsKey(nid) ) {
		Carp.die (tag + ": duplicate node id " + nid +
			  " in topology mapping " + topology_mapping);
	    }
	    node_id_to_vertex.put( nid, vertex );
	    
	    tok.nextToken ();
	    while (tok.ttype == StreamTokenizer.TT_EOL) {
		tok.nextToken ();
		lineno++;
	    }
	}
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );	
    }

    /**
     * Checks state and restarts/retries whatever is necessary.
     */
    private void handle_alarm( Signal.AlarmSignal sig )
    {
	final String tag = class_tag+".handle_alarm";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	if( sig.target.equals( resender_name ) ) {
	    // Used by the <code>Resender</code> internally.
	    // Let it resend and/or time-out requests.
	    resender.process_queues();

	    // Now look for any that timed out.
	    QueueElementIF[] timed_out =
		resender.clear( false /* do not clear still-pending events */ );

	    if( timed_out.length > 0 ) {
		for( int i=0; i<timed_out.length; i++ ) {
		    ConnectionState to_kill = (ConnectionState)timed_out[i];
		    // We were either waiting for an <code>AuthenticationMsg</code> 
		    // or a <code>CloseConnectionMsg</code>.
		    if( states_pending_authentication.remove(to_kill)
			!= null ) {
			// Authentication timed out.
			Carp.carp( tag + ": authentication timed out: " +
				   to_kill );
			NodeId contact_id = get_contact_id( to_kill.peer_id );
			ConnectionState active_state = (ConnectionState)
			    connection_states_by_contact_id.get( contact_id );

			// Try to close cleanly, even though that will
			// probably just time out, too.
			close_connection( to_kill );
			if( active_state == to_kill ) {
			    // Nodes are effectively unreachable.
			    signal_death( to_kill );
			}
		    } else if( closing_connections.remove(to_kill) != null ) {
			// Close timed out.
			Carp.carp( tag + ": connection shutdown timed out: "
				   + to_kill );
			delete_connection_state( to_kill );
		    } else {
			Carp.die( tag + ": What timed out? :" + to_kill);
		    }
		}
	    }
	    
	} else {
	    // Just an alarm for the stage.  Keep books.
	    
	    // Close some connections if necessary.
	    prune_cache();
	    
	    // Print out a progress mark.
	    print_status();

	    // Set a new alarm.
	    Signal.getSignal( my_contact_id ).alarm( name, ALARM_PERIOD );
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Prints some brief stats on current connections.
     * Note that "connected" and "active" are different.
     */
    private void print_status()
    {
	final String tag = class_tag+".print_status";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	Debug.printtagln( tag, " fr: " + free_connections + /*free*/
			  " wa: " + waiting_connections.size() + /*waiting*/
			  " pa: " + /*pending authentication*/
			  states_pending_authentication.size() +
			  /*active (sendable virtual ids)*/
			  " ac: " + connection_states_by_contact_id.size() +
			  /*open*/
			  " co: " + connection_states_by_connection.size() +
			  " ic: " + connection_cache.size() + /*in cache*/
			  " cl: " + closing_connections.size() ); /*closing*/
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Prints out some useful information about known connections.
     * Currently prints out the peers for each active connection.
     */
    private void print_connections()
    {
	final String tag = class_tag+".print_connections";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	StringBuffer sbuf = new StringBuffer();
	Iterator conn_iter = connection_states_by_contact_id.keySet().iterator();
	while( conn_iter.hasNext() ) {
	    NodeId n = (NodeId)conn_iter.next();
	    sbuf.append( n.toString() );
	    sbuf.append( " " );
	}
	Debug.printtagln( tag, " peers: (" + sbuf + ")" );
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Records the newly opened server port and prints out a status
     * message.  That message is used by run scripts; do not modify it.
     */
    private void handle_listen_success( ATcpListenSuccessEvent event )
    {
	final String tag = class_tag+".handle_listen_success";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	ATcpServerSocket server_socket = event.getSocket();
	int server_port = server_socket.getPort();
	Carp.out( class_tag + " now listening on port " + server_port + "." );
	// Remember the socket for later shutdown.
	server_sockets.add( server_socket );
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Prints an error message and dies.  Almost always happens because
     * another server is on the same port.
     */
    private void handle_listen_failure( ATcpListenFailedEvent event )
    {
	final String tag = class_tag+".handle_listen_failure";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	ATcpServerSocket server_socket = event.getSocket();
	int listen_port = server_socket.getPort();

	Carp.carp( tag +  ": Could not listen on port " + listen_port + ".");
	if( infinite_connect_retry ) {
	    Carp.carp( tag +  ": Trying again." );
	    try {
		new ATcpServerSocket( listen_port, my_sink );
	    } catch( IOException ioe ) {
		Carp.die( ioe );	// Shouldn't ever happen.
	    }
	} else {
	    Carp.croak( tag + ": There is probably another " + 
			"program using the port.  Dying..." );
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }

    /**
     * Maps the <code>NodeId</code>s of virtual nodes to the primary
     * <code>NodeId</code> that represents all virtual nodes for a given
     * physical machine/network stage.
     */
    private NodeId get_contact_id( NodeId virtual_node )
    {
	final String tag = class_tag+".get_contact_id";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	NodeId contact_id;
	if( assume_single_remote_network_stage ) {
	    // HACK -- ignore the port number.
	    InetAddress addr = virtual_node.address();
	    contact_id = (NodeId)this.contact_ids.get(addr);
	    if( contact_id == null ) {
		// Unknown.  Use this id for this address from now on.
		contact_id = virtual_node;
		this.contact_ids.put( addr, contact_id );
	    }
	    
	} else {
	    // First check the hashtable
	    contact_id = (NodeId)this.contact_ids.get(virtual_node);
	    if( contact_id == null ) {
		// Unknown.  Use the id directly.
		contact_id = virtual_node;
	    }
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
	return contact_id;
    }
    
    /**
     * Adds a mapping from the <code>NodeId</code>s of a virtual node to
     * the primary <code>NodeId</code> that represents all virtual nodes
     * for its machine/network stage.
     */
    private void set_contact_id( NodeId virtual_node, NodeId contact_id )
    {
	final String tag = class_tag+".set_contact_id";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	// Just update the hashtable
	if( assume_single_remote_network_stage ) {
	    // HACK -- ignore the port number.
	    InetAddress addr = virtual_node.address();
	    this.contact_ids.put( addr, contact_id );	    
	} else {
	    this.contact_ids.put( virtual_node, contact_id );
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
    }
    
    /**
     * Deals with unreachable peers.
     */
    private void handle_connect_failed( ATcpConnectFailedEvent event)
    {
	final String tag = class_tag+".handle_connect_failed";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );
	
	ATcpClientSocket socket = event.getSocket();
	NodeId peer_id = new NodeId(socket.getPort(), socket.getAddress());
	if( DEBUG || DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
	    // TODO: Should be an error in most cases, but not at end of
	    // successful experiments.  For now, don't print "ERROR".
	    Carp.out( tag + ": could not connect to " + peer_id );
	}
	NodeId contact_id = get_contact_id( peer_id );
	ConnectionState state = (ConnectionState)
	    connection_states_by_contact_id.get( contact_id );

	if( ( state == null ) || 	
	    ( state.client_socket != socket ) ) {
	    // The failed connection has been superceded by a later one.
	    if( DEBUG ) Debug.printtagln( tag, "DOA" );
	    state = (ConnectionState)doa_connections.remove( socket );
	    if( state == null ) {
		// Could not find the connection that failed.
		Carp.die( tag + ": Unknown connection: " + contact_id);
	    }
	}
	// Now <code>state</code> points correctly to the connection
	// that failed.
	if( state.get_state() != ConnectionState.STATE_STARTING ) {
	    Carp.die( "Unexpected state: " + state );
	}

	// Retry or give up?
	if( infinite_connect_retry ) {
	    // Possibly just a temporary network problem.
	    // Try resending the messages in a bit.
	    List to_send = state.remove_pending();
	    if( to_send != null ) {
		Iterator pending_iter = to_send.iterator();
		while( pending_iter.hasNext() ) {
		    NetworkMessage msg = (NetworkMessage)pending_iter.next();
		    Classifier classifier = (Classifier)
			node_id_to_classifier.get(my_contact_id);
		    classifier.dispatch_later( msg, RECONNECT_DELAY );
		}
	    }
	} 
	// Clean up the failed connection's state.
	delete_connection_state( state );

	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Dispatches an error signal when a node becomes unreachable.
     */
    private void signal_death( ConnectionState state )
    {
	final String tag = class_tag+".signal_death";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );
	
	try {
	    // Send a message for each remote virtual node as well as the
	    // remote connection id, if distinct.
	    Set unreachable_nodes = new HashSet( state.authenticated_peers );
	    unreachable_nodes.add( state.peer_id );
	    Iterator peer_iter = unreachable_nodes.iterator();
	    while( peer_iter.hasNext() ) {
		NodeId peer = (NodeId)peer_iter.next();
		if( DEBUG ) Debug.printtagln( tag, "dead: " + peer );
		if( peer == null ) {
		    continue;
		}
		NetworkNodeStatusMsg obit = new NetworkNodeStatusMsg( peer );
		Iterator classifiers =
		    node_id_to_classifier.values().iterator();
		while( classifiers.hasNext() ) {
		    ((Classifier)classifiers.next()).dispatch( obit );
		}
	    }
	} catch( SinkFullException sfe ) {
	    Carp.die( sfe );
	}

	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Processes a newly opened connection.  Passes the connection to one
     * of two helper methods, depending on whether the connection was
     * initiated by this node or the remote peer.
     */
    private void handle_new_connection( ATcpConnection conn )
    {
	final String tag = class_tag+".handle_new_connection";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called: " + conn );

	if( conn.getClientSocket() != null ) {
	    // locally initiated.  Start using it.
	    handle_outgoing_connection( conn );

	} else {
	    // unexpected.  Wait for authentication.
	    handle_incoming_connection( conn );
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Finishes opening an outgoing connection.  First an
     * <code>AuthenticationMsg</code> is sent, so that the remote peer can
     * learn and verify our contact <code>NodeId</code>.
     * Then the <code>ConnectionState</code> object is marked as
     * "open" and pending messages are sent.
     */
    private void handle_outgoing_connection( ATcpConnection conn )
    {
	final String tag = class_tag+".handle_outgoing_connection";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	ATcpClientSocket client_socket = conn.getClientSocket();
	NodeId peer_id = new NodeId( conn.getPort(), conn.getAddress() );
	NodeId contact_id = get_contact_id( peer_id );
	if( DEBUG ) Debug.printtagln( tag, "opened to: " + contact_id +
				      "(" + peer_id + ")" );
	// Would be different from <code>peer_id</code> only if we received an
	// <code>AuthenticationMsg</code> on a duplicate connection, but
	// chose to keep this one instead.
	ConnectionState state =
	    (ConnectionState)connection_states_by_contact_id.get( contact_id );
	
	boolean is_doa = false;
	if( (state == null) || 	
	    ( state.client_socket != client_socket ) ) {
	    // We have (or had and then closed) a different (better)
	    // connection active for this address.  This connection
	    // must be DOA.  Close it.
	    if( DEBUG ) Debug.printtagln( tag, "DOA: " + conn );
	    is_doa = true;
	    state = (ConnectionState)doa_connections.remove( client_socket );
	    if( state == null ) {
		// Could not find this connection. 
		Carp.die( tag + ": Unexpected connection: " + conn );
	    }
	} else if( state.get_state() != ConnectionState.STATE_STARTING ) {
	    Carp.die( "unexpected state: " + state );
	}
	
	if( DEBUG ) Debug.printtagln( tag, "found state: " + state );
	// Store the connection for later.
	state.set_connection( conn );
	if( DEBUG ) Debug.printtagln( tag, "local connection id: " +
				      state.local_connection_id );
	connection_states_by_connection.put( conn, state );

	// Start receiving incoming packets
	conn.startReader( my_sink );

	if( is_doa ) {
	    close_connection( state );
	} else {
	    if( DEBUG ) Debug.printtagln(tag, "opening connection: " + state);
	    // We were waiting for this connection.  Send an
	    // <code>AuthentcationMsg</code> followed by any pending messages.
	    
	    // Add the new state to the cache clock algorithm
	    connection_cache.add( state );
	    state.set_referenced();	// clock algorithm update.
	    cache_clock.insert( state );
	
	    // Must fetch pending queue before advancing state:
	    List msgs_to_send = state.remove_pending();
	    // Mark the connection open.
	    state.set_state( ConnectionState.STATE_OPEN );

	    // Tell the remote peer our real name
	    Object token = send_authentication_message( state );
	    // Don't accept messages from remote peer until authenticated.
	    // TODO: We need not authenticate the peer that we addressed,
	    // just other virtual nodes on same machine.
	    states_pending_authentication.put( state, token );

	    // Now send all the pending messages.
	    if( msgs_to_send != null ) {
		Iterator pending_iter = msgs_to_send.iterator();
		while( pending_iter.hasNext() ) {
		    NetworkMessage msg = (NetworkMessage)pending_iter.next();
		    enqueue_message( conn, msg );
		}
	    }
	}		
	// Finally, consider making some room on free list.
	prune_cache(); 
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Accepts incoming connections.  A new state object is opened and 
     * we wait for an <code>AuthenticationMsg</code> from the remote peer.
     */
    private void handle_incoming_connection( ATcpConnection conn )
    {
	final String tag = class_tag+".handle_incoming_connection";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	// Create a new state object.
	NodeId peer_id = new NodeId( conn.getPort(), conn.getAddress() );
	if( DEBUG ) Debug.printtagln( tag, "opened by: " + peer_id );
	ConnectionState state = new ConnectionState( peer_id );	// STATE_WAITING
	state.set_connection( conn );
	if( DEBUG ) Debug.printtagln( tag, "local connection id: " +
				      state.local_connection_id );
	connection_states_by_connection.put( conn, state );
	
	// Reserve slot in cache
	free_connections--;	// Might be negative now.

	// Start receiving incoming packets
	conn.startReader( my_sink );

	// Send an <code>AuthenticationMsg</code> to the remote peer and
	// wait for its to arrive.
	Object token = send_authentication_message( state );

	if( DEBUG ) Debug.printtagln(tag, "on hold: " + state);
	states_pending_authentication.put( state, token );
	
	// Finally, consider making some room on free list.
	prune_cache(); 
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Constructs and enqueues an <code>AuthenticationMsg</code> for a
     * given connection
     * @return a token from the <code>Resender</code>.
     */
    private Object send_authentication_message( ConnectionState state )
    {
	final String tag = class_tag+".send_authentication_message";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called: " + state );
	
	// TODO: prepare (and sign) certs ahead of time.
	VirtualizedNetCert[] certs =
	    new VirtualizedNetCert[node_id_to_classifier.size()];
	Iterator virtual_node_iter =
	    node_id_to_classifier.keySet().iterator();
	int i=0;
	while( virtual_node_iter.hasNext() ) {
	    NodeId virtual_node_id = (NodeId)virtual_node_iter.next();
	    certs[i] = new VirtualizedNetCert( virtual_node_id,
					       my_contact_id );
	    i++;
	}
	AuthenticationMsg auth_msg =
	    new AuthenticationMsg( state.peer_id, my_contact_id,
				   state.local_connection_id, certs );
	enqueue_message( state.get_connection(), auth_msg );
	// Set a timeout that we will clear iff we receive an
	// <code>AuthenticationMsg</code> from the remote peer.
	Object token = resender.remember( state, PEER_TIMEOUT) ;
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
	return token;
    }
    
    /**
     * Verifies the contact id received from a peer.
     * Detects redundant connections and eliminates duplicates. 
     */
    private void process_authentication_message( ConnectionState state,
						 AuthenticationMsg msg )
    {
	final String tag = class_tag+".process_authentication_message";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called: " + msg  );

	if( msg.get_connection_id().equals( state.peer_id ) ) {
	    // First check the sender id and certificate count.
	    
	    // TODO: check signatures on message.
	    // For now, assume that it clears.

	    // Stop waiting for this message.
	    Object token = states_pending_authentication.remove(state);
	    if( token != null ) {
		resender.cancel( token );
	    } // else: we already authenticated this connection once.

	    // (Re)Assign connection to primary contact id
	    NodeId contact_id = msg.get_contact_id();
	    if( DEBUG ) Debug.printtagln( tag, "contact id: " + contact_id );

	    // Keep track of duplicate connections while setting this one up.
	    Set duplicate_connections = new HashSet();
	    ConnectionState collision = (ConnectionState)
		connection_states_by_contact_id.get( contact_id );
	    if( (collision != null) && (collision != state) ) {
		if( DEBUG ) Debug.printtagln( tag, "collision: " + collision );
		// resolve at end of this method.
		duplicate_connections.add( collision );
	    }

	    ////////////////////
	    // Process each virtual node's certificate
	    VirtualizedNetCert [] certs = msg.get_virtualized_net_certs();
	    for( int i=0; i<certs.length; i++ ) {
		if( ! contact_id.equals( certs[i].get_contact_id() ) ) {
		    if( DEBUG ) {
			Debug.printtagln( tag, "invalid contact id: "
					  + certs[i] + " vs. " + contact_id);
		    }
		    // Just ignore this cert.  Maybe others are valid?
		} else {
		    NodeId virtual_id = certs[i].get_virtual_node_id();
		    if( DEBUG ) Debug.printtagln( tag, "virtual node: " +
						  virtual_id );
		    // TODO: check signature on certificate?
		    state.authenticated_peers.add( virtual_id );
		    
		    NodeId old_contact_id = get_contact_id( virtual_id );
		    if( ! old_contact_id.equals( contact_id ) ) {
			// Moved to new address.
			if( DEBUG ) Debug.printtagln( tag, "moved: " +
						      virtual_id );
			// Remember new contact mapping
			set_contact_id( virtual_id, contact_id );

			// Check for duplicate (outgoing) connections.
			collision = (ConnectionState)
			    connection_states_by_contact_id.remove(old_contact_id);
			if( (collision != null) && (collision != state) ) {
			    if( DEBUG ) Debug.printtagln( tag, "collision: " +
							  collision );
			    // resolve at end of this method.
			    duplicate_connections.add( collision );
			}
		    }
		}
	    }
		
	    if( state.get_state() == ConnectionState.STATE_CLOSING ) {
		if( DEBUG ) Debug.printtagln(tag, "Should be closing: " +
					     state);
		// We wanted to initialize the authenticated_peers, in case
		// this connection has other messages in flight still.
		// But we don't want to use the connection ourselves, so
		// stop here.
		if( duplicate_connections.size() > 1 ) {
		    // The only possible collision should be a connection
		    // that we chose over this one.
		    Carp.die( "Too many collisions: " +
			      duplicate_connections );
		}

	    } else {
		// Connection is open or opening.  Finish setup.
		if( state.get_state() == ConnectionState.STATE_OPEN ) {
		    // Already opened.  Must be a re-authentication.
		    if( DEBUG ) Debug.printtagln(tag, "Already open: " +
						 state);
		}		
		if( state.client_socket == null ) {    // Incoming connection. 
		    if( state.get_state() == ConnectionState.STATE_WAITING ) {
			//  We've been ignoring it (STATE_WAITING).
			if( DEBUG ) Debug.printtagln(tag, "Marking as open: " +
						     state);
			// Skip straight to OPEN
			state.set_state( ConnectionState.STATE_STARTING );
			state.set_state( ConnectionState.STATE_OPEN );
			
			// Add the new state to the cache clock algorithm
			connection_cache.add( state );
			state.set_referenced();	// clock algorithm update.
			cache_clock.insert( state );
		    }
		    
		} else {	// outgoing connection
		    // We've been talking to one virtual node already.
		    // Make sure we can still talk to it.
		    if((! state.authenticated_peers.contains( state.peer_id ))
		       ||(!contact_id.equals(get_contact_id(state.peer_id)))){
			Carp.die( "Lost mapping to original peer: " +
				  state.peer_id );
		    }
		}
		
		////////////////////
		// Now close duplicate connections.
		if( ! duplicate_connections.isEmpty() ) {
		    duplicate_connections.add( state );
		    close_duplicate_connections( duplicate_connections,
						 contact_id );
		} else {
		    // No competition.
		    connection_states_by_contact_id.put( contact_id, state );
		}
	    }
	} else {
	    // message sender does not match connection port.
	    if( DEBUG || DebugFlags.LOG_BAD_NETWORK_EVENTS ) {	    
		Debug.printtagln( tag, "wrong connection id.  Expected: " +
				  state.peer_id + ". Dropping it." );
	    }
	    // Drop it.
	}

	if( DEBUG ) print_status();	// Show current status.
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }

    /**
     * Chooses one of a set of connections, supposedly all connected to the
     * same remote network stage, and closes all but one.
     * <p>
     * If the set contains both locally- and remote-initiated connections,
     * this method first uses <code>local_dominates</code> to choose which
     * kind of connection to keep.  It then chooses the connection from
     * that set whose client-side port number is lowest.
     * <p>
     * This algorithm may not always choose the "best" connection by many
     * metrics.  For example, it may choose a connection that is still
     * starting up over one that is already open and authenticated.  Right
     * now it's more important to make sure that both peers make the
     * <b>same</b> decision, and this algorithm suffices for that purpose.
     * It's also slightly biased towards older connections, which may help
     * in 
     */
    private void close_duplicate_connections( Set connections,
					      NodeId contact_id )
    {
	final String tag = class_tag+".close_duplicate_connections";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	// First pass - pick best locally- and best remote-initiated
	// connection.  Check for consistent remote addresses.
	InetAddress remote_addr = null;
	int lowest_local_port = Integer.MAX_VALUE;
	int lowest_remote_port = Integer.MAX_VALUE;
	ConnectionState best_local = null;
	ConnectionState best_remote = null;

	Iterator conn_iter = connections.iterator();
	while( conn_iter.hasNext() ) {
	    ConnectionState state = (ConnectionState)conn_iter.next();
	    
	    if( remote_addr == null ) {
		remote_addr = state.peer_id.address();
		if( DEBUG ) Debug.printtagln(tag, "Remote address: "
					     + remote_addr );
		
	    } else if( ! state.peer_id.address().equals( remote_addr )) {
		Carp.die( "Connections have different remote addresses: "
			  + remote_addr + ", " + state.peer_id.address() );
	    }
	    if( state.client_socket == null ) {	// remote
		int port = state.peer_id.port();
		if( port < lowest_remote_port ) {
		    lowest_remote_port = port;
		    best_remote = state;
		}
	    } else {	// local
		if( state.get_state() == ConnectionState.STATE_OPEN ) {
		    // We are open, so we have a port:
		    int port = state.local_connection_id.port();
		    if( port < lowest_local_port ) {
			lowest_local_port = port;
			best_local = state;
		    }
		} else {
		    // Not open yet.  Only keep it if it's the only one.
		    // Hopefully the remote peer won't choose this
		    // connection if we don't.  If it does, we'll close
		    // everything and start over.
		    if( best_local == null ) {
			best_local = state;
		    }
		}
	    }
	}
	
	// Now choose the best overall 
	ConnectionState best;
	if( (best_local != null) && (best_remote != null) ) {
	    if( remote_addr.equals(my_contact_id.address()) ) {
		if( lowest_local_port < lowest_remote_port ) {
		    best = best_local;
		} else {
		    best = best_remote;
		}
	    } else if( local_dominates( remote_addr ) ) {
		best = best_local;
	    } else {
		best = best_remote;
	    }
	} else if( best_local != null ) {
	    best = best_local;
	} else {
	    best = best_remote;
	}
	if( DEBUG ) Debug.printtagln(tag, "keeping: " + best );
	
	// Second pass - close connections and send any pending messages.
	List msgs_to_send = new LinkedList();
	HashSet authenticated_peers = new HashSet();
	conn_iter = connections.iterator();
	while( conn_iter.hasNext() ) {
	    ConnectionState state = (ConnectionState)conn_iter.next();

	    if( state != best ) {
		if( DEBUG ) Debug.printtagln(tag, "dropping: " + state);
		// Remember all authenticated_peers.
		authenticated_peers.addAll( state.authenticated_peers );
		if( state.client_socket == null ) {	// remote
		    // Must be OPEN already.
		    close_connection( state );
		} else {	// local
		    // send pending messages, set doa if necessary.
		    switch( state.get_state() ) {
		    case ConnectionState.STATE_WAITING:
			state.set_state(ConnectionState.STATE_STARTING);
			msgs_to_send.addAll(state.remove_pending());
			// Never started.  Just drop it.
			waiting_connections.remove( state );
			break;
		    case ConnectionState.STATE_STARTING:
			msgs_to_send.addAll(state.remove_pending());
			// Will close after it opens.
			doa_connections.put( state.client_socket, state );
			break;
		    case ConnectionState.STATE_OPEN:
			// Already started a connection.  Close it down.
			close_connection( state );
			break;
		    case ConnectionState.STATE_CLOSING:
		    default:
			Carp.die( "Illegal state: " + state );
		    }
		}
	    }
	}

	// Now we have only the best connection left.
	connection_states_by_contact_id.put( contact_id, best );
	best.authenticated_peers.addAll( authenticated_peers );
	
	// Now send all the pending messages.
	Iterator pending_iter = msgs_to_send.iterator();
	while( pending_iter.hasNext() ) {
	    NetworkMessage msg = (NetworkMessage)pending_iter.next();
	    if( best.get_state() == ConnectionState.STATE_OPEN ) {
		enqueue_message( best.get_connection(), msg );
	    } else {
		// This connection is not yet ready.  Just wait.
		best.add_pending( msg );
	    }
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
    }

    /**
     * A socket was closed (most probably because we closed it ourselves).
     */
    private void handle_socket_closed( ATcpServerSocketClosedEvent event ) {
	
	final String tag = class_tag+".handle_socket_closed";

	// don't remove this until we know we're not listening on the port
	// because otherwise we could try to open a new connection
	// to ourselves and confuse everyone.
	node_id_to_classifier.remove(new NodeId(event.server_socket.getPort(), 
						  my_contact_id.address()));

	if( DEBUG ) {
	    Debug.printtagln( tag, "Socket closed event: " + event );
	}

    }
    
    /**
     * Processes an incoming packet.  If this packet completes a
     * <code>NetworkMessage<code>, that message is reconstructed
     * (header only) and dispatched.
     */
    private void handle_incoming_packet( ATcpInPacket incoming )
    {
	final String tag = class_tag+".handle_incoming_packet";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	ATcpConnection conn = incoming.getConnection();
	// Check closing connections first
	ConnectionState state = (ConnectionState)
	    connection_states_by_connection.get( conn );
	if( state == null ) {
	    // Logically should be an error, but I've seen late packets
	    // arrive for connections that were already terminated by a
	    // <code>SinkClosedEvent</code>.  Probably draining a queue.
	    if( DEBUG || DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
		Carp.carp(tag + ": Received message from unknown connection "
			  + conn + ".  Dropping it.  " );
	    }
	    return;	// EARLY EXIT!!
	}
	state.set_referenced();	// clock algorithm update.

	if( DEBUG ) Debug.printtagln( tag, "adding packet " + incoming +
				      ", seq " + incoming.getSequenceNumber());
	ConnectionStream input_stream = state.input_stream;
	input_stream.add_packet( incoming );

	// Now parse any <code>NetworkMessages</code> available in the stream.
	while( input_stream.message_available() ) {
	    long msg_size = 0;
	    if( record_peer_stats ) {
		msg_size = input_stream.next_message_size();
	    }
	    InputBuffer in_buf = input_stream.get_next_message_buffer();
            NodeId wire_peer = new NodeId (in_buf);
            NodeId wire_sender = new NodeId (in_buf);
	    NetworkMessage msg = null;
	    try {
		msg = (NetworkMessage)TypeTable.new_object( in_buf );
	    } catch( Exception e ) {
		if( DEBUG || DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
		    Carp.carp(tag + ": Received message from " + state.peer_id
			      + " that caused exception " + e + 
			      " on decode.  Dropping it.  " + 
			      "(Stack trace follows.)");
		    Carp.error( e );
		}
		continue;	// Next message.
	    }
            msg.peer = wire_peer;
            msg.sender = wire_sender;

	    if( DEBUG_VERBOSE ) Debug.printtagln( tag, "decoded: " + msg );

	    // Set inbound flag before counting bytes
	    msg.inbound = true;	    
	    if( record_peer_stats ) {
		// Increment the bandwidth counters for this peer.
		record_bytes( msg, msg_size );
	    }
	    // Switch the outbound form into an inbound form.
	    NodeId dest = msg.peer;
	    msg.peer = msg.sender;
	    msg.sender = null;

	    // Now process the message.
	    // First check for messages handled directly by this stage.
	    if( (msg instanceof AuthenticationMsg) ||
		(msg instanceof CloseConnectionMsg) ) {
		if( ! msg.peer.equals( state.peer_id ) ) {
		    // Must be sent from connection peer.
		    if( DEBUG || DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
			Debug.printtagln( tag, "invalid sender: " +
					  msg.peer + " vs. " +
					  state.peer_id + ". Dropping it." );
		    } // Drop message.
		} else if( msg instanceof AuthenticationMsg ) {
		    // Check the sender and its aliases.
		    process_authentication_message( state,
						    (AuthenticationMsg)msg );
		} else {
		    // CloseConnectionMsg
		    finish_close( state.get_connection() );
		}
	    } else {
		// Common case: message addressed to (virtual) node.
		if( ! state.authenticated_peers.contains( msg.peer ) ) {
		    // Not an authenticated peer.
		    if( DEBUG || DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
			Debug.printtagln( tag, "unauthenticated sender: " +
					  msg.peer + ". Dropping it." );
		    } // Drop message.
		} else if(states_pending_authentication.containsKey(state)) {
		    // This connection is not authenticated yet..
		    if( DEBUG || DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
			Debug.printtagln( tag, "received message from " +
					  "unauthenticated connection: " +
					  state + "\tmsg: " + msg 
					  + ". Dropping it." );
		    } // Drop message.		    
		} else {
		    // Message sent from authenticate peer node.
		    // Check local address.

		    Classifier local_classifier =
			(Classifier)node_id_to_classifier.get( dest );
		    if( local_classifier == null ) {
			if( DEBUG || DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
			    Debug.printtagln(tag, "Received message for " +
					     "unknown node: " + dest +
					     ". Dropping it." );
			}
		    } else {
			// This message is for another stage.
			
			if( DEBUG_VERBOSE ) {
			    Debug.printtagln( tag, "dispatching locally: "
					      + msg );
			} else if( DEBUG ) {
			    Debug.printtagln( tag, "dispatching locally: " +
					      msg.getClass().getName() );
			}
			try {
			    local_classifier.dispatch( msg );
			} catch( SinkFullException e ) {
			    Carp.die( e );
			}
		    }	// end valid local address block
		}	// end valid remote address block
	    }	// end not-control-message block
	}	// end while loop
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Sends a message to a (possibly remote) virtual node.
     * If necessary, stores the message temporarily to emulate the delays
     * of an artificial topology.
     */
    private void handle_outbound_message( NetworkMessage to_send )
    {
	final String tag = class_tag+".handle_outbound_message";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	if( add_delays == true ) {
	    long delay = calculate_delay( to_send );
	    if( delay == 0 ) {
		send_message( to_send );
	    } else {
		if( DEBUG_VERBOSE ) {
		    Debug.printtagln( tag, "delaying " + to_send +
				      " for " + delay + " ms." );
		}
		DelayWrapper wrapper = new DelayWrapper( to_send );
		Classifier classifier = (Classifier)
		    node_id_to_classifier.get(my_contact_id);
		classifier.dispatch_later( wrapper, delay );
	    }
	} else {
	    send_message( to_send );
	}
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }

    /**
     * Stops listening on the port specified in the event.
     */
    private void handle_close_local_port( CloseLocalPortMsg event ) {

	final String tag = class_tag+".handle_close_local_port";

	// can't close the port the contact id is listening on
	if( event.port == my_contact_id.port() ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Cannot close contact id port " +
				  event.port + ", ignoring." );
	    }
	    return;
	}

	Iterator socket_iter = server_sockets.iterator();
	while( socket_iter.hasNext() ) {
	    ATcpServerSocket server_socket =
		(ATcpServerSocket)socket_iter.next();
	    // close it only if it's the right port
	    if( server_socket.getPort() == event.port ) {
		server_socket.close();
		if( DEBUG ) {
		    Debug.printtagln( tag, "Closing connection on port " +
				      event.port );
		}
		return;
	    }
	}
	if( DEBUG ) {
	    Debug.printtagln( tag, "Got a CloseLocalPortMsg for port " +
			      event.port + ", but was not listening " + 
			      "on that port; ignoring.");
	}
	
    }

    /**
     * Calculates the number of milliseconds to delay a message, based on
     * the destination, using the artificial topology.
     */
    private long calculate_delay( NetworkMessage msg )
    {
	final String tag = class_tag+".to_delay";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	if( add_delays == false ) Carp.die( "no artificial topology" );
	// Copied from MultiplexedNetwork.java:
	Integer Dst = (Integer) node_id_to_vertex.get (msg.peer);
	if (Dst == null) {
	    Carp.die ("Got message " + msg + " to " + msg.peer + 
		      ", who is not in the topology mapping.");
	}
	Integer Src = (Integer) node_id_to_vertex.get (msg.sender);
	if (Src == null) {
	    Carp.die ("We (" + msg.sender + ") are somehow " +
		      "not in the topology mapping.");
	}
	int dst = Dst.intValue ();
	int src = Src.intValue ();
	// Change by geels: do not record type_code
	//	long type_code = TypeTable.type_code (msg);
	CountBuffer cb = new CountBuffer ();
	cb.add ((int)0);
	cb.add (msg);
	int size = cb.size ();
	TransitStub.PathInfo pi =
	    topology.shortest_path (src, dst, size, false /* no record */,
				    0 /* dummy type_code */ );
	long delay_ms = Math.round (pi.route_time * 1000.0);
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
	return delay_ms;
    }
    
    /**
     * Sends a message to a (possibly remote) virtual node.  If necessary,
     * opens a new connection to the remote machine.
     */
    private void send_message( NetworkMessage to_send )
    {
	final String tag = class_tag+".send_message";
	if( DEBUG_VERBOSE ) Debug.printtagln( tag, "called: " + to_send );
	else if( DEBUG_TRACE ) Debug.printtagln( tag, "called:" );

	// First check for local delivery
	Classifier local_classifier =
	    (Classifier)node_id_to_classifier.get( to_send.peer );
	if( local_classifier != null ) {
	    // Local destination.  Clone the message (for safety) and send it.
	    NetworkMessage copy = (NetworkMessage)QSClone.qs_clone( to_send );
	    // Set inbound flag before counting bytes
	    copy.inbound = true;
	    if( record_peer_stats ) {
		// Increment the bandwidth counters for this peer.
		CountBuffer cb = new CountBuffer();
		cb.add( copy );
		record_bytes( to_send, cb.size() );	// outbound
		record_bytes( copy, cb.size() );	// inbound
	    }
	    // Switch the outbound form into an inbound form.
	    copy.peer = to_send.sender;
	    copy.sender = null;

	    if( DEBUG ) Debug.printtagln( tag, "dispatching locally: " +
					  to_send.getClass().getName() );
	    try {
		local_classifier.dispatch( copy );
	    } catch( SinkFullException e ) {
		Carp.die( e );
	    }

	} else {
	    // Not a local node: find the connection for the remote peer.
	    NodeId contact_id = get_contact_id( to_send.peer );
	    ConnectionState state = (ConnectionState)
		connection_states_by_contact_id.get( contact_id );
	    if( state == null ) {
		if( DEBUG ) {
		    Debug.printtagln( tag, "no connection" );
		    print_connections();
		}
		// Need to create a new connection.
		start_new_connection( to_send );
		
	    } else {
		if( DEBUG ) Debug.printtagln( tag, "found state: " + state );
		
		if( state.get_state() != ConnectionState.STATE_OPEN ) {
		    if( DEBUG ) Debug.printtagln( tag, "waiting" );
		    state.add_pending( to_send );
		    
		} else {		// Finall - SENDING!
		    enqueue_message( state.get_connection(), to_send );
		    state.set_referenced();	// clock algorithm update.
		}
	    }
	}
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }

    /**
     * Starts a connection. Just a helper method.
     */
    private void start_connection( ConnectionState state )
    {
	final String tag = class_tag+".start_connection";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	state.set_state( ConnectionState.STATE_STARTING );
	
	// Reserve a slot in the cache.
	free_connections--;
	// Don't actually insert into the clock algorithm until the
	// connection is open, to avoid victim selection process.
	
	// Initiate a new connection through the aSocket layer.
	// It will return an <code>ATcpConnection</code> event later.
	if( DEBUG ) Debug.printtagln( tag, "opening new socket to " +
				      state.peer_id );	    	
	state.client_socket = new ATcpClientSocket( state.peer_id.address(),
						    state.peer_id.port(),
						    my_sink );

	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Starts a waiting connection if there is room in the cache.
     */
    private void start_waiting_connection()
    {
	final String tag = class_tag+".start_waiting_connection";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	if( waiting_connections.isEmpty() ) {
	    if( DEBUG ) Debug.printtagln( tag, "nothing waiting" );

	} else if( free_connections > 0 ) {
	    // We don't have too many open connections.  Start one.
	    ConnectionState state =
		(ConnectionState)waiting_connections.remove( 0 );
	    start_connection( state );
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    
    /**
     * Initiates a new connection to the destination machine and stores the
     * message for later transmission.  If the connection cache is full, this
     * connection is put on the waiting list.
     */
    private void start_new_connection( NetworkMessage to_send )
    {
	final String tag = class_tag+".start_new_connection";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	NodeId contact_id = get_contact_id( to_send.peer );
	if( DEBUG ) Debug.printtagln( tag, "new peer: " + contact_id );
	
	// First set up a new data structure.
	ConnectionState state = new ConnectionState( contact_id );
	state.add_pending( to_send );
	connection_states_by_contact_id.put( contact_id, state );
	
	if( free_connections <= 0 ) {
	    if( DEBUG ) Debug.printtagln( tag, "waiting" );	    
	    // We're full.  Start the connection later.
	    waiting_connections.add( state );	// append
	    prune_cache();	// Maybe make some room on free list.
	    
	} else {
	    // We don't have too many open connections.  Start immediately.
	    start_connection( state );
	    prune_cache();	// May have just filled cache.
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Copes with a failed connection.  Deletes local state.
     */
    private void handle_sink_closed( SinkClosedEvent event )
    {
	final String tag = class_tag+".handle_sink_closed";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	SinkIF dead_sink = event.sink;
	if( dead_sink instanceof ATcpConnection ) {
	    // One of our remote connections has closed unexpectedly
	    ConnectionState closed = (ConnectionState)
		connection_states_by_connection.remove( dead_sink );
	    if( closed == null ) {
		if( DEBUG ) {
		    Debug.printtagln( tag, "Received " + event +
				      " for unknown connection" );
		}
	    } else {
		delete_connection_state( closed );
	    }
		
	} else {
	    Carp.die( "Received unexpected: " + event );
	}
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }

    /**
     * Follow the handshake protocol to close a connection.
     * Respond if necessary, then clear out local state and check the
     * waiting list for new connections.
     */
    private void finish_close( ATcpConnection conn )
    {
	final String tag = class_tag+".finish_close";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	ConnectionState state =
	    (ConnectionState)connection_states_by_connection.get( conn );

	if( state.get_state() == ConnectionState.STATE_CLOSING ) {
	    // Invariant: We already called <code>close_connection</code>.
	} else if( (state.get_state() == ConnectionState.STATE_OPEN) ||
		   ((state.get_state() == ConnectionState.STATE_WAITING)
		    && (state.client_socket == null)) ) {
	    // A normal, open connection, or an unauthenticated (but
	    // otherwise open) incoming connection.
	    // Reply and disable the connection.
	    close_connection( state );
	} else {
	    Carp.die( "Unexpected state: " + state );
	}

	// Our half of the handshake is now complete.  Close the connection.
	// We will finish clearing the connection state once the aSocket
	// layer sends us a <code>SinkClosedEvent</code>.
	// WARNING: SandStorm uses <code>enqueue_lossy</code> for this
	// event, so we might never see it.  Cache leak possible.
	// We should add code to remember the close and time out manually.
	try {
	    conn.close( my_sink );
	} catch( SinkClosedException sce ) {
	    if( DEBUG ) Debug.printtagln( tag, "apparently already closed: "
					  + sce );
	    // No <code>SinkClosedEvent</code> coming.  Clean up now.
	    delete_connection_state( state );
	}

	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Cleans up local state corresponding to a connection to a remote
     * peer.  The connection should already be closed.
     */
    private void delete_connection_state( ConnectionState state_to_delete )
    {
  	final String tag = class_tag+".delete_connection_state";
 	if( DEBUG_TRACE ) Debug.printtagln( tag, "called: " +
					    state_to_delete );
	// Stop waiting for close.
	Object token = closing_connections.remove( state_to_delete );
	if( token != null ) {
	    // Normal, properly closed connection.  Mostly done.
	    resender.cancel( token );
	    
	} else {
	    if( DEBUG ) {
		Debug.printtagln( tag, "looks like an unexpected close.");
	    }
	    // Must be an unexpected close, because
	    // <code>close_connection</code> was not called.  Finish the
	    // cleanup that that method would have done.

	    boolean was_active = clean_state( state_to_delete );
	    if( was_active ) {
		// Unexpected close to active connections.
		// dispatch an error message to every node
		signal_death( state_to_delete );
	    }
	}

	// Last bit of cleanup
	ATcpConnection conn = state_to_delete.get_connection();
	if( conn != null )
	    connection_states_by_connection.remove(conn);
	free_connections++;
	    
	// Start another connection if any are waiting.
	if( DEBUG ) print_status();	
	start_waiting_connection();
	
 	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Closes open connections as necessary and possible to ensure enough
     * room on the free list.
     */
    private void prune_cache()
    {
	final String tag = class_tag+".prune_cache";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );
	if( DEBUG ) print_status();
	if( DEBUG ) print_connections();

	// We need to close enough so that we can start all waiting
	// connections and still have room on the free list.
	int num_to_evict = ((waiting_connections.size() + min_free_conn) -
			    (free_connections + closing_connections.size()));
	if( num_to_evict > 0 ) {
	    if( DEBUG ) Debug.printtagln( tag, "evicting " + num_to_evict );
	    // Our free list is too small, even when we count the
	    // connections that will be closed soon.

	    if( connection_cache.size() < num_to_evict ) {
		num_to_evict = connection_cache.size();

		if( DEBUG ) {
		    Debug.printtagln( tag, "only " + num_to_evict +
				      " available");
		    print_status();
		}
	    }

	    // Choose and close connections.
	    for( int i=0; i<num_to_evict; i++ ) {
		ConnectionState state = (ConnectionState)
		    cache_clock.select_victim( true /* modified ok */ );
		// (we don't really use the "modified" bit)
		connection_cache.remove( state );
		
		close_connection( state );
	    }
	    if( DEBUG ) print_status();	    
	    
	} else {
	    // The cache is fine.
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }
    
    /**
     * Serializes a <code>NetworkMessage</code> and enqueues it onto a
     * connection. 
     */
    private void enqueue_message( ATcpConnection conn, NetworkMessage msg )
    {
	final String tag = class_tag+".enqueue_message";
	if( DEBUG_VERBOSE ) Debug.printtagln( tag, "called: " + msg );
	else if( DEBUG ) Debug.printtagln( tag, "called: " +
					   msg.getClass().getName() );
	else if( DEBUG_TRACE ) Debug.printtagln( tag, "called:" );

	int msg_size = 0;
	try {
	    // Serialize the message and record size.
	    NetworkOutputBuffer buffer = new NetworkOutputBuffer();
	    buffer.add( (int)0 );	// reserve space for size
	    msg.peer.serialize (buffer);
	    msg.sender.serialize (buffer);
	    buffer.add( msg );
	    msg_size = buffer.size();
	    buffer.overwrite( 0, msg_size-4 );	// backpatch size.
	    // Now pass the bytes to the connection.
	    buffer.enqueue( conn );

	} catch( SinkFullException e ) {
	    if( DEBUG ) Debug.printtagln( tag, "connection busy" );
	    if( DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
		Debug.printtagln( tag, "connection sink full.  Cannot " +
				  "enqueue " + msg + ".  Dropping it." );
	    }
	} catch( SinkClosedException e ) {
	    if( DEBUG ) Debug.printtagln( tag, "connection closed; " +
					  "packet loss possible" );
	    if( DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
		Debug.printtagln( tag, "connection closed.  Cannot " +
				  "enqueue " + msg + ".  Dropping it." );
	    }
	    // Remove the connection.
	    ConnectionState closed = (ConnectionState)
		connection_states_by_connection.remove( conn );
	    if( closed != null ) {
		delete_connection_state( closed );
	    } else if( DEBUG ) {
		Debug.printtagln( tag, "Connection already deleted" );
	    }

	} catch( Exception e ) {
	    Carp.carp( tag + ": could not send message " + msg +
		       " on connection " + conn );
	    Carp.die( e );
	}
	
	if( record_peer_stats ) {
	    // Increment the bandwidth counters for this peer.
	    record_bytes( msg, msg_size );
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }

    /**
     * Returns <code>true</code> iff the local machine has "priority" over
     * the specified remote machine when each initiates connections to the
     * other simultaneously.
     * <p>
     * The decision is a deterministic but somewhat-random-looking one
     * based on both the relative magnitude of the two addresses and the
     * parity of their XOR.  We add the parity bit to reduce the bias.
     */
    private boolean local_dominates( InetAddress remote_peer )
    {
	final String tag = class_tag+".local_dominates";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	boolean choose_local;
	
	// Technically, use the hashcode of the address rather than the
	// address itself.  In the current implementation, there is no
	// difference.
	int local = my_contact_id.address().hashCode();
	int remote = remote_peer.hashCode();

	if( DEBUG ) Debug.printtagln( tag, "local: " + local +
				      " remote: " + remote );
	if( local == remote ) {
	    Carp.die( "same machine: " + local );
	}

	// First find the parity of the XOR.
	int xor = (local ^ remote);
	int parity = 0;
	for( int i=0; i<32; i++ ) {
	    parity += ( (xor >> i) & 0x1 );
	}
	parity &= 0x1;
	
	if( parity == 0 ) {
	    // Choose the lesser address.
	    if( local < remote ) {
		choose_local = true;
	    } else {
		choose_local = false;
	    }
	} else {
	    // Choose the greater address.
	    if( local < remote ) {
		choose_local = false;
	    } else {
		choose_local = true;
	    }
	}

	if( DEBUG ) Debug.printtagln( tag, "chose: " +
				      (choose_local?"local":"remote") );
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
	return choose_local;
    }

    /**
     * Prepares a connection for shutdown by clearing most related state.
     * @return <code>true</code> iff this connection was active.
     */
    private boolean clean_state( ConnectionState state )
    {
	final String tag = class_tag+".clean_state";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );
	
	// First, stop waiting for authentication.
	Object token = states_pending_authentication.remove(state);
	if( token != null )
	    resender.cancel( token );
	// It was possibly a DOA that never arrived:
	if( state.client_socket != null )
	    doa_connections.remove( state.client_socket );

	// Remove from clock algorithm if necessary.
	boolean in_clock = connection_cache.remove( state );
	if( in_clock ) cache_clock.remove( state );
	
	// Still an active connection?  See if we would use it to send a
	// message to one of the nodes on the other end.
	NodeId contact_id;
	if( state.authenticated_peers.isEmpty() ) {
	    // Unauthenticated connection.  Check the connection peer.
	    contact_id = get_contact_id( state.peer_id );
	} else {
	    // Check any of the authenticated peers:
	    Iterator peer_iter = state.authenticated_peers.iterator();
	    contact_id = get_contact_id( (NodeId)peer_iter.next() );
	}
	ConnectionState state_in_use = (ConnectionState)
	    connection_states_by_contact_id.get( contact_id );
	// Use instance equality check:
	boolean is_active = (state_in_use == state);

	if( is_active ) {
	    // We were still using this connection.
	    // That means a few more data structures need cleaning.
	    connection_states_by_contact_id.remove( contact_id );

	    // Print out any pending messages that we're dropping.
	    if( state.get_state() == ConnectionState.STATE_STARTING ) {
		List to_send = state.remove_pending();
		if( to_send != null ) {
		    if( DEBUG || DebugFlags.LOG_BAD_NETWORK_EVENTS ) {
			StringBuffer msg_sbuf =
			    new StringBuffer( "Dropping it: (  " );
			Iterator pending_iter = to_send.iterator();
			while( pending_iter.hasNext() ) {
			    msg_sbuf.append( pending_iter.next() );
			    msg_sbuf.append(", ");
			}
			msg_sbuf.setCharAt( msg_sbuf.length()-2, ')' );
			Carp.carp( tag + ": " + msg_sbuf.toString() );
		    }
		}
	    }
	}

	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
	return is_active;
    }
	
    /**
     * Starts the process of closing a connection.  This method removes the
     * connection from use, marks the state as "closed", and possibly sends
     * notification to the remote peer.<p>
     * The <code>ATcpConnection.close</code> method will not be called
     * until the remote peer responds.  This delay prevents the loss of
     * incoming packets.  <p>
     * It is possible to close an outgoing connection before it is open
     * (because of a collision with an incoming connection).  In this case
     * this method must be called again once the connection opens.
     */
    private void close_connection( ConnectionState state )
    {
	final String tag = class_tag+".close_connection";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );
	if( DEBUG ) {
	    print_status();	
	    Debug.printtagln( tag, "closing: " + state );
	}

	if( (state.get_state() == ConnectionState.STATE_STARTING)
	    && (state.client_socket != null) ) {
	    // Outgoing, DOA connection that just opened.
	    state.set_state( ConnectionState.STATE_OPEN );	    
	} else if( (state.get_state() == ConnectionState.STATE_WAITING)
		   && (state.client_socket == null) ) {
	    // Incoming, unauthenticated connection.
	    // Treat it as a fully-open connection during shutdown.
	    state.set_state( ConnectionState.STATE_STARTING );
	    state.set_state( ConnectionState.STATE_OPEN );
	}
	
	// At this point the only valid state is STATE_OPEN.
	if( state.get_state() != ConnectionState.STATE_OPEN ) {
	    Carp.die( "Illegal state: " + state );
	}

	// Clear out just about all the related state.
	clean_state( state );

	// Send a <code>CloseConnectionMsg</code> to the remote peer.
	// It will respond to complete the handshake protocol.
	if( state.get_connection() == null )
	    Carp.die( "no connection: " + state );
	CloseConnectionMsg stop_please =
	    new CloseConnectionMsg( state.local_connection_id,
				    state.peer_id );
	enqueue_message( state.get_connection(), stop_please );
	// Set a timeout that we will clear iff we receive an
	// <code>CloseConnectionMsg</code> from the remote peer
	// and shut down cleanly.
	Object token = resender.remember( state, PEER_TIMEOUT) ;
	closing_connections.put( state, token );
	    state.set_state( ConnectionState.STATE_CLOSING );
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );		
    }

    /**
     * Records the bandwidth consumed by a message.
     * A separate map is kept for each local virtual node.
     * In each map a <code>PeerStats</code> object is maintained for each 
     * peer,classname pair.
     */
    private void record_bytes( NetworkMessage msg, long size )
    {
	final String tag = class_tag+".record_bytes";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	NodeId our_id;
	NodeId peer;
	if( msg.inbound ) {
	    our_id = msg.peer;
	    peer = msg.sender;
	} else {
	    our_id = msg.sender;
	    peer = msg.peer;
	}
	String class_name = msg.getClass().getName();
	Pair key = new Pair( msg.peer, class_name );

	// First get the map for this virtual node.
	HashMap per_node_map = (HashMap)peer_stats.get( our_id );
	if( per_node_map == null ) {
	    per_node_map = new HashMap();
	    peer_stats.put( our_id, per_node_map );
	}
	
	PeerStats stats = (PeerStats)per_node_map.get( key );
	if( stats == null ) {
	    stats = new PeerStats();
	    per_node_map.put( key, stats );
	}
	if( msg.inbound ) {
	    stats.msgs_received++;
	    stats.bytes_received += size;
	} else {
	    stats.msgs_sent++;
	    stats.bytes_sent += size;
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
    }
    
    /**
     * Prints out the current peer statistics and clear the counters.
     * Does <b>not</b> send a reply.
     */
    private void handle_clear_bw_req( ClearBandwidthCountsReq req )
    {
	final String tag = class_tag+".handle_clear_bw_req";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	if( ! record_peer_stats ) {
	    if( DEBUG ) Debug.printtagln( tag, "No stats recorded" );
	} else {

	    Iterator id_iter = peer_stats.keySet().iterator();
	    while( id_iter.hasNext() ) {
		NodeId our_id = (NodeId)id_iter.next();
		Carp.out( tag + " bandwidth counts for " + our_id + ":" );
		HashMap per_node_map = (HashMap)peer_stats.get( our_id );
		Iterator key_iter = per_node_map.keySet().iterator();
		while( key_iter.hasNext() ) {
		    Pair key = (Pair)key_iter.next();
		    NodeId peer = (NodeId)key.first();
		    String class_name = (String)key.second();
		    PeerStats stats = (PeerStats)per_node_map.get( key );
		    
		    Carp.out( req.user_data + "\t" + peer + "\t" + class_name +
			      "\t" + stats.msgs_received + "\t" +
			      stats.msgs_sent + "\t" + stats.bytes_received +
			      "\t" + stats.bytes_sent );
		}
	    }
	    
	    // Now clear the statistics.
	    peer_stats = new HashMap();
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
    }
    
    /* Specified by seda.sandStorm.api.EventHandlerIF */
    public void destroy()
    {
	final String tag = class_tag+".destroy";
	if( DEBUG_TRACE ) Debug.printtagln( tag, "called" );

	Iterator socket_iter = server_sockets.iterator();
	while( socket_iter.hasNext() ) {
	    ATcpServerSocket server_socket =
		(ATcpServerSocket)socket_iter.next();
	    server_socket.close();
	}
	
	if( DEBUG_TRACE ) Debug.printtagln( tag, "done" );
    }
    
    ////////////////////     INNER CLASSES	////////////////////

    /**
     * A data structure to store the state of a connection to a remote
     * machine.  The connection can be in one of four states, defined by
     * the <code>STATE_*</code> constants.
     * <p>
     * This class extends {@link CachePolicy.Line} so that we can easily
     * set the reference bits used by the connection cache replacement
     * algorithm.  We do not use the "modified" bits (they remain
     * permanently cleared).
     * <p>
     * It implements <code>QueueElementIF</code> only so that we can store
     * it using {@link Resender#remember}.
     */
    private class ConnectionState extends CachePolicy.Line
					  implements QueueElementIF
    {
	////////////////////     CONSTANTS	////////////////////

	// The various valid states for a <code>ConnectionState</code>.
	public static final int STATE_WAITING = 0;
	public static final int STATE_STARTING = 1;
	public static final int STATE_OPEN = 2;
	public static final int STATE_CLOSING = 3;
	
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////

	/**
	 * One of the <code>STATE_*</code> values above.
	 */
	private int state;

	/**
	 * The address and port of the remote peer.  May differ from that
	 * peer's contact id if it initiated the connection or hosts
	 * multiple virtual nodes.
	 */
	public NodeId peer_id;

	/**
	 * The <code>NodeId</code> of all (possibly virtual) remote nodes
	 * that have authenticated themselves on this connection.
	 */
	public Set authenticated_peers;

	/**
	 * The underlying connection
	 */
	private ATcpConnection conn;

	/**
	 * The local address and port used by the connection.
	 */
	public NodeId local_connection_id;
	
	/**
	 * The local client socket for locally-initiated connections, or
	 * <code>null</code> for remote-initiated connections.
	 */
	public ATcpClientSocket client_socket;

	/**
	 * A reordering layer on top of the <code>ATcpConnection</code>.
	 */
	public ConnectionStream input_stream;

	/**
	 * Messages waiting to be sent (because
	 * <code>state=STATE_WAITING</code> or
	 * <code>state=STATE_STARTING</code>).
	 * or <code>null</code>.
	 */
	private List pending_outbound_messages;
	
	////////////////////    CONSTRUCTORS	////////////////////

	/**
	 * Create a new state object.
	 */
	public ConnectionState( NodeId peer_id  )
	{
	    super( false /* "modified" means nothing here */);
	    this.peer_id = peer_id;
	    this.authenticated_peers = new HashSet();
	    this.state = STATE_WAITING;
	    this.input_stream = new ConnectionStream();
	}
	
	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////

	/**
	 * Returns the state of the connection.
	 */
	public int get_state()
	{
	    return this.state;
	}

	/**
	 * Changes the state of the connection.
	 * Currently only single transitions forward are allowed.
	 */
	public void set_state( int new_state )
	{
	    if( this.state != (new_state - 1) ) {
		Carp.die( "Invalid ConnectionState transition: " +
			  this.state + " to " + new_state );
	    }
	    this.state = new_state;
	}

	/**
	 * Removes and returns the internal list storing all messages
	 * waiting to be sent on this connection.
	 */
	public List remove_pending()
	{
	    if( state != STATE_STARTING ) {
		Carp.die( "Cannot access pending messages in current "
			  + " state: " + toString() );
	    }
	    List to_return = pending_outbound_messages;
	    pending_outbound_messages = null;
	    return to_return;
	}

	/**
	 * Appends a new message to the list waiting to be sent.
	 */
	public void add_pending( NetworkMessage msg )
	{
	    if( (state != STATE_WAITING) && (state != STATE_STARTING) ) {
		Carp.die( "Cannot add pending messages in current "
			  + " state: " + toString() );
	    }
	    if( pending_outbound_messages == null )
		pending_outbound_messages = new LinkedList();		
	    pending_outbound_messages.add( msg );
	}

	/**
	 * Associates a socket with this state.
	 */
	public void set_connection( ATcpConnection conn )
	{
	    this.conn = conn;
	    Socket local_socket = conn.getSocket();
	    this.local_connection_id =
		new NodeId(local_socket.getLocalPort(),
			   local_socket.getLocalAddress());
	}
	
	/**
	 * Returns the underlying connection.
	 */
	public ATcpConnection get_connection()
	{
	    return this.conn;
	}
	
	/**
	 * Returns a human-readable representation of this state.
	 */
	public String toString()
	{
	    String state_name;
	    switch( state ) {
	    case STATE_WAITING: state_name = "WAITING"; break;
	    case STATE_STARTING: state_name = "STARTING"; break;
	    case STATE_OPEN: state_name = "OPEN"; break;
	    case STATE_CLOSING: state_name = "CLOSING"; break;
	    default: state_name = "UNKNOWN";
	    }
	    StringBuffer msg_sbuf;
	    if( DEBUG_VERBOSE ) {
		msg_sbuf = new StringBuffer("pending: (  ");
		if( pending_outbound_messages != null ) {
		    Iterator pending_iter = pending_outbound_messages.iterator();
		    while( pending_iter.hasNext() ) {
			msg_sbuf.append( pending_iter.next() );
			msg_sbuf.append(", ");
		    }
		}
		// overwrite last comma.	    
		msg_sbuf.setCharAt( msg_sbuf.length()-2, ')' );
	    } else {
		int num_pending = 0;
		if( pending_outbound_messages != null ) {
		    num_pending = pending_outbound_messages.size();
		}
		msg_sbuf = new StringBuffer("(" + num_pending + " pending)");
	    }
	    
	    return "(ConnectionState for " + peer_id + " connection: " + conn +
		" (state " + state_name + ") " + msg_sbuf + ")";
	}
    }

    /**
     * We add artificial delay to outbound messages by wrapping them in
     * <code>DelayWrapper</code>s and sending the wrapped messages to
     * ourselves with some delay.  The wrapper serves only to differentiate
     * between these resent messages and those sent from other stages.
     */
    private class DelayWrapper implements QueueElementIF
    {
	////////////////////     VARIABLES	////////////////////

	/**
	 * The message to send.
	 */
	public NetworkMessage to_send;
	
	////////////////////    CONSTRUCTORS	////////////////////

	/**
	 * Create a new wrapper.
	 */
	public DelayWrapper( NetworkMessage to_send )
	{
	    this.to_send = to_send;
	}
	
	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////

	/**
	 * Returns a human-readable representation of this wrapper.
	 */
	public String toString()
	{
	    return "(DelayWrapper " + to_send + ")";
	}
    }

    /**
     * Simple little data structure to count messages sent or received from
     * a remote peer.
     */
    private class PeerStats
    {
	////////////////////     VARIABLES	////////////////////

	/** The number of messages sent to this peer. */
	public long msgs_sent;
	
	/** The number of bytes sent to this peer. */
	public long bytes_sent;
	
	/** The number of messages received from this peer. */
	public long msgs_received;
	
	/** The number of bytes received from this peer. */
	public long bytes_received;
	
    }
}
