/**
 * InjectFaultMsg.java
 *
 * @author Jeremy Stribling
 * @version    $Id: InjectFaultMsg.java,v 1.1 2003/02/08 00:49:37 strib Exp $
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;
import ostore.network.NetworkMessage;
import ostore.util.NodeId;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.QSException;

/**
 * Injects faults into Patchwork, either of a specific link or an overall node 
 * failure.
 *
 * @author Jeremy Stribling
 * @version    $Id: InjectFaultMsg.java,v 1.1 2003/02/08 00:49:37 strib Exp $
 */
public class InjectFaultMsg extends NetworkMessage 
    implements QuickSerializable {

    public boolean isPercent;
    public boolean allNodes;
    public int value;

    public NodeId nodeToDrop;

    /**
     * Constucts a fault injection message, which will let a peer
     * know that it should begin dropping messages.  Either an
     * absolute number of messages can be specified, or a percentage 
     * of messages for an indefinite amount of time.  Note that if
     * a percentage is given, it can be removed by sending another
     * fault injection message with the value at 0.  Any arriving
     * fault message overrides the previous one.
     *
     * @param peer  the node to contact
     * @param value the amount of messages to drop, either as a 
     *              number of messages or a percent of messages
     * @param isPercent  true if <code>value</code> specifies a percentage
     */
    public InjectFaultMsg (NodeId peer, int value, boolean isPercent) {
	this( peer, value, isPercent, false );
    }

    /**
     * Constucts a fault injection message, which will let a peer
     * know that it should begin dropping messages.  Either an
     * absolute number of messages can be specified, or a percentage 
     * of messages for an indefinite amount of time.  Note that if
     * a percentage is given, it can be removed by sending another
     * fault injection message with the value at 0.  Any arriving
     * fault message overrides the previous one.
     *
     * @param peer  the node to contact
     * @param value the amount of messages to drop, either as a 
     *              number of messages or a percent of messages
     * @param isPercent  true if <code>value</code> specifies a percentage
     * @param boolean   true iff this fault is to apply to all nodes probing it
     */
    public InjectFaultMsg (NodeId peer, int value, boolean isPercent, 
			      boolean allNodes) {

	super (peer, false /* outbound */);

	this.value = value;
	this.isPercent = isPercent;
	this.allNodes = allNodes;

    }

    /**
     * Constucts a fault injection message, which will let a peer
     * know that it should begin dropping messages.  Either an
     * absolute number of messages can be specified, or a percentage 
     * of messages for an indefinite amount of time.  Note that if
     * a percentage is given, it can be removed by sending another
     * fault injection message with the value at 0.  Any arriving
     * fault message overrides the previous one.
     *
     * @param peer  the node to contact
     * @param value the amount of messages to drop, either as a 
     *              number of messages or a percent of messages
     * @param isPercent  true if <code>value</code> specifies a percentage
     * @param nodeToDrop Drop messages from a specific node, instead of sender 
     */
    public InjectFaultMsg (NodeId peer, int value, boolean isPercent, 
			      NodeId nodeToDrop) {

	super (peer, false /* outbound */);

	this.value = value;
	this.isPercent = isPercent;
	allNodes = false;
	this.nodeToDrop = nodeToDrop;

    }

    public InjectFaultMsg (InputBuffer buffer) throws QSException {
	super( buffer );
	isPercent = buffer.nextBoolean();
	value = buffer.nextInt();
	allNodes = buffer.nextBoolean();
	if( buffer.nextBoolean() ) {
	    nodeToDrop = (NodeId) buffer.nextObject();
	}
    }

    public void serialize (OutputBuffer buffer) {
	super.serialize (buffer);
	buffer.add( isPercent );
	buffer.add( value );
	buffer.add( allNodes );
	if( nodeToDrop == null ) {
	    buffer.add( false );
	} else {
	    buffer.add( true );
	    buffer.add( nodeToDrop );
	}
    }

    public Object clone () throws CloneNotSupportedException {
	InjectFaultMsg newone = (InjectFaultMsg) super.clone ();
	newone.isPercent = isPercent;
	newone.value = value;
	newone.allNodes = allNodes;
	newone.nodeToDrop = nodeToDrop;
	return newone;
    }

    public String toString () {
	return "(InjectFaultMsg, value: " + 
	    value + ", isPercent: " + isPercent + ", allNodes:"
	    + allNodes + ", nodeToDrop: " + nodeToDrop + 
	    " " + super.toString () + ")";
    }
}

