/**
 * Period.java
 *
 * @author Jeremy Stribling
 * @version    $Id: Period.java,v 1.5 2003/03/06 01:10:31 strib Exp $
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;
import java.util.HashMap;
import java.util.Iterator;
import ostore.util.NodeId;
import ostore.util.Pair;

/**
 * Representing a periodic function, used for corresponding probing levels
 * and nodes to probing periods. 
 *
 * @author Jeremy Stribling
 * @version    $Id: Period.java,v 1.5 2003/03/06 01:10:31 strib Exp $
 */
public class Period {

    private static final int OP_ADD = 0;
    private static final int OP_MULT = 1;
    private static final int OP_SUB = 2;    
    private static final int OP_DIV = 3;    
    private static final int OP_EXP = 4;

    // max timeout is two minutes
    private static final long MAX_TIMEOUT = 30000;
    // min timeout is 100 ms
    private static final long MIN_TIMEOUT = 100;

    private int _base_period;
    private int _op;
    private int _factor;
    private long _timeout;

    private long _cached_period;
    private int _cached_level;

    private NodeId _self_node_id;

    // hashtable of nodes to levels
    private HashMap _node_levels;

    // hashmap of each nodes hashmaps of nodes->timeouts
    private static HashMap _node_timeouts;

    static {
	_node_timeouts = new HashMap();
    }

    private static HashMap getTimeoutHash( NodeId node_id ) {
	HashMap hash = null;
	synchronized( _node_timeouts ) {
	    hash = (HashMap) _node_timeouts.get(node_id);
	    if( hash == null ) {
		hash = new HashMap();
		_node_timeouts.put( node_id, hash );
	    }	    
	}
	return hash;
    }

    /**
     */
    public Period( NodeId self_node_id, int basePeriod, String op, 
		   int factor, long timeout ) {

	_base_period = basePeriod;
	_op = stringToOp( op );
	_factor = factor;
	_timeout = timeout;
	_cached_level = -1;
	_self_node_id = self_node_id;

	_node_levels = new HashMap();
    }

    public int getBasePeriod() {
	return _base_period;
    }

    public int getOp() {
	return _op;
    }

    public int getFactor() {
	return _factor;
    }

    public void setPeriod( int basePeriod, String op, int factor ) {
	setPeriod( basePeriod, stringToOp( op ), factor );
    }

    public void setPeriod( int basePeriod, int op, int factor ) {

	_base_period = basePeriod;
	_op = op;
	_factor = factor;
	resetLevels();

    }

    private void resetLevels() {
	_cached_level = -1;
	_cached_period = -1;

	// go through each node and redo it
	Iterator i = _node_levels.keySet().iterator();
	while( i.hasNext() ) {

	    NodeId node = (NodeId) i.next();
	    Pair p = (Pair) _node_levels.get( node );
	    p.second = new Long( getPeriod( ((Integer) p.first).intValue() ) );

	}
    }

    public synchronized boolean changePeriod( NodeId node, int level ) {

	// see if it's there already
	Pair p = (Pair) _node_levels.get( node );

	// if it's there, and the same level, no action needed
	if( p != null && ((Integer) p.first).intValue() == level ) {
	    //do nothing
	    return false;
	} else {
	    p = new Pair( new Integer( level ), 
			  new Long( getPeriod( level ) ) );
	    _node_levels.put( node, p );
	    return true;
	}

    }

    public synchronized void removeNode( NodeId node ) {
	_node_levels.remove( node );
	_node_timeouts.remove( node );
    }

    public synchronized long getPeriod( NodeId node ) {

	Pair p = (Pair) _node_levels.get( node );

	if( p == null ) {
	    return -1;
	} else {
	    long per = ((Long) p.second).longValue();
	    if( per <= 0 ) {
		return -1;
	    } else {
		return per;
	    }
	}

    }

    public synchronized long getPeriod( int level ) {

	// no negative levels
	if( level < 0 ) {
	    return -1;
	}

	// no non-positive periods
	if( _base_period <= 0 ) {
	    return -1;
	}

	// if we've cached it, just return it
	if( _cached_level == level ) {
	    return _cached_period;
	}

	// otherwise, let's compute . . .
	long returnTime = _base_period;
	
	switch( _op ) {
	case OP_ADD:
	    for( int i = 0; i < level; i++ ) {
		returnTime += _factor;
	    }    
	    break;
	case OP_MULT:
	    for( int i = 0; i < level; i++ ) {
		returnTime *= _factor;
	    }    
	    break;
	case OP_SUB:
	    for( int i = 0; i < level; i++ ) {
		returnTime -= _factor;
	    }    
	    break;
	case OP_DIV:
	    for( int i = 0; i < level; i++ ) {
		returnTime /= _factor;
	    }    
	    break;
	case OP_EXP:
	    for( int i = 0; i < level; i++ ) {
		returnTime = (long) Math.pow(returnTime, _factor);
	    }
	}
	
	if( returnTime <= 0 ) {
	    returnTime = -1;
	}

	_cached_level = level;
	_cached_period = returnTime;

	return returnTime;

    }

    /**
     * Get what the timeout value should be for a node.
     * NOTE: Each instance of a Period shares its table of timeouts
     *       with all other Periods used by the same NodeId.  That
     *       means, for example, a LossPatch and LatencyPatch within 
     *       the same Patchwork look at the same timeout values.
     */
    public long getTimeout( NodeId node ) {
	Long to = (Long) getTimeoutHash(_self_node_id).get( node );
	if( to == null ) {
	    return _timeout;
	} else {
	    return to.longValue();
	}
    }

    /**
     * Set what the timeout value should be for a node.
     * NOTE: Each instance of a Period shares its table of timeouts
     *       with all other Periods used by the same NodeId.  That
     *       means, for example, a LossPatch and LatencyPatch within 
     *       the same Patchwork look at the same timeout values.
     *       THIS MEANS IF YOU SET IT, ALL OTHER PATCHES ON YOUR NODE
     *       WILL SEE IT.
     */
    public void setTimeout( NodeId node, long value ) {
	if( value > MAX_TIMEOUT ) {
	    value = MAX_TIMEOUT;
	} else if( value < MIN_TIMEOUT ) {
	    value = MIN_TIMEOUT;
	}
	getTimeoutHash(_self_node_id).put( node, new Long(value) );
    }

    private static int stringToOp( String opString ) {

	if( opString == null ) {
	    throw new RuntimeException( "null passed to stringToOp");
	} else if( opString.equals( "add" ) || opString.equals( "+" ) ) {
	    return OP_ADD;
	} else if( opString.equals( "mult" ) || opString.equals( "*" ) ) {
	    return OP_MULT;
	} else if( opString.equals( "sub" ) || opString.equals( "-" ) ) {
	    return OP_SUB;
	} else if( opString.equals( "div" ) || opString.equals( "/" ) ) {
	    return OP_DIV;
	} else if( opString.equals( "exp" ) || opString.equals( "^" ) ) {
	    return OP_EXP;
	} else {
	  throw new RuntimeException( "An \"op\" tag cannot be " + 
				      opString );
	}
    }

}

