/**
 * QuickStatReq.java
 *
 * @author Jeremy Stribling
 * @version    $Id: QuickStatReq.java,v 1.6 2003/05/08 01:55:05 strib Exp $
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;
import ostore.util.NodeId;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;

/**
 * A <code>QuickStatReq</code> event can be sent by any stage to the Patchwork
 * stage.  It triggers the measurement of a set of nodes, but only for a 
 * small amount of time.  After that amount of time has passed, a 
 * <code>QuickStatResp</code> is sent back with the results of the 
 * measurements.
 *
 * @author Jeremy Stribling
 * @version    $Id: QuickStatReq.java,v 1.6 2003/05/08 01:55:05 strib Exp $
 */
public final class QuickStatReq implements QueueElementIF, PatchMsg {

    public byte type;
    public NodeId[] peers;
    public Object userData;
    public SinkIF sink;
    // only collect one statistic for every node
    // (the timeout is still used just in case)
    public boolean just_once;
    // in the just_once case, counts how many of 
    // the probes have finished
    protected int how_many_done = 0;
    protected boolean which_done[];

    /**
     * When the response will be generated (in ms), default 1 sec
     */
    public long time_to_resp = 1000;
   
    /**
     * Construct a new <code>QuickStatReq</code>.
     *
     * @param peer  The node to measure quickly
     * @param type  What type of measurement to make (see <code>Patch</code>
     *                for valid types.
     */
    public QuickStatReq (NodeId peer, byte type, Object userData, 
			 SinkIF sink ) {
	NodeId[] tmp = {peer};
	peers = tmp;
	this.type = type;
	this.userData = userData;
	this.sink = sink;
	this.which_done = new boolean[peers.length];
    }

    /**
     * Construct a new <code>QuickStatReq</code>.
     *
     * @param peers  The nodes to measure quickly
     * @param type  What type of measurement to make (see <code>Patch</code>
     *                for valid types.
     */
    public QuickStatReq (NodeId[] peers, byte type,  Object userData, 
			 SinkIF sink ) {
	this.peers = peers;
	this.type = type;
	this.userData = userData;
	this.sink = sink;
	this.which_done = new boolean[peers.length];
    }

    /**
     * Construct a new <code>QuickStatReq</code>.
     *
     * @param peer  The node to measure quickly
     * @param type  What type of measurement to make (see <code>Patch</code>
     *                for valid types.
     * @param timeToResp   How to long to perform the measurement for (in ms)
     */
    public QuickStatReq (NodeId peer, byte type, long timeToResp,
			 Object userData, SinkIF sink ) {
	this( peer, type, userData, sink );
	time_to_resp = timeToResp;
    }

    /**
     * Construct a new <code>QuickStatReq</code>.
     *
     * @param peers  The nodes to measure quickly
     * @param type  What type of measurement to make (see <code>Patch</code>
     *                for valid types.
     * @param timeToResp   How to long to perform the measurement for
     */
    public QuickStatReq (NodeId[] peers, byte type, long timeToResp,
			 Object userData, SinkIF sink ) {
	this( peers, type, userData, sink );
	time_to_resp = timeToResp;
    }

    public Object clone () throws CloneNotSupportedException {
	QuickStatReq newone = (QuickStatReq) super.clone ();
	newone.type = type;
	newone.peers = peers;
	newone.userData = userData;
	newone.sink = sink;
	return newone;
    }

    public byte getType() {
	return type;
    }

    public String toString () {
	String s =  "(QuickStatReq, type " + type + ", peers {";
	for( int i = 0; i < peers.length; i++ ) {
	    s += peers[i] + " ";
	}
	s += "})";
	return s;
    }
}

