/**
 * HMMBuilder.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import java.io.IOException;
import ostore.dispatch.Classifier;
import ostore.dispatch.Signal;
import ostore.util.QuickSerializable;
import ostore.util.NodeId;
import ostore.util.QSIO;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.InvalidConfigFileException;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkIF;

/**
 * A stage which automatically fits a Hidden Markov Model to a monitored
 * event stream.
 * This class extends <code>HMMStage</code>, adding support for
 * <code>Builder</code>-specific config tags and ipmlementing the
 * <code>handleEvent</code> method.
 *
 * @see HMM
 *
 * @author	Dennis Geels
 * @version	$Id: HMMBuilder.java,v 1.9 2002/07/22 20:54:53 srhea Exp $
 */

public class HMMBuilder extends HMMStage
{
    ////////////////////       CONSTANTS	////////////////////

    private static final boolean DEBUG = DebugFlags.OIL_HMMBUILDER;

    // A few strings used in the config files:
    private static final String RBLD_PERIOD_STR	= "period";
    private static final String SNAPSHOT_STR	= "write_snapshot";
    private static final String DELTA_STR	= "emit_deltas";
    
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The number of events to observe before recalculating model parameters.
     * If 0, recalculate only when a <code>RebuildSignal</code> is received.
     */
    protected int recalculation_period;
    
    /**
     * <code>true</code> iff this stage should write out snapshots of its
     * <code>Model</code> every time it recalculates its parameters. 
     */
    protected boolean write_snapshots;
    
    /**
     * The name of the snapshot file to write, if <code>write_snapshots ==
     * true</code>.  Otherwise this parameter is ignored.
     */
    protected String snapshot_name;
    
    /**
     * <code>true</code> iff this <code>Builder</code> should publish
     * <code>Delta</code>s for its <code>Model</code> whenever the model
     * parameters are recalculated.
     */
    protected boolean emit_deltas;
    
    /**
     * The total number of observations so far.
     */
    protected int num_observed;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	class_tag = "HMMBuilder";
    }

    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////


    /* Specified by seda.sandStorm.api.EventHandlerIF */
    /**
     * Processes the monitored event stream and handles
     * <code>RebuildSignal</code>s.
     *
     * @param elem	Either an event to monitor or a
     * 			<code>Signal.RebuildSignal</code>.
     *
     * @exception UnknownEventException if <code>elem<code> is neither of
     * the expected event types.
     */
    public void handleEvent(QueueElementIF elem) throws UnknownEventException
    {
	final String tag = class_tag+".handleEvent";
	if( DEBUG ) Debug.printtagln( tag, name + " called:\t" + elem );

	if( observed_type.isInstance( elem ) ) {

	    // Extract the desired field from the Event.
	    QuickSerializable observation = null;
	    try {
		observation = (QuickSerializable)accessor.get(elem);
	    } catch( IllegalAccessException iae ) {
		// Should never happen.
		Carp.die( iae );
	    }

	    // Add the observation to the list.
	    model.record( observation );
	    num_observed++;

	    // Now recalculate the model parameters based on the list of
	    // recent observations.
	    if( (recalculation_period > 0) &&
		((num_observed % recalculation_period) == 0) ) {

		recalculate();
	    }
	    
	} else if( elem instanceof Signal.RebuildSignal ) {
	    recalculate();

	} else {
	    throw new
		UnknownEventException( "Neither a RebuildSignal nor a "
				       + observed_type.getName(), elem );
	}
	if( DEBUG ) Debug.printtagln( tag, name + " done" );
    }

    /* Specified by seda.sandStorm.api.EventHandlerIF */
    /**
     * Initializes this <code>HMMBuilder</code> using the configuration
     * parameters.
     *
     * @exception InvalidConfigFileException if the
     * <code>ConfigDataIF</code> contains an invalid parameter or does not
     * contain a required field.
     * @exception InvalidTypeException if a parameter specifies a
     * <code>Class</code> or <code>Field</code> which does not implement
     * the proper java interface.
     *
     * @see HMMStage#init
     */
    public void init(ConfigDataIF config) throws Exception
    {
	final String tag = class_tag+".init";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	// Initialize the shared HMMStage fields.
	super.init( config );

	// Listen for <code>RebuildSignal</code>s
	Signal signal = Signal.getSignal(
		new NodeId (config.getString ("node_id")));
	SinkIF my_sink = config.getStage().getSink();
	signal.register( Signal.RBLD, name, my_sink );
	
	// Initialize generic Builder parameters

	recalculation_period = 0;	// Do not recalculate automatically
	if( config.contains( RBLD_PERIOD_STR ) ) {
	    recalculation_period = config.getInt( RBLD_PERIOD_STR );
	    if( recalculation_period < 0 )
		throw new InvalidConfigFileException(name, RBLD_PERIOD_STR);
	}
	if( DEBUG ) {
	    if( recalculation_period == 0 ) {
		Debug.printtagln( tag, "disabling period recalculation" );
	    } else {
		Debug.printtagln( tag, "will recalculate every " +
				  recalculation_period + " events." );
	    }
	}

	write_snapshots = true;
	snapshot_name = config.getString( SNAPSHOT_STR );
	if( snapshot_name == null ) {
	    write_snapshots = false;
	    if( DEBUG ) Debug.printtagln( tag, "no \"" +
					  SNAPSHOT_STR + "\" tag;" +
					  " disabling periodic snapshots");
	}

	emit_deltas = false;
	if( config.contains( DELTA_STR ) ) 
	    emit_deltas = config.getBoolean( DELTA_STR );
	if( DEBUG ) {
	    Debug.printtagln( tag, (emit_deltas ? "enabling" : "disabling")
			      + " perodic deltas" );
	}

	num_observed = 0;

	if( DEBUG ) Debug.printtagln( tag, "done: " + name );	
    }

    /**
     * Recalculates the model parameters, writes out snapshots, and
     * dispatches deltas.
     */
    private void recalculate()
    {
	final String tag = class_tag+".recalculate";
	if( DEBUG ) {
	    Debug.printtagln( tag, "Total observations:\t" + num_observed );
	}

	Model.Delta d = model.recalculate();

	if( d == null ) {
	    if( DEBUG )
		Debug.printtagln( tag, "Model produced a null delta" );
	} else if( emit_deltas ) {
	    // Models should be robust to dropped deltas.
	    classifier.dispatch_lossy( d );
	}
	
	if( write_snapshots ) {
	    try {
		// Write out the QuickSerializable form.
		QSIO.write( model, snapshot_name );
	    } catch( IOException ioe ) {
		// Print out an error no matter what.
		Carp.carp( "Could not write snapshot to file \""
			   + snapshot_name + "\"" );
		if( DEBUG ) Carp.error( ioe );
	    }
	}

	if( DEBUG ) Debug.printtagln( tag, "done" );
    }
	
    ////////////////////     INNER CLASSES	////////////////////
}
