/**
 * ThreshTest.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security.thresh;
import java.io.FileInputStream;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.Provider;
import java.security.Security;
import java.security.Signature;
import java.security.SignatureException;
import java.util.Random;
import ostore.security.QSPublicKey;
import ostore.security.SignedQS;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.QSByteArray;
import ostore.util.QSIO;
import ostore.util.TypeTable;

/**
 * Tests the threshold signature scheme (abeit not completely).
 *
 * <p>Reference: "Practical Threshold Signatures",
 * Victor Shoup (sho@zurich.ibm.com), IBM Research Paper RZ3121, 4/30/99
 *
 * @author  <a href=mailto:sweis@theory.lcs.mit.edu>Steve Weis</a> 
 * @version $Id: ThreshTest.java,v 1.11 2002/08/02 19:39:56 geels Exp $
 */
 class ThreshTest {

    public static void FAIL (String msg) {
	System.err.println (msg);
	System.exit (1);
    }

    public static void main(String args[]) 
    throws SignatureException, InvalidKeyException,
    TypeTable.NoSuchTypeCode {

	final String tag = "ThreshTest.main";

	Debug.turnOn ("ThresholdSignature");
	ostore.util.Types.registerAllTypes ();

	Security.addProvider (new cryptix.jce.provider.CryptixCrypto ());
	Security.addProvider (new ThresholdProvider ());

	ThresholdPublicKey public_key = null;
	String keyfilename = args [0];
	try {
	    FileInputStream keyfile = new FileInputStream (keyfilename);
	    public_key = (ThresholdPublicKey)  QSIO.read (keyfile);
	    keyfile.close ();
	}
	catch (Exception e) {
	    Carp.die (tag + ": caught exception " + e +
			" while trying to read pr keys from disk.");
	}

	KeyShare[] keys = new KeyShare [4];
	for (int i = 0; i < 4; ++i) {
	    String ksfilename = args [i+1];
	    try {
		FileInputStream keyfile = new FileInputStream (ksfilename);
		keys [i] = (KeyShare)  QSIO.read (keyfile);
		keyfile.close ();
	    }
	    catch (Exception e) {
		Carp.die (tag + ": caught exception " + e +
			" while trying to read " + ksfilename + " from disk.");
	    }
	}

	Dealer d = null;
	int k = 2;
	int l = 4;

	long qs_sig_type = TypeTable.type_code ((new ThresholdSignature ()));

	SignedQS signed_data = null;
	QSByteArray qsdata = null;
	// These are the indices of the shares
	int[] S = {0,2};
	{
	    // Generating a KB of random data and taking a hash
	    byte[] data = new byte[1024];
	    (new Random()).nextBytes(data);
	    qsdata = new QSByteArray (data);

	    // Sanity check. Make sure the hash is working..
	    // debug("Hash Value: " + new SafeBigInt(b));

	    debug("Attempting to verify a valid set of signatures...");
	    // Pick a set of shares to attempt to verify
	    SigShare[] sigs = new SigShare[k];

	    for (int i=0; i<k; i++)
		sigs[i] = ThresholdSignedQS.create_sig_share (
			keys[S[i]], qsdata, qs_sig_type);

	    byte [] signature = SigShare.package_up (sigs, public_key);
	    signed_data = new ThresholdSignedQS (qsdata, signature);
	}

	Signature engine = null;
	try {
	    engine = Signature.getInstance ("Threshold", "Threshold");
	}
	catch (Exception e) {
	    if (engine == null) Carp.die( "Caught exception " + e + 
		    " while instanciating signature " +
		    "engine.  Dying...");
	}
	if (engine == null) Carp.die( "Could not instanciate signature " +
		"engine.  Dying...");

	if (signed_data.verify (public_key, engine))
	    debug ("Signature Correctly Verified");
	else 
	    FAIL ("Signature Verification Failed!");

	debug ("----------");
	debug("Attempting to verify a different set of shares...");

	{
	    // Create k sigs to verify using different keys
	    int[] T = {2,3};
	    SigShare[] sigs = new SigShare[k];
	    for (int i=0; i<k; i++)
		sigs[i] = ThresholdSignedQS.create_sig_share (
			keys[T[i]], qsdata, qs_sig_type);

	    byte [] signature = SigShare.package_up (sigs, public_key);
	    signed_data = new ThresholdSignedQS (qsdata, signature);
	}

	if (signed_data.verify (public_key, engine))
	    debug ("Signature Correctly Verified");
	else 
	    FAIL ("Signature Verification Failed!");

	debug ("----------");

	debug("Attempting to identify an invalid set of signatures...");
	{
	    // Resign it correctly
	    int[] T = {2,3};
	    SigShare[] sigs = new SigShare[k];
	    for (int i=0; i<k; i++)
		sigs[i] = ThresholdSignedQS.create_sig_share (
			keys[T[i]], qsdata, qs_sig_type);
	    // Corrupt one signature
	    byte [] c = "corrupt data".getBytes();
	    sigs[1] = ThresholdSignedQS.create_sig_share (
		    keys[3], new QSByteArray (c), qs_sig_type);

	    byte [] signature = SigShare.package_up (sigs, public_key);
	    signed_data = new ThresholdSignedQS (qsdata, signature);
	}

	if (!signed_data.verify (public_key, engine))
	    debug ("Signature correctly did not verify bad sig");
	else
	    FAIL ("Signature should not have verified, but it did.");

	debug ("----------");
	debug("Running performance test..");
	int RUNS = 25, START = 5;

	long start = 0, elapsed;
	SigShare[] sigs = new SigShare[k];
	for (int i=0; i<RUNS; i++) {
	    if (i == START) start = System.currentTimeMillis();
	    sigs[0] = ThresholdSignedQS.create_sig_share (
		    keys[S[0]], qsdata, qs_sig_type);
	}
	elapsed = System.currentTimeMillis()-start;
	debug("\tSigning total ("+ (RUNS-START) + " sigs) (ms): "
	      + elapsed + " Average: " +
	      (((double) elapsed)/(RUNS-START)));

	for (int i=0; i<k; i++)
	    sigs[i] = ThresholdSignedQS.create_sig_share (
		    keys[S[i]], qsdata, qs_sig_type);

	{
	    byte [] signature = SigShare.package_up (sigs, public_key);
	    signed_data = new ThresholdSignedQS (qsdata, signature);
	}

	RUNS=100; START=20;
	for (int i=0; i<RUNS; i++) {
	    if (i == START) start = System.currentTimeMillis();
	    if (! signed_data.verify (public_key, engine))
		FAIL ("Sig Failed to verify correctly");
	}
	elapsed = System.currentTimeMillis()-start;
	debug("\tVerification total ("+ (RUNS-START)+" sigs) (ms): "
			   + elapsed + " Average: " +
			   (((double) elapsed)/(RUNS-START)));
	debug ("----------");

	System.exit(0);
    }

// Debugging
//............................................................................
   
    private final static boolean DEBUG = true;
   
    private static void debug(String s)
    {
	System.out.println("ThreshTest: " + s);
    }
}

