/**
 * PublishMsg.
 *
 * Internal messages for doing either publishing or unpublishing an object 
 * in the Tapestry network.  Whether it's publish or unpublish is indicated
 * by a boolean flag. Publish=true, unpublish=false. This internal message 
 * is routed between Tapestry nodes, where each overlay node examines the
 * message and does the correct thing w.r.t. their local object location
 * pointer map.
 *
 * @author     Ben Y. Zhao
 * @version    $Id: PublishMsg.java,v 1.21 2003/02/26 23:26:35 strib Exp $
 */

package ostore.tapestry.impl;
import ostore.network.NetworkMessage;
import ostore.tapestry.channel.MsgQos;
import ostore.tapestry.api.TapestryTag;
import ostore.util.NodeId;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;

public class PublishMsg 
  extends TapestryMessage implements QuickSerializable {

    public SecureHash src;
    public SecureHash guid;
    public TapestryTag tag;
    public int hop;
    public boolean publish; /*true if publishing, false if unpublishing*/

    /**
     * Maximum number of hops this msg can take.
     */
    public int TTL;

    /**
     * Local optimizations flags and time-to-live
     */
    public int local_opts;
    public int local_opts_ttl;

    //This is moved to TapestryMessage by Ling Huang 07/31/02.
    //Added by Ling Huang for reliable msg delivery.
    //    protected int qos;                //The QOS of this message, should it be reliable or un-.
    //    protected SecureHash msg_id;      //Unique identity for a message for retransmission.

    public PublishMsg (NodeId peer, SecureHash src, SecureHash guid,
	    TapestryTag tag, int hop, boolean publish) {
	super (peer, false /* outbound */);
	this.src = src;
	this.guid = guid;
	this.tag = tag;
	this.hop = hop;
	this.publish = publish;

	this.qos = MsgQos.UNRELIABLE;
	this.msg_id = null;

	// defaults: no optimization
	local_opts = 0;
	local_opts_ttl = 0;
   }

    public PublishMsg (InputBuffer buffer) throws QSException {
	super (buffer);
	src = (SecureHash) buffer.nextObject ();
	guid = (SecureHash) buffer.nextObject ();
	hop = buffer.nextInt ();
	publish = buffer.nextBoolean ();
	local_opts = buffer.nextInt ();
	// only if there is some optimization (saves space)
	if( local_opts > 0 ) {
	    local_opts_ttl = buffer.nextInt ();
	}
	if (buffer.nextBoolean ())
	    tag = (TapestryTag) buffer.nextObject ();

	/*This is moved to TapestryMessage by Ling Huang 07/31/02.
	qos = buffer.nextInt ();
	if (qos == MsgQos.RELIABLE)
	    msg_id = (SecureHash) buffer.nextObject ();
	*/
    }

    public void serialize (OutputBuffer buffer) {
	super.serialize (buffer);

	buffer.add (src);
	buffer.add (guid);
	buffer.add (hop);
	buffer.add (publish);
	buffer.add (local_opts);
	// only if there is some optimization (saves space)
	if( local_opts > 0 ) {
	    buffer.add (local_opts_ttl);
	}
	if (tag == null) 
	    buffer.add (false);
	else 
	{
	    buffer.add (true);
	    buffer.add (tag);
	}

	/*This is moved to TapestryMessage by Ling Huang 07/31/02.
	buffer.add (qos);
	if (qos == MsgQos.RELIABLE)
	    buffer.add (msg_id);
	*/
    }

    public Object clone () throws CloneNotSupportedException {
	PublishMsg result = (PublishMsg) super.clone ();
	result.src = src;
	result.guid = guid;
	result.tag = tag;
	result.hop = hop;
	result.publish = publish;

	result.local_opts = local_opts;
	result.local_opts_ttl = local_opts_ttl;

	//This is moved to TapestryMessage by Ling Huang 07/31/02.
	//	result.qos = qos;
	//	result.msg_id = msg_id;

	return result;
    }


  /*This is moved to TapestryMessage by Ling Huang 07/31/02.
    // Added by Ling Huang for msg retransmission.
    public int get_qos()                    {   return qos; }
    public boolean is_reliable()
    {
	if (qos == MsgQos.RELIABLE)
	    return true;
	return false;
    }
    public SecureHash get_msg_id()          {   return msg_id; }
    public void set_msg_id(SecureHash id)   {   msg_id = id; }
  */

    public void set_qos(int qos)
    {   
        this.qos = qos;
	  //	if (qos == MsgQos.RELIABLE && msg_id == null)
	if (msg_id == null)
	    generate_msg_id();
    }
    public SecureHash generate_msg_id()
    {
	double  id;
	String  id_string;

	id = java.lang.Math.random();
	id_string = src.toString() + guid.toString() + Double.toString(id);
	msg_id = new SHA1Hash(id_string);

	return msg_id;
    }
    public SecureHash generate_msg_id(Object id)
    {
	String  id_string;

	id_string = src.toString() + guid.toString() + id.toString();
	msg_id = new SHA1Hash(id_string);

	return msg_id;
    }
    

    public String toString () {
      if (publish)
      	return "(Publish" +
	    " hop=" + hop +
	    " guid=" + guid +
	    " tag=" + tag +
	    " src=" + src +
	    " to=" + peer +
	    " " + super.toString () + 
	    ")";
      else 
	return "(Unpublish" +
	    " hop=" + hop +
	    " guid=" + guid +
	    " tag=" + tag +
	    " src=" + src +
	    " " + super.toString () + 
	    ")";
    }
}











