/**
 * RouterFaultMsg.java
 *
 * @author Jeremy Stribling
 * @version    $Id: RouterFaultMsg.java,v 1.5 2003/02/05 22:35:04 hling Exp $
 */

package ostore.tapestry.patchwork;
import ostore.network.NetworkMessage;
import ostore.util.ByteUtils;
import ostore.util.NodeId;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.QSException;

/**
 * This is a message sent out by a testing stage
 * to inform a peer's Router to start dropping messages.
 *
 * @author Jeremy Stribling
 * @version    $Id: RouterFaultMsg.java,v 1.5 2003/02/05 22:35:04 hling Exp $
 */
public class RouterFaultMsg extends NetworkMessage 
    implements QuickSerializable {

    public boolean isPercent;

    public int value;

    public boolean allNodes;

    public NodeId nodeToDrop;


    /**
     * Constucts a fault injection message, which will let a peer
     * know that it should begin dropping messages.  Either an
     * absolute number of messages can be specified, or a percentage 
     * of messages for an indefinite amount of time.  Note that if
     * a percentage is given, it can be removed by sending another
     * fault injection message with the value at 0.  Any arriving
     * fault message overrides the previous one.
     *
     * @param peer  the node to contact
     * @param value the amount of messages to drop, either as a 
     *              number of messages or a percent of messages
     * @param isPercent  true if <code>value</code> specifies a percentage
     * @param nodeToDrop  Drop messages from this node, rather than the sender
     */
    public RouterFaultMsg (NodeId peer, int value, boolean isPercent, 
			   NodeId nodeToDrop) {

	super (peer, false /* outbound */);

	this.value = value;
	this.isPercent = isPercent;
	this.nodeToDrop = nodeToDrop; 

    }

    /**
     * Constucts a fault injection message, which will let a peer
     * know that it should begin dropping messages.  Either an
     * absolute number of messages can be specified, or a percentage 
     * of messages for an indefinite amount of time.  Note that if
     * a percentage is given, it can be removed by sending another
     * fault injection message with the value at 0.  Any arriving
     * fault message overrides the previous one.
     *
     * @param peer  the node to contact
     * @param value the amount of messages to drop, either as a 
     *              number of messages or a percent of messages
     * @param isPercent  true if <code>value</code> specifies a percentage
     * @param allNodes  true if the router should apply these faults to all 
     *                  nodes
     */
    public RouterFaultMsg (NodeId peer, int value, boolean isPercent, 
			   boolean allNodes) {

	super (peer, false /* outbound */);

	this.value = value;
	this.isPercent = isPercent;
	this.allNodes = allNodes; 

    }

    /**
     * Constucts a fault injection message, which will let a peer
     * know that it should begin dropping messages.  Either an
     * absolute number of messages can be specified, or a percentage 
     * of messages for an indefinite amount of time.  Note that if
     * a percentage is given, it can be removed by sending another
     * fault injection message with the value at 0.  Any arriving
     * fault message overrides the previous one.
     *
     * @param peer  the node to contact
     * @param value the amount of messages to drop, either as a 
     *              number of messages or a percent of messages
     * @param isPercent  true if <code>value</code> specifies a percentage
     */
    public RouterFaultMsg (NodeId peer, int value, boolean isPercent) {

	this (peer, value, isPercent, false);

    }

    public RouterFaultMsg (InputBuffer buffer) throws QSException {
	super( buffer );
	isPercent = buffer.nextBoolean();
	value = buffer.nextInt();
	allNodes = buffer.nextBoolean();
	if( buffer.nextBoolean() ) {
	    nodeToDrop = (NodeId) buffer.nextObject();
	}
    }

    public void serialize (OutputBuffer buffer) {
	super.serialize (buffer);
	buffer.add( isPercent );
	buffer.add( value );
	buffer.add( allNodes );
	if( nodeToDrop == null ) {
	    buffer.add( false );
	} else {
	    buffer.add( true );
	    buffer.add( nodeToDrop );
	}
    }


    public Object clone () throws CloneNotSupportedException {
	RouterFaultMsg newone = (RouterFaultMsg) super.clone ();
	newone.isPercent = isPercent;
	newone.value = value;
	return newone;
    }

    public String toString () {
	return "(RouterFaultMsg, value: " + 
		value + ", isPercent: " + isPercent + ", allNodes: "
	        + allNodes + ", nodeToDrop: " + nodeToDrop + " " 
		+ super.toString () + ")";
    }
}
