/**
 * InputBufferImpl.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.util;
import java.math.BigInteger;

/**
 * Implements all of the functionality of <code>InputBuffer</code> except
 * <code>nextByte</code> and <code>nextBytes</code>.  A subclass can
 * inherit from this class to implement an InputBuffer interface while only
 * having to define <code>nextByte</code> and <code>nextBytes</code>.
 *
 * @author	Sean C. Rhea
 * @version	$Id: InputBufferImpl.java,v 1.9 2002/09/09 05:04:15 srhea Exp $
 */
public abstract class InputBufferImpl implements InputBuffer {
    /**
     * Read the next short out of the buffer.
     */
    public short nextShort () { return nextShort (this); }

    /**
     * Read the next integer out of the buffer.
     */
    public int nextInt () { return nextInt (this); }

    /** 
     * Read the next long out of the buffer.
     */
    public long nextLong () { return nextLong (this); }

    /** 
     * Read the next double out of the buffer.
     */
    public double nextDouble () { return nextDouble (this); }

    /**
     * Read the next boolean out of the buffer.
     */
    public boolean nextBoolean () { return nextBoolean (this); }

    /**
     * Read the next String out of the buffer.
     */
    public String nextString () { return nextString (this); }

    /**
     * Read the next BigInteger out of the buffer.
     */
    public BigInteger nextBigInteger () {
	return nextBigInteger (this);
    }
    
    /**
     * Read the next object out of the buffer.
     */
    public QuickSerializable nextObject () throws QSException { 
	return nextObject (this); 
    }

    /**
     * Read the next integer out of the buffer.
     */
    public static final short nextShort (InputBuffer buffer) {
	short result = (short) ((short) 0xff & (short) buffer.nextByte ());
	result <<= 8;
	result |= (short) ((short) 0xff & (short) buffer.nextByte ());
	return result;
    }

    /**
     * Read the next integer out of the buffer.
     */
    public static final int nextInt (InputBuffer buffer) {
	int result = 0;
	for (int i = 0; i < 4; ++i)
	    result = (result << 8) | (0xff & (int) buffer.nextByte ());
	return result;
    }

    /** 
     * Read the next long out of the buffer.
     */
    public static final long nextLong (InputBuffer buffer) {
	long result = 0;
	for (int i = 0; i < 8; ++i)
	    result = (result << 8) | (0xff & (long) buffer.nextByte ());
	return result;
    }

    /** 
     * Read the next double out of the buffer.
     */
    public static final double nextDouble (InputBuffer buffer) {
	long result = nextLong(buffer);
	return Double.longBitsToDouble(result);
    }

    /**
     * Read the next boolean out of the buffer.
     */
    public static final boolean nextBoolean (InputBuffer buffer) {
	return buffer.nextByte () == 1;
    }

    /**
     * Read the next String out of the buffer.
     */
    public static final String nextString (InputBuffer buffer) {
        // TODO: get rid of copy.
	int length = buffer.nextInt ();
        if (length > 0)
        {
            byte [] data = new byte [length];
            buffer.nextBytes (data, 0, length);
            return new String (data, 0, length);
        }
        else
        {
            return new String();
        }
    }

    /**
     * Read the next BigInteger out of the buffer.
     */
    public static final BigInteger nextBigInteger (InputBuffer buffer) {
	int length = buffer.nextInt ();
	byte [] bytes = new byte [length];
	buffer.nextBytes (bytes, 0, length);
	return new BigInteger (bytes);
    }

     /**
     * Read the next object out of the buffer.
     */
    public static final QuickSerializable nextObject (InputBuffer buffer) 
    throws QSException {
	return TypeTable.new_object (buffer);

    }

    public Object clone () throws CloneNotSupportedException {
	// does nothing -- no fields to copy
	return super.clone ();
    }
}

