/**
 * SHA1HashBuffer.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.util;


import java.security.MessageDigest;
import java.security.DigestException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.math.BigInteger;

/**
 * A <code>SecureHashBuffer</code> that produces <code>SHA1Hash</code>s.
 *
 * @see SecureHashBuffer
 * @see SHA1Hash
 * @see java.security.MessageDigest
 *
 * @author	Dennis Geels
 * @version	$Id: SHA1HashBuffer.java,v 1.12 2002/08/28 03:20:19 hweather Exp $
 */

public class SHA1HashBuffer extends SecureHashBuffer
{
    ////////////////////       CONSTANTS	////////////////////
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The internal <code>MessageDigest</code> object, which does all the
     * real work.
     */
    MessageDigest md;
    
    ////////////////////  STATIC CONSTRUCTORS	////////////////////
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Constructs a new <code>SHA1HashBuffer</code>.
     */
    public SHA1HashBuffer()
    {
	try {
	    if(SHA1NativeIF.available)
		md = MessageDigest.getInstance("SHA", "SHA1Native");
	    else
		md = MessageDigest.getInstance("SHA", "SUN");

	} catch( NoSuchAlgorithmException nsae ) {
	    Carp.die( nsae );
	} catch (java.security.NoSuchProviderException nspe) {
	    Carp.die( nspe );
	}
    }
    
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Adds the specified byte to the digest.
     */
    public void add( byte input )
    {
	md.update( input );
    }

    /**
     * Adds <code>length</code> bytes of the specified array to the digest,
     * starting at <code>offset</code>.
     */
    public void add( byte[] input, int offset, int length )
    {
	md.update( input, offset, length );
    }

    /**
     * Computes and returns the digest of all inputs.  Also resets internal
     * state, so do <b>not</b> attempt to produce intermediate hashes by
     * calling this method prematurely.
     * <p>
     * This implementation always returns a {@link SHA1Hash}.
     */
    public SecureHash toHash()
    {
	return new SHA1Hash( this );
    }
    
    /**
     * Computes and returns the digest of all inputs.  Also resets internal
     * state.
     * <p>
     * <code>SHA1Hash</code> uses this method to produce efficient hashes.
     * No other class should call this method.
     */
    byte[] compute_bytes()
    {
	byte[] bytes = new byte[SHA1Hash.NUM_BYTES];
	try{
	    md.digest(bytes, 0, SHA1Hash.NUM_BYTES);
	} catch(DigestException de) {
	    Carp.die( de );
	}
	return bytes;
	//return md.engineDigest();
    }
	
    ////////////////////     INNER CLASSES	////////////////////
    
    ////////////////////       TEST CODE	////////////////////

    public static void main( String args[] )
    {
	//////////////////////////////
	// First test that creating hashes using this class works.
	Carp.out( "\nPhase 1\n" );
	
	// Create some objects to hash
	int zeros_len = 200;
	byte[] zeros = new byte[zeros_len];

	int trash_len = 200;
	byte[] trash = new byte[trash_len];
	java.util.Random rand = new java.util.Random( 0 );
	rand.nextBytes( trash );

	// TODO: test other primitives once we add some interesting add()
	// methods to SecureHashBuffer.
	//	QSDate now = new QSDate( Clock.current_date( null ) );
	int[] offset = {0};
	//	now.to_bytes( null, offset );
	//	int now_len = offset[0];

	// Now serialize them
//  	byte[] all = new byte[ zeros_len + trash_len + now_len ];
//  	ByteUtils.memcpy( all, 0, zeros, 0, zeros_len );
//  	ByteUtils.memcpy( all, zeros_len, trash, 0, trash_len );
//  	offset[0] = zeros_len + trash_len;
//  	now.to_bytes( all, offset );
	byte[] all = new byte[ zeros_len + trash_len  ];
	ByteUtils.memcpy( all, 0, zeros, 0, zeros_len );
	ByteUtils.memcpy( all, zeros_len, trash, 0, trash_len );
	offset[0] = zeros_len + trash_len;

	// Try three different ways to get the same hash
	SHA1Hash old_way = new SHA1Hash( all );
	Carp.out( "Old way: " + old_way );

	SHA1HashBuffer digest = new SHA1HashBuffer();
	digest.add( all );
	SHA1Hash digest_all = (SHA1Hash)digest.toHash();
	Carp.out( "All to digest: " + digest_all );
	Carp.out( "Equal (t)? " + digest_all.equals( old_way ) );
	if( !digest_all.equals( old_way ) ) {
	    Carp.die( "Doesn't match QuickSerializable hash." );
	}

	digest = new SHA1HashBuffer();
	digest.add( zeros );
	digest.add( trash, 0, trash_len );
	//	digest.add( now );
	SHA1Hash incremental = (SHA1Hash)digest.toHash();
	Carp.out( "Incremental digest: " + incremental );
	Carp.out( "Equal to last (t)? " + incremental.equals( digest_all ) );
	Carp.out( "Equal to old way (t)? " + incremental.equals( old_way ) );
	if( !incremental.equals( old_way ) ) {
	    Carp.die( "Piecewise doesn't match QuickSerializable hash." );
	}
	if( !incremental.equals( digest_all ) ) {
	    Carp.die( "Piecewise doesn't match hash of serialized form." );
	}

	//////////////////////////////	
	// Now test odd behavior
	Carp.out( "\nPhase 2\n" );
	
	// empty hashes:
	digest = new SHA1HashBuffer();
	SHA1Hash empty = (SHA1Hash)digest.toHash();
	Carp.out( "Empty digest:" + empty );
	
	// intermediate hashes:
	digest = new SHA1HashBuffer();
	digest.add( zeros );
	SHA1Hash zeros_only = (SHA1Hash)digest.toHash();
	Carp.out( "intermediate digest:" + zeros_only );
	SHA1Hash repeat = (SHA1Hash)digest.toHash();
	Carp.out( "repeat digest:" + repeat );
	Carp.out( "Equal to last (f)? " + repeat.equals( zeros_only ) );
	Carp.out( "Equal to empty (t)? " + repeat.equals( empty ) );
	if( repeat.equals( zeros_only ) || !repeat.equals( empty ) ) {
	    Carp.die( "Repeated hash didn't reset state." );
	}
	
	digest.add( trash, 0, trash_len );
	//	digest.add( now );
	SHA1HashBuffer half_digest = new SHA1HashBuffer();
	half_digest.add( trash, 0, trash_len );
	
	SHA1Hash continuing = (SHA1Hash)digest.toHash();
	SHA1Hash second_half = (SHA1Hash)half_digest.toHash();
	Carp.out( "final digest:" + continuing );
	Carp.out( "Equal to old way (f)? " + continuing.equals( old_way ) );
	Carp.out( "Equal to starting over (t)? " +
		  continuing.equals( second_half ) );
	if( continuing.equals( old_way ) ||
	    !continuing.equals( second_half) ) {
	    Carp.die( "Repeated hash didn't reset state." );
	}
	
	//////////////////////////////	
	// Now test new QuickSerializable stuff.
	Carp.out( "\nPhase 3\n" );
	
	final byte field_a = 4;
	final SecureHash field_b = new SHA1Hash( "foo" );
	QuickSerializable morsel = new VID( field_a, field_b );
	
	/* I can't think of a way I approve of doing this these days, and
	 * this class is fully tested by all the other regression tests, so
	 * I'm taking this section out. --srhea

	digest = new SHA1HashBuffer();
	digest.add( field_a );
	digest.add( TypeTable.type_code( field_b ) );
	digest.add( field_b );
	SHA1Hash piecewise = (SHA1Hash)digest.toHash();
	Carp.out( "Piecewise hash: " + piecewise );
	
	digest = new SHA1HashBuffer();
	digest.add( morsel );
	SHA1Hash whole_morsel = (SHA1Hash)digest.toHash();
	Carp.out( "Object digest: " + whole_morsel );
	Carp.out( "Equal to last (t)? " + whole_morsel.equals( piecewise ) );
	if( !whole_morsel.equals( piecewise ) ) {
	    Carp.die( "QuickSerializable class didn't digest properly" );
	}
	
	SHA1Hash from_constructor = new SHA1Hash( morsel );
	Carp.out( "From constructor: " + from_constructor );	
	Carp.out( "Equal to last (t)? " +
		  from_constructor.equals( whole_morsel ) );
	Carp.out( "Equal to piecewise (t)? " +
		  from_constructor.equals( piecewise ) );
	if( !from_constructor.equals( whole_morsel ) ||
	    !from_constructor.equals( piecewise ) ) {
	    Carp.die( "SHA1Hash( QuickSerializable ) constructor doesn't match." );
	}
	*/

	//////////////////////////////	
	// Test new support for primitives
	Carp.out( "\nPhase 4\n" );

	SHA1HashBuffer buffer = new SHA1HashBuffer ();
	buffer.add (0xdeadbeef);
	SecureHash hash1 = buffer.toHash ();

	SecureHash hash2 = null;
	{
	    int [] off = {0};
	    byte [] data = new byte [4];
	    ByteUtils.intToBytes (0xdeadbeef, data, off);
	    hash2 = new SHA1Hash (data);
	}

	if (! hash1.equals (hash2)) 
	    Carp.die ("Doesn't work for ints.");

	buffer.add (0xdeadbeef12345678L);
	hash1 = buffer.toHash ();

	{
	    int [] off = {0};
	    byte [] data = new byte [8];
	    ByteUtils.longToBytes (0xdeadbeef12345678L, data, off);
	    hash2 = new SHA1Hash (data);
	}

	if (! hash1.equals (hash2)) 
	    Carp.die ("Doesn't work for longs.");

	BigInteger i = new BigInteger ("123456789012345678901234567890");

	buffer.add (i);
	hash1 = buffer.toHash ();

	{
	    byte [] data = null;
	    while (true) {
		int [] off = {0};
		ByteUtils.bigIntegerToBytes (i, data, off);
		if (data == null)
		    data = new byte [off [0]];
		else
		    break;
	    }
	    hash2 = new SHA1Hash (data);
	}

	if (! hash1.equals (hash2)) 
	    Carp.die ("Doesn't work for big integers.");
	
	// Performance test

	byte [] lotsofbytes = new byte [1024*1024];
	rand.nextBytes (lotsofbytes);

	final int ITER = 100;
	long start = System.currentTimeMillis ();
	for (int j = 0; j < ITER; ++j) {
	    SHA1HashBuffer buf = new SHA1HashBuffer ();
	    buf.add (lotsofbytes);
	    buf.toHash ();
	}
	long stop = System.currentTimeMillis ();

	System.out.println ("Throughput = " +
		(((double) (ITER*lotsofbytes.length)) 
		 / (((double) (stop - start)) / 1000.0)) + " bytes/sec");

	Carp.out( "\nTest complete" );


    }	// end main()
}
