#!/usr/bin/perl -w
#
# $Id: call_iterator.pl,v 1.2 2002/11/13 22:57:57 geels Exp $

# This script serves as a back end for all the useful tools (call_remote,
# killall_java_remote, etc.) that execute a command on, or copy files to,
# every machine on the cluster, or in $OSTORE_SERVERS.

use strict;

BEGIN {
    # Find the directory with Tools.pm.  Should this be easier?
    my $run_dir = "$0";
    if( $run_dir !~ m%^/% ) {	# relative pathname
	$run_dir = $ENV{"PWD"} . "/$run_dir";
    } 
    $run_dir =~ s%/(./)?[^/]*$%%;	# strip off "call_iterator"
    $ENV{"OSTORE_RUN_DIR"} = $run_dir;
}

use lib $ENV{"OSTORE_RUN_DIR"};
use Tools;
use Getopt::Long;
$Getopt::Long::order = $REQUIRE_ORDER;	# Stop processing at command arguments..

####################
# Process options.

my %options;
&GetOptions( \%options, "help|?", "mode=s", "parallel|j=i", "remote-dir=s",
	     "tag=s", "force" )
    or &usage() and exit( -1 );

my $mode = $options{"mode"};
if( (! defined $mode) or (($mode ne "ssh") and ($mode ne "scp")) ) {
    warn "Missing or invalid mode\n" and &usage() and exit( -1 );
}

my $num_parallel = $options{"parallel"};

my $remote_dir;
if( $options{"remote-dir"} ) {
    if( $mode ne "scp" ) {
	warn "Invalid use of remote-dir\n" and &usage() and exit( -1 );
    }
    $remote_dir = $options{"remote-dir"};
} else {
    $remote_dir = "";
}

if( $options{"tag"} and $mode ne "ssh" ) {
    warn "Invalid use of tag\n" and &usage() and exit( -1 );
}

if( $options{"help"} ) {
    &usage() and exit( 0 );
}

if( ! @ARGV ) {
    warn "Missing script name\n" and &usage() and exit( -1 );
}
my $script_name = shift @ARGV;

# Rest is arguments to ssh or scp.
my @tool_args = @ARGV;
if( ! @tool_args ) {
    &usage and exit( -1 );
}

sub usage {
    print <<HERE;

  USAGE: call_iterator.pl --mode=<mode> [options] <script_name> <args>

  WARNING: You probably should not call this script directly.
	   See "call_remote.pl".

  REQUIRED OPTION:
	--mode			Specifies how to interpret <args>. One of:
		ssh		Run <args> as a command on each machine.
		scp		Copy <args> (as files) to each machine.

  OTHER OPTIONS:
	--help			Displays this help message.

	--force			Ignore errors and do not save logs.	    
 
	--parallel <num>	Fork a thread to run at most <num> commands
	    			simultaneously (-1 means unlimited).

	--remote-dir <dir>	Valid for "scp" mode only.  Names the destination
	    			directory on the remote machines.

	--tag <name>		Valid for "ssh" mode only.  Refer to command
	    			as <name>.
	    
  ARGUMENTS:
	script_name		Something like "call_remote"
	args			The command to run or files to copy.

HERE
}

####################
# Choose what machines to use.
my @machines;
if( $ENV{"OSTORE_SERVERS"} ) {
    my @raw_machines = split /\s+/, $ENV{"OSTORE_SERVERS"};
    my %added;
    foreach my $machine (@raw_machines) {
	next if( $added{$machine} );	# Prune out duplicates.
	push @machines, $machine;
	$added{$machine} = "true";
    }
} else {
    foreach( 1 .. 42 ) {
	push @machines, "x$_.millennium.berkeley.edu";
    }
}

####################
# Now loop across the machines and call the command.
my $exit_code = 0;

my @ssh_cmd = ("ssh");
my @scp_cmd = ("scp");
my $total = ($#machines+1);
my @child_pids;
my @child_logs;
for( my $i=0; $i < $total; $i++ ) {
    my $machine = $machines[$i];
    my $log = "$tmp_dir/$script_name-$mode-$$-$machine.log";
    my @ssh_args = &ssh_args( $machine );
    
    my $command_status;
    my @command;
    if( $mode eq "ssh" ) {
	@command = (@ssh_cmd, @ssh_args, $machine, @tool_args);
	my $command_name = ($options{"tag"} or $tool_args[0]);
	$command_status = "Calling $command_name on $machine";
    } else { # $mode eq "scp" 
	@command = (@scp_cmd, @ssh_args, @tool_args, "$machine:$remote_dir");
	$command_status = "Copying files to $machine:$remote_dir";
    }
    
    print "[" . ($i+1) . "/$total] $command_status\n";

    if( $options{"parallel"} ) {
	# Call command in forked thread.
	push @child_pids, &fork_process( $log, @command );
	push @child_logs, $log;

	if( ($num_parallel > 0) and ($num_parallel <= @child_pids) ) {
	    # too many in flight.  Wait for one to exit.
	    my $child_pid = shift @child_pids;
	    my $child_log = shift @child_logs;
	    if( (! &returned_zero( "--quiet", $child_pid )
		 or ! &is_clean( $child_log ))
		and ! $options{"force"} ) {
		$exit_code = -1;
		warn( "ERROR: See log $child_log\n" );
	    } else {
		unlink $child_log if( -f $child_log);
	    }
	}
	
    } else {
	# Just call inline.
	my $ret_code = system( @command );
	if( $ret_code != 0 and ! $options{"force"} ) {
	    $exit_code = -1;	    
	    warn( "ERROR: Failed on $machine.\n" );
	}
    }
}

####################
# Finally, Clean up any logs and forked threads.
if( ! $options{"parallel"} ) {
    # No forked threads.
} else {
    # Make sure the threads exit cleanly.
    for( my $i=0; $i <= $#child_pids; $i++ ) {
	my $child_pid = $child_pids[$i];
	my $child_log = $child_logs[$i];
	if( (! &returned_zero( "--quiet", $child_pid )
	     or ! &is_clean( $child_log ))
	    and ! $options{"force"} ) {
	    $exit_code = -1;
	    warn( "ERROR: See log $child_log\n" );
	} else {
	    unlink $child_log if( -f $child_log);
	}
    }
}
exit( $exit_code );
