#!/usr/bin/perl -w
#
# author: Dennis Geels
# $Id: extract_latencies.pl,v 1.10 2002/12/31 20:18:55 eaton Exp $
#
# Reads and correllates timestamped logs from several machines.
# All timestamps are adjusted to compensate for clock skew.

use strict;
use Getopt::Std;
use vars qw($opt_f $opt_g);

$| = 1;

sub usage {
    warn( "USAGE: extract_latencies.pl <log1> <host1> ... <logN> <hostN> \n" );
    warn( "       extract_latencies.pl -f <log1> ... <logN> \n" );
}

getopts('fg') or die( "could not read options" );

my $ignore_clock_skew = $opt_f;
my $compensate_for_gc = $opt_g;

##################################################
# Variables

# log name -> hostname map.
my %log_hosts;
    
# hostname -> clock skew map.
my %skews;

# per-request stats:
# Each request maps to a hash table with the following fields:
# "START" -> timestamp			- request sent
# "CATCH" -> ref of list of timestamps	- request received
# "REPLY" -> ref of list of timestamps	- response sent
# "ARRIVE" -> ref of list of timestamps	- response reaches replica.
# "DONE" -> ref of list of timestamps	- response completely processed.
# "FAIL" -> ref of list of timestamps	- rejection hint received.

# Update digest -> hash ref map.
my %updates;

# DTree join digest -> hash ref map.
my %joins;

# per-request stat summaries:
# Each summary is a hash table with the following fields:
# "CATCH", "REPLY", "ARRIVE", "DONE", and "FAIL".
# Each field is a reference to a list of the following form:
# ($count, $min, $max, $average, $std_dev )
#  Where all entries (other than $count) are computed over the time difference
#  between the named event ("CATCH", etc) and the corresponding "START" event.
my @stat_tags = qw( CATCH REPLY ARRIVE DONE FAIL );
my %update_stats;
my %join_stats;


# per-phase stats:
# Each request maps to a "START" and (possibly many) "END" fields.
my %join_phase_one;	# Request from new replica
my %join_phase_two;	# Parent replica processing
my %join_phase_three;	# Reply from parent
my %join_phase_four;	# Child processing reply

my %update_phase_one;	# Request from client to general
my %update_phase_two;	# Inner ring processing
my %update_phase_three;	# Root of tree to replica - subindexed by tree
my %update_phase_four;	# Update applied locally - subindexed by replica

my %read_phase_one;	# Opening replica
my %read_phase_two;	# Processing version predicate
my %read_phase_three;	# Waiting for ReadStage

##################################################
# Script

# Process arguments.
if( ! @ARGV ) {
    &usage() and exit(1);
} elsif( $ignore_clock_skew ) {
    # Only logs on command line.
    for (@ARGV) {
	$log_hosts{$_} = "who cares";
    }
} else {
    if( (@ARGV % 2) != 0 ) {
	&usage() and exit(1);
    } else {
	# non-empty, even list.  Read directly into hashtable.
	%log_hosts = @ARGV;
    }
}

unless( $ignore_clock_skew ) {
# find clock skews.
    open( SKEWS, "<clock_skew.txt" ) or
	die ("Cannot open ./clock_skew.txt.  Be sure to run calculate-skew.\n" );
    while( <SKEWS> ) {
	/^(\S*)\t(\S*)$/ or die( "Invalid clock skew entry: $_" );
	$skews{$1} = $2;
    }
    close( SKEWS );
}

# process logs.
my $file;
for $file (keys %log_hosts) {
    my $hostname = $log_hosts{$file};
    open( LOG, "<$file" ) or die ("Cannot open $file\n" );
#    print "Reading $file (from $hostname)\n";

    my $skew;
    unless( $ignore_clock_skew ) {
	$skew = $skews{$hostname};
	if( ! defined $skew ) {
	    die( "No entry for $hostname in clock_skew.txt\n" );
	}
    }
    
    while( <LOG> ) {
#	print "entry: $_";
	if( m/^			# start of line
	    (\d{13})		# timestamp
	    \s:(JOIN|UPDATE|READ)	# type
	    \s(\w+)		# op
	    \s\(SHA1Hash:([ 0-9a-f]{13})\)		# request digest
	    (\s\(SHA1Hash:([ 0-9a-f]{13})\))?	# local guid - optional
	    (\s\(SHA1Hash:([ 0-9a-f]{13})\))?	# source guid - optional
	    $/x ) {	# use extended regexp.
	    
	    my( $time, $type, $op, $digest, $here, $source )
		= ($1, $2, $3, $4, $5, $7);	# ignore $6, $8.
	    
	    $time -= $skew unless( $ignore_clock_skew );
	    &add_entry( $time, $type, $op, $digest, $here, $source );
	}
    }
}

# Now process the times we extracted from the logs.
#print "Processing...\n";

#&process_stats( \%joins, \%join_stats );
#&print_stats( \%join_stats );

#&process_stats( \%updates, \%update_stats );
#&print_stats( \%update_stats );

&process_phases();

##################################################
# METHODS

# Process a log entry
#
# usage: &add_entry( <timestamp>, <JOIN|UPDATE|READ>, <op>, <req digest>, \
#			[<local guid>, [<response source guid>]] );

sub add_entry {
    my( $time, $type, $op, $digest, $here, $source ) = @_;  

    #####
    # First save the per-request stats.
    my $request_table;
    if( $type eq "JOIN" ) {
	$request_table = \%joins;
    } elsif( $type eq "UPDATE" ) {
	$request_table = \%updates;
    } elsif( $type eq "READ" ) {
	# no per-requests stats.
    } else {
	die( "invalid type: $type\n" );
    }

    if( defined $request_table ) {
	my $request_stats = $$request_table{$digest};
	if( ! defined $request_stats ) {
	    $request_stats = {};
	    $$request_table{$digest} = $request_stats;
	}
	
	if( $op eq "START" ) {
	    if( defined $$request_stats{"START"} ) {
		die( "$digest started twice." );
	    }
	    $$request_stats{"START"} = $time;
	
	} elsif( ($op eq "CATCH") ||
		 ($op eq "REPLY") ||
		 ($op eq "ARRIVE") ||
		 ($op eq "DONE") ||
		 ($op eq "FAIL") ) {
	    
	    push @{$$request_stats{$op}}, $time;
	    
	} else {
	    die( "unknown op: $op\n" );
	}
    }

    #####
    # Now save per-phase stats.
    if( $type eq "UPDATE" ) {
	if( $op eq "START" ) {
	    $update_phase_one{$digest}{"START"} = $time;
	} elsif( $op eq "CATCH" ) {
	    $update_phase_one{$digest}{"END"} = $time;
	    $update_phase_two{$digest}{"START"} = $time;
	} elsif( $op eq "REPLY" ) {
	    push @{$update_phase_two{$digest}{"END"}}, $time;
	    $update_phase_three{$digest}{$here}{"START"} = $time;
	} elsif( $op eq "ARRIVE" ) {
	    push @{$update_phase_three{$digest}{$source}{"END"}}, $time;
	    $update_phase_four{$digest}{$here}{"START"} = $time;
	} elsif( $op eq "DONE" ) {
	    $update_phase_four{$digest}{$here}{"END"} = $time;
	}
    } elsif( $type eq "READ" ) {
	if( $op eq "START" ) {
	    $read_phase_one{$digest}{"START"} = $time;
	} elsif( $op eq "OPEN" ) {
	    $read_phase_one{$digest}{"END"} = $time;
	    $read_phase_two{$digest}{"START"} = $time;
	} elsif( $op eq "SEND" ) {
	    $read_phase_two{$digest}{"END"} = $time;
	    $read_phase_three{$digest}{"START"} = $time;
	} elsif( $op eq "DONE" ) {
	    $read_phase_three{$digest}{"END"} = $time;
	}
    }
    
}

# Process a table of requests stats
#
# usage: &process_stats( <req hash ref>, <summary hash ref> );

sub process_stats {
    my( $table, $summary_ref ) = @_;

    # accumulator lists
    my %accumulators;
    my $tag;

    # collect times across all requests.
    my $digest;
    for $digest (keys %$table) {
	my $start_time = $$table{$digest}{"START"};
	if( ! defined $start_time ) {
	    warn( "Request never started: $digest\n" );
	    next;
	}

	foreach $tag (@stat_tags) {
	    my $times_ref = $$table{$digest}{$tag};
	    push @{$accumulators{$tag}}, map( ($_ - $start_time), @$times_ref );
	}
    }
    
    # Now process each list.
    foreach $tag (@stat_tags) {
	if( $accumulators{$tag} ) {
	    my @stats = &process_list( $accumulators{$tag} );
	    $$summary_ref{$tag} = \@stats;
	}
    }
}

# Processes the per-phase stats
sub process_phases {

    my @update_phase_times;
    my $digest;

    # first process update phases.
    for $digest (keys %update_phase_one) {

	# phase 1
	my $start_time = $update_phase_one{$digest}{"START"};
	next if( ! defined $start_time );
	my $end_time = $update_phase_one{$digest}{"END"};
	push @{$update_phase_times[0]}, ($end_time - $start_time);

	# phase 2
	$start_time = $update_phase_two{$digest}{"START"};
	my @end_times = map( ($_ - $start_time),
			     @{$update_phase_two{$digest}{"END"}} );
	push @{$update_phase_times[1]}, @end_times;

	# phase 3
	my $tree;
	for $tree (keys %{$update_phase_three{$digest}}) {
	    $start_time = $update_phase_three{$digest}{$tree}{"START"};
	    @end_times = map( ($_ - $start_time),
			      @{$update_phase_three{$digest}{$tree}{"END"}} );
	    push @{$update_phase_times[2]}, @end_times;
	}
	
	# phase 4
	my $node;
	for $node (keys %{$update_phase_four{$digest}}) {
	    $start_time = $update_phase_four{$digest}{$node}{"START"};
	    $end_time = $update_phase_four{$digest}{$node}{"END"};
	    if( defined $end_time) {
		push @{$update_phase_times[3]}, ($end_time - $start_time);
	    }
	}
    }

    my $phase;
    for $phase (1..4) {
	print "Update phase $phase:\t";
	if( $update_phase_times[$phase-1] ) {
	    my @stats = &process_list( $update_phase_times[$phase-1] );
	    &print_list( \@stats );
	} else {
	    print "empty\n";
	}
    }


#     # Now process read phases
#     my @read_phase_times;
#     for $digest (keys %read_phase_one) {

# 	# phase 1
# 	my $start_time = $read_phase_one{$digest}{"START"};
# 	next if( ! defined $start_time );
# 	my $end_time = $read_phase_one{$digest}{"END"};
# 	push @{$read_phase_times[0]}, ($end_time - $start_time);

# 	# phase 2
# 	$start_time = $read_phase_two{$digest}{"START"};
# 	$end_time = $read_phase_two{$digest}{"END"};
# 	push @{$read_phase_times[1]}, ($end_time - $start_time);
	
# 	# phase 3
# 	$start_time = $read_phase_three{$digest}{"START"};
# 	$end_time = $read_phase_three{$digest}{"END"};
# 	push @{$read_phase_times[2]}, ($end_time - $start_time);
#     }

#     for $phase (1..3) {
# 	print "Read phase $phase:\t";
# 	if( $read_phase_times[$phase-1] ) {
# 	    my @stats = &process_list( $read_phase_times[$phase-1] );
# 	    &print_list( \@stats );
# 	} else {
# 	    print "empty\n";
# 	}
	
#     }

}


# Calculate the count, minimum, maximum, average, and standard deviation of
#  the numbers in a list.
# usage: ($cnt, $min, $max, $avg, $stddev, $ninetyfive, $ninetynine)
#		= &process_list( <list_ref> );

sub process_list {
    my $list_ref = shift;
    my( $count, $min, $med, $ninetyfive, $ninetynine, $max, $avg, $stddev );

    my @list = sort { $a <=> $b; } @$list_ref;
    $count = @list;

    if( $count > 0 ) {
	my $ninety_five_index = int(($#list+1)*0.95);
	my $ninety_nine_index = int(($#list+1)*0.99);
	$ninetyfive = $list[$ninety_five_index];
	$ninetynine = $list[$ninety_nine_index];
	if( ($compensate_for_gc) && ($ninety_nine_index > 0) ) {
	    splice @list, $ninety_nine_index;
	    $count = @list;
	}
	
	$min = $list[0];
	$med = $list[int ($#list/2)];
	$max = $list[$#list];
	
	my $num;
	my $sum = 0;
	foreach $num (@list) {
	    $sum += $num;
	}
	$avg = $sum/$count;
    
	if( $count > 1 ) {
	    $sum = 0;
	    foreach $num (@list) {
		$sum += ($num - $avg) * ($num - $avg);
	    }
	    $stddev = sqrt( $sum/($count-1) );
	}
    }
    return( $count, $min, $med, $max, $avg, $stddev, $ninetyfive, $ninetynine );
}


# Prints out the numbers returned by &process_list.
#
# usage: &print_stats( <summary_ref> );

sub print_stats {
    my $summary_ref = shift;
    my $tag;
    foreach $tag (@stat_tags) {
	my $list_ref = $$summary_ref{$tag};
	print "$tag: ";
	if( $list_ref ) {
	    &print_list( $list_ref );
	} else {
	    print "empty\n";
	}
    }
}

# Prints out a single set of statistics.
sub print_list {
    my $list_ref = shift;
    my($cnt, $min, $med, $max, $avg, $stddev, $ninetyfive, $ninetynine ) = @$list_ref;
    
    print "$cnt entries";
    if( $cnt > 0 ) {
	if( $cnt > 1 ) {
	    printf "\tMIN: %.3f\tMED: %.3f\t95: %.3f\t99: %.3f\tMAX: %.3f\tAVG: %.3f\tSTDDEV: %.3f\n",
	    $min, $med, $ninetyfive, $ninetynine, $max, $avg, $stddev;
	} else {
	    printf "\t%.3f\n", $med;
	}
    } else {
	print "\n";
    }
}
