#!/usr/bin/perl -w
#
# author: Jeremy Stribling
# $Id: ping_times.pl,v 1.7 2003/02/18 00:14:37 geels Exp $
#
# calculates the ping times to all hosts given as input
# usage: ping_times.pl <ping_count> <host1> ... <hostN>
# if no hosts are specified, defaults to $OSTORE_SERVERS

BEGIN {
    # Find the directory with Tools.pm.  Should this be easier?
    my $run_dir = "$0";
    if( $run_dir !~ m%^/% ) {	# relative pathname
	$run_dir = $ENV{"PWD"} . "/$run_dir";
    } 
    $run_dir =~ s%/(./)?[^/]*$%%;	# strip off "run-experiment"
    $ENV{"OSTORE_RUN_DIR"} = $run_dir;
}

use lib $ENV{"OSTORE_RUN_DIR"};
use strict;
use Tools;
use Traceroute;

my $PINGS_PER_SERVER = 10;
if( $#ARGV >= 0 ) {
    $PINGS_PER_SERVER = shift( @ARGV );
}

my $TIMEOUT_SECS = 120;
my $TMP_FILE = "$tmp_dir/pings.$$.tmp";

my @servers = @ARGV;
if( $#servers == -1 ) {print "\n";

    @servers = split(/\s+/, $ENV{"OSTORE_SERVERS"}) 
	if( $ENV{"OSTORE_SERVERS"} );
}
# if it's still empty, die
if( $#servers == -1 ) {
    die( "No host names provided, and \$OSTORE_SERVERS not set." );
}

# some ping programs don't support -w
my $deadline = "-w $TIMEOUT_SECS";
if( system( "ping -c 1 $deadline 127.0.0.1" ) ) {
    $deadline = "";
}

# let the pings begin
my @pingtimes;
my @original_indices = (0 .. $#servers);
for( my $i=@servers; $i>0; $i-- ) {

    # pings servers at random, so they don't get overloaded
    my $rand_server = int rand (@servers);
    my $server = splice @servers, $rand_server, 1;
    my $original_index = splice @original_indices, $rand_server, 1;

    print "$server\n";
    
    my $result = system( "ping -c $PINGS_PER_SERVER $deadline " . 
			 "$server &> $TMP_FILE" );

    if( $result != 0 ) {
	#clean_death( "Could not ping $server" );

	# instead, issue a traceroute to that server and see where 
	# it failed
	my $tr = Net::Traceroute->new(host=> $server,
				      trace_program=> "/usr/sbin/traceroute",
				      max_ttl=> 30,
				      queries=> 1,
				      query_timeout=> 5);
	if(defined($tr->found) and $tr->stat == $tr->TRACEROUTE_OK and
	   $tr->hop_query_host($tr->hops, 0) eq $server) {
	    # somehow we found it, even though we couldn't ping it
	    # record the traceroute round trip time
	    my $rtt = $tr->hop_query_time( $tr->hops, 0 );
	    my $val = "$rtt/$rtt/$rtt";
	    $pingtimes[$original_index] = $val;
	} else {
	    # return the last hop
	    my $hop = 1;
	    while( defined $tr->hop_query_host( $hop, 0 ) ) {
		$hop++;
	    }
	    my $last_hop = $tr->hop_query_host($hop-1, 0);
	    $pingtimes[$original_index] = "***$last_hop***";
	}
	next;
    }

    # now parse the results
    open( TMP, "<$TMP_FILE" ) or clean_death( "Could not open $TMP_FILE" );

    my $line;
    foreach $line (<TMP>) {

	# split first by the equals sign, then by the /
	my @equalstats = split( /=/, $line );
	
	if( $#equalstats < 1 ) {
	    next;
	}

	my @pingstats = split( /\//, $equalstats[1] );
	if( $#pingstats > 0 ) {
	    my $val = "$pingstats[0]/$pingstats[1]/$pingstats[2]";
	    # clear out any whitespace, newlines, or "ms" 
	    $val =~ s/\s//;
	    $val =~ s/\n//;
	    $val =~ s/ms//;
	    $pingtimes[$original_index] = $val;
	}
    }

    close( TMP );
    
}

unlink( $TMP_FILE );

# now print em out and we're done
my $time;
foreach $time (@pingtimes) {
    print "$time ";
}
print "\n";

################

sub clean_death {
    my $death_msg = shift;
    unlink( $TMP_FILE );
    print $death_msg . "\n";
    exit( 1 );
}
