#!/usr/bin/perl -w

use strict;
use Getopt::Std;
use File::Copy;
use Tools;
use vars qw($opt_h $opt_a $opt_c $opt_e $opt_k $opt_n $opt_s $opt_x $opt_1 $opt_2 $opt_3);

sub usage
{
    select(STDERR);
    print <<'EOUsage';

run-tput-test.pl [options]
  Options:
    -h           - Display this help message.
    -c MACHINES  - Pass the token among MACHINES machines.
    -n NODES     - Create a network of NODES nodes.  If this number is
                   greater than the number of token-passing nodes, the excess
                   nodes are assigned to be passive tapestry nodes.
    -x ROC	 - Relocate the test in the X-cluster to start using
                   machine xROC.millennium.berkeley.edu.
		   NOTE: if no -x option is specified, the test will use
		   the machines listed in the OSTORE_SERVERS environment
		   variable.
    -s START     - Start the measurement phase when the leader receives
                   the token the STARTth time.  The default value of
                   START is 5.
    -e END       - Terminate the test when the leader receives the
                   token the ENDth time.  The default value of END is 25.
    -k KEY_SIZE  - Use keys of of length KEY_SIZE bits.  Valid values
                   are 512 and 1024.  The default value is 512.
    -a   	 - Turn the archive on (w/dissemination and everything)
    -1           - Do not use OceanStore.  Instead run a comparison
                   system that passes the token using point-to-point
                   communication.
    -2           - Do not use OceanStore.  Instead run a comparison
                   system that passes the token through the master.
                   The node hosting the primary copy notifies the new 
		   holder of the token.
    -3           - Do not use OceanStore.  Instead run a comparison
                   system that passes the token by updating a primary
                   copy.  Clients must query the primary copy to
                   learn who holds the token.

EOUsage

    exit(1);
}

# Set a nice signal handler.
$SIG{INT} = \&slow_death;
$SIG{QUIT} = \&slow_death;
$SIG{HUP} = \&slow_death;
$SIG{PIPE} = \&slow_death;
$SIG{TERM} = \&slow_death;

#  Constants.
my $TAG_RESULT_LOG;
my $PEER_MACHINE_OFFSET = 5;

#  Test parameters.
my $SKIP_ARCHIVE = 'true';
my $TOTAL_NODES;
my $MODE = -1;
my $TEST_MACHINES = 1;
my $START = 5;
my $END = 25;
my @SERVERS = split /\s+/, $ENV{"OSTORE_SERVERS"} if ($ENV{"OSTORE_SERVERS"});;
my $KEY_SIZE = 512;

my $run_experiment_pid;


#  Get the user-defined parameters.
my $opts_ok = getopts('habc:e:k:n:s:x:123');
if ($opt_a) { $SKIP_ARCHIVE = 'false'; }
if ($opt_h) { &usage; }
if ($opt_c) { $TEST_MACHINES = $opt_c; }
if ($opt_e) { $END = $opt_e; }
if ($opt_s) { $START = $opt_s; }
if ($opt_x) { @SERVERS = map "x$_.millennium.berkeley.edu", ($opt_x .. 42); }
if ($opt_k) { $KEY_SIZE = $opt_k; }
if ($opt_n) { $TOTAL_NODES = $opt_n; } else { $TOTAL_NODES = $TEST_MACHINES; }
if ($opt_1) { $MODE = 1; }
if ($opt_2) { $MODE = 2; }
if ($opt_3) { $MODE = 3; }

#  The command line is valid if getopts returned without error and
#  there are no arguments left over.  $#ARGV returns the number of
#  remaining arguments minus one.
if ((! $opts_ok) || ($#ARGV != -1)) { &usage; }

#  Check that the set of input parameters is valid.
if ($START < 1 || $END <= $START)
{
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Invalid START and END parameters.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}

if ($KEY_SIZE != 512 && $KEY_SIZE != 1024)
{
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Invalid key size.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}

if( ! @SERVERS ) {
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Undefined cluster machine.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}

#  Compute the result log name.
$TAG_RESULT_LOG = "$Tools::tmp_dir/tag_result_log.$$.log";

#  Set the environment variable for the keys.
my $key_cmd;
if ($KEY_SIZE == 512)
{
    $ENV{KEYSHARE_PREFIX} = "ir";
    $key_cmd = "KEYSHARE_PREFIX=ir";
}
elsif ($KEY_SIZE == 1024)
{
    $ENV{KEYSHARE_PREFIX} = "ir-1024-4-2";
    $key_cmd = "KEYSHARE_PREFIX=ir-1024-4-2";
}
else
{
    print(STDERR "Script error.  Invalid key length.");
}

#  Create configuration files.
my @peer_cfgs    = &write_peer_config_files($TEST_MACHINES);
my $passive_cfg	 = &write_passive_config_file();
my $exp_cfg      = &write_experiment_file($TEST_MACHINES, $TOTAL_NODES, $passive_cfg);
my $tag_fork_log = "$Tools::tmp_dir/tag_fork_log.$$.log";

#  Run the experiment.
print("Running tag benchmark: $key_cmd ./run-experiment $exp_cfg\n" .
      "  TestMachines = $TEST_MACHINES\n" .
      "  TotalMachines = $TOTAL_NODES\n" .
      "  StartMeasure = $START\n" .
      "  EndMeausre   = $END\n" .
      "  KeySize      = $KEY_SIZE\n");

$run_experiment_pid = &fork_process($tag_fork_log, "./run-experiment", $exp_cfg);
if(! &returned_zero($run_experiment_pid))
{
    die("Experiment returned non-zero.\n");
}

#  Read the result log.
my $result_log_name = &get_result_log_name($run_experiment_pid);
my @results         = &analyze_result_log($result_log_name);

#  Compute the results.
my $time           = $results[2] - $results[0];
my $tokens         = $results[3] - $results[1];
my $time_per_token = $time / $tokens;
print("  Time         = $time ms\n" .
      "  Tokens       = $tokens\n" .
      "  Time/Token   = $time_per_token ms/token\n");

#  Write to log.
&write_result_to_log($TEST_MACHINES, $START, $END, $KEY_SIZE,
                     $time, $tokens, $time_per_token);

#  Remove the files we created.
unlink(@peer_cfgs, $exp_cfg, $tag_fork_log, $passive_cfg,
       $result_log_name );

exit(0);


sub write_result_to_log
{
    my $peers          = shift(@_);
    my $start          = shift(@_);
    my $end            = shift(@_);
    my $key_size       = shift(@_);
    my $time           = shift(@_);
    my $tokens         = shift(@_);
    my $time_per_token = shift(@_);

    my $log_msg =
      "Peers=${peers} " .
      "StartMeasure=${start} " .
      "EndMeasure=${end} " .
      "KeySize=${key_size} " .
      "Time=${time} ms " .
      "Tokens=${tokens} " .
      "Time/Token=${time_per_token} ms/token\n";

    write_to_log($log_msg);

    return;
}

sub write_to_log
{
    my $log_msg = shift(@_);

    open(RESULT_LOG, ">> $TAG_RESULT_LOG") ||
      die("Could not open result log ($TAG_RESULT_LOG).");
    print(RESULT_LOG "$log_msg");
    close(RESULT_LOG);

    return;
}

sub get_result_log_name
{
    my $pid = shift(@_);

    my @internal_logs = glob("$Tools::tmp_dir/*-${pid}*-internal.log-tag.log");
    (scalar(@internal_logs) == 1) ||
      die("Cannot determine result log name.\n");
    my $log = shift(@internal_logs);
    return $log;
}

sub analyze_result_log
{
    my $log = shift(@_);

    open(RESULT_LOG, "$log") || die("Could not open result log.\n");

    my $start_time;
    my $start_token;
    my $end_time;
    my $end_token;
    while(<RESULT_LOG>)
    {
        if (/StartTime=(.*)/)
        {
            $start_time= $1;
        }
        if (/StartToken=(.*)/)
        {
            $start_token = $1;
        }
        if (/EndTime=(.*)/)
        {
            $end_time = $1;
        }
        if (/EndToken=(.*)/)
        {
            $end_token = $1;
        }
    }

    close(RESULT_LOG);

    my @results;
    push(@results, $start_time);
    push(@results, $start_token);
    push(@results, $end_time);
    push(@results, $end_token);

    return @results;
}

sub write_experiment_file
{
    my( $test_machines, $total_nodes, $passive_cfg ) = @_;
    
    my $exp_cfg = "$Tools::tmp_dir/tag.$$.exp";
    open(EXP, ">$exp_cfg") ||
      die("Could not create experiment configuration file ($exp_cfg).\n");
    my $old_fd = select(EXP);
    my $fed = 0;
    my $ir0 = 1;
    my $ir1 = 2;
    my $ir2 = 3;
    my $ir3 = 4;

    print <<EOExp;
hosts      @SERVERS
topology   NONE
Federation \${run_dir}/federation.cfg  fed     daemon $fed
RP         \${run_dir}/rp_replica.cfg  static  daemon $fed
Ring0      \${run_dir}/inner_replica.cfg       static  daemon $ir0
Ring1      \${run_dir}/inner_replica.cfg       static  daemon $ir1
Ring2      \${run_dir}/inner_replica.cfg       static  daemon $ir2
Ring3      \${run_dir}/inner_replica.cfg       static  daemon $ir3
EOExp

    my $i;
    for ($i = 0; $i < $test_machines; $i++)
    {
        my $peer_cfg = "$Tools::tmp_dir/tag-peer-$i.$$.cfg";
        my $machine_num = $PEER_MACHINE_OFFSET + $i;
        if ($i == 0)
        {
            print("Peer$i $peer_cfg static mortal $machine_num\n");
        }
        else
        {
            print("Peer$i $peer_cfg static daemon $machine_num\n");
        }
    }
    # Fill out Tapestry network:
    for( ; $i < $total_nodes; $i++ )
    {
        my $machine_num = $PEER_MACHINE_OFFSET + $i;
	print( "Tapestry$i $passive_cfg static daemon $machine_num\n" );
    }

    # Finally, override a few runtime parameters:
    print <<EOOverrides;    
set	Ring*.Inner.skip_archive 		$SKIP_ARCHIVE
set	Ring*.DisseminatorStage.AlarmTimeout 	3600000
set	*.RecoalesceBlkStage.AlarmTimeout 	3600000
set	*.DTree.flood_publish			true
set	*.ReplicaStage.apply_updates 		true
set	*.Router.local_opt_nearest		30
set	*.TClient.all_pairs_pings_file	\${allPairsPingsFile}
set	*.TClient.remember_all_static_nodes	true
EOOverrides
    
    close(EXP);
    select($old_fd);

    return $exp_cfg;
}

# Like a peer config file, but no TagMicroBenchmark stage.
sub write_passive_config_file
{
    my $passive_cfg = "$Tools::tmp_dir/tag-passive.$$.cfg";
    open(PASSIVE, ">$passive_cfg") ||
	die("Could not create passive configuration file ($passive_cfg).\n");

    my $old_fd = select(PASSIVE);

    &write_peer_header();
    &write_peer_footer();

    close(PASSIVE);
    select($old_fd);
    return $passive_cfg;
}

sub write_peer_config_files
{
    my $test_machines = shift(@_);

    my $peers;
    for (my $i = 0; $i < $test_machines; $i++)
    {
	$peers .= "\${Peer${i}ID},";
    }

    my @peer_config_files;
    for (my $i = 0; $i < $test_machines; $i++)
    {
        my $peer_cfg = "$Tools::tmp_dir/tag-peer-$i.$$.cfg";
        push(@peer_config_files, $peer_cfg);
        open(PEER, ">$peer_cfg") ||
          die("Could not create peer configuration file ($peer_cfg).\n");

        my $old_fd = select(PEER);

        &write_peer_header();
        if ($MODE != -1)
        {
            &write_base_peer_stage($peers);
        }
        else
        {
            &write_peer_stage($peers);
        }
        &write_peer_footer();

        close(PEER);
        select($old_fd);
    }

    return @peer_config_files;
}

sub write_peer_header
{
    print <<'EOHeader';
<sandstorm>
    <!include Generic.hdr>    
    <stages>
	<!include Network.stg>
    
        <!include Router_fake_keys.stg>
	
	<!include StaticTClient.stg>
	    
	<!include Client.stg>

	<!include Replica.stg>

	<!include Read.stg>

	<!include RecoalesceBlk.stg>

	<!include Update.stg>

	<!include DTreeNode.stg>

	<!include Requestor.stg>

	<!include MemoryCache.stg>
	
	<!include ArchivalStorage.stg>

EOHeader
}

sub write_base_peer_stage
{
    my $peers = shift(@_);

    print <<EOStage;
  <TagMicroBenchmark>
    class ostore.apps.test.TagMicroBenchmarkControl
    queueThreshold 1000

    <initargs>
      Mode               $MODE
      Peers              $peers
      StartMark          $START
      EndMark            $END
      ResultLog          \${InternalLog}-tag.log

      rp_pkeyfile        \${RPPublicKey}
      principal_pkeyfile \${PrincipalPublicKey}
      principal_skeyfile \${PrincipalPrivateKey}
    </initargs>
  </TagMicroBenchmark>
EOStage
}

sub write_peer_stage
{
    my $peers = shift(@_);

    print <<EOStage;
  <TagMicroBenchmark>
    class ostore.apps.test.TagMicroBenchmark
    queueThreshold 1000

    <initargs>
      Peers              $peers
      StartMark          $START
      EndMark            $END
      ResultLog          \${InternalLog}-tag.log

      rp_pkeyfile        \${RPPublicKey}
      principal_pkeyfile \${PrincipalPublicKey}
      principal_skeyfile \${PrincipalPrivateKey}
    </initargs>
  </TagMicroBenchmark>
EOStage
}

sub write_peer_footer
{
    print <<EOFooter;
</stages>
</sandstorm>
EOFooter
}


# try to kill run-experiment nicely.
sub slow_death {
    print(  "*** Caught SIG@_ ***\n" );
    if( defined $run_experiment_pid ) {
	print "  Killing run-experiment...\n";
	kill 2, $run_experiment_pid;
	# Now wait for it to die.
	&returned_zero( $run_experiment_pid );
    }
    &exit_all( 1 );
}

